@@ -7,8 +7,8 @@ import platform
 import sys
 
 if '-lboost_python-mt' == '':
-	print 'You need to pass --enable-python-binding to configure in order ',
-	print 'to properly use this setup. There is no boost.python library configured now'
+	print('You need to pass --enable-python-binding to configure in order '),
+	print('to properly use this setup. There is no boost.python library configured now')
 	sys.exit(1)
 
 def parse_cmd(cmdline, prefix, keep_prefix = False):
@@ -33,7 +33,7 @@ if platform.system() == 'Windows':
 	# msvc 9.0 (2008) is the official windows compiler for python 2.6
 	# http://docs.python.org/whatsnew/2.6.html#build-and-c-api-changes
 	if os.system('bjam boost=source link=static geoip=static boost-link=static release msvc-9.0 optimization=space') != 0:
-		print 'build failed'
+		print('build failed')
 		sys.exit(1)
 	try: os.mkdir(r'build')
 	except: pass
@@ -43,7 +43,7 @@ if platform.system() == 'Windows':
 	except: pass
 	shutil.copyfile(r'bin\msvc-9.0\release\boost-source\geoip-static\link-static\optimization-space\threading-multi\libtorrent.pyd', r'.\build\lib\libtorrent.pyd')
 	setup( name='python-libtorrent',
-		version='0.15.10',
+		version='0.16.3',
 		author = 'Arvid Norberg',
 		author_email='arvid@cs.umu.se',
 		description = 'Python bindings for libtorrent-rasterbar',
@@ -64,10 +64,10 @@ if "OPT" in config_vars and "-Wstrict-prototypes" in config_vars["OPT"]:
 source_list = os.listdir(os.path.join(os.path.dirname(__file__), "src"))
 source_list = [os.path.join("src", s) for s in source_list if s.endswith(".cpp")]
 
-extra_cmd = ' -DTORRENT_USE_OPENSSL  -DWITH_SHIPPED_GEOIP_H  -DBOOST_ASIO_HASH_MAP_BUCKETS=1021  -DBOOST_EXCEPTION_DISABLE  -DBOOST_FILESYSTEM_VERSION=2  -DTORRENT_LINKING_SHARED  -I/opt/local/include    -lboost_system-mt -lboost_filesystem-mt -lboost_thread-mt -lboost_python-mt  -lssl -lcrypto -lz   -L/opt/local/lib   -I/opt/local/include  '
+extra_cmd = ' -DTORRENT_USE_OPENSSL  -DWITH_SHIPPED_GEOIP_H  -DBOOST_ASIO_HASH_MAP_BUCKETS=1021  -DBOOST_EXCEPTION_DISABLE  -DBOOST_ASIO_ENABLE_CANCELIO  -DBOOST_ASIO_DYN_LINK -DTORRENT_LINKING_SHARED  -I/opt/local/include    -lboost_system-mt -lboost_python-mt  -lssl -lcrypto   -L/opt/local/lib   -I/opt/local/include  '
 
 setup( name='python-libtorrent',
-	version='0.15.10',
+	version='0.16.3',
 	author = 'Arvid Norberg',
 	author_email='arvid@cs.umu.se',
 	description = 'Python bindings for libtorrent-rasterbar',
