@@ -13,19 +13,19 @@ ses.listen_on(6881, 6891)
 
 info = lt.torrent_info(sys.argv[1])
 h = ses.add_torrent({'ti': info, 'save_path': './'})
-print 'starting', h.name()
+print('starting', h.name())
 
 while (not h.is_seed()):
 	s = h.status()
 
 	state_str = ['queued', 'checking', 'downloading metadata', \
 		'downloading', 'finished', 'seeding', 'allocating', 'checking fastresume']
-	print '\r%.2f%% complete (down: %.1f kb/s up: %.1f kB/s peers: %d) %s' % \
+	print('\r%.2f%% complete (down: %.1f kb/s up: %.1f kB/s peers: %d) %s' % \
 		(s.progress * 100, s.download_rate / 1000, s.upload_rate / 1000, \
-		s.num_peers, state_str[s.state]),
+		s.num_peers, state_str[s.state]), end=' ')
 	sys.stdout.flush()
 
 	time.sleep(1)
 
-print h.name(), 'complete'
+print(h.name(), 'complete')
 
