@@ -37,6 +37,11 @@ std::string dht_announce_alert_ip(dht_announce_alert const& pa)
     return pa.ip.to_string(ec);
 }
 
+tuple incoming_connection_alert_ip(incoming_connection_alert const& ica)
+{
+    return endpoint_to_tuple(ica.ip);
+}
+
 list stats_alert_transferred(stats_alert const& alert)
 {
    list result;
@@ -46,6 +51,18 @@ list stats_alert_transferred(stats_alert const& alert)
    return result;
 }
 
+list get_status_from_update_alert(state_update_alert const& alert)
+{
+   list result;
+
+   for (std::vector<torrent_status>::const_iterator i = alert.status.begin(); i != alert.status.end(); ++i)
+   {
+      result.append(*i);
+   }
+   return result;
+}
+
+
 void bind_alert()
 {
     using boost::noncopyable;
@@ -99,6 +116,15 @@ void bind_alert()
         .def_readonly("url", &tracker_alert::url)
         ;
 
+    class_<torrent_added_alert, bases<torrent_alert>, noncopyable>(
+        "torrent_added_alert", no_init)
+        ;
+
+    class_<torrent_removed_alert, bases<torrent_alert>, noncopyable>(
+        "torrent_removed_alert", no_init)
+        .def_readonly("info_hash", &torrent_removed_alert::info_hash) 
+        ;
+
     class_<read_piece_alert, bases<torrent_alert>, noncopyable>(
         "read_piece_alert", 0, no_init)
         .add_property("buffer", get_buffer)
@@ -116,6 +142,7 @@ void bind_alert()
         .def_readonly("msg", &tracker_error_alert::msg)
         .def_readonly("times_in_row", &tracker_error_alert::times_in_row)
         .def_readonly("status_code", &tracker_error_alert::status_code)
+        .def_readonly("error", &tracker_error_alert::error)
         ;
 
     class_<tracker_warning_alert, bases<tracker_alert>, noncopyable>(
@@ -233,9 +260,10 @@ void bind_alert()
     class_<portmap_error_alert, bases<alert>, noncopyable>(
         "portmap_error_alert", no_init)
         .def_readonly("mapping", &portmap_error_alert::mapping)
-        .def_readonly("type", &portmap_error_alert::type)
+        .def_readonly("map_type", &portmap_error_alert::map_type)
         .def_readonly("error", &portmap_error_alert::error)
 #ifndef TORRENT_NO_DEPRECATE
+        .def_readonly("type", &portmap_error_alert::map_type)
         .def_readonly("msg", &portmap_error_alert::msg)
 #endif
         ;
@@ -244,13 +272,17 @@ void bind_alert()
         "portmap_alert", no_init)
         .def_readonly("mapping", &portmap_alert::mapping)
         .def_readonly("external_port", &portmap_alert::external_port)
-        .def_readonly("type", &portmap_alert::type)
+#ifndef TORRENT_NO_DEPRECATE
+        .def_readonly("type", &portmap_alert::map_type)
+#endif
+        .def_readonly("map_type", &portmap_alert::map_type)
         ;
 
     class_<portmap_log_alert, bases<alert>, noncopyable>(
         "portmap_log_alert", no_init)
-        .def_readonly("type", &portmap_log_alert::type)
+        .def_readonly("map_type", &portmap_log_alert::map_type)
 #ifndef TORRENT_NO_DEPRECATE
+        .def_readonly("type", &portmap_log_alert::map_type)
         .def_readonly("msg", &portmap_log_alert::msg)
 #endif
         ;
@@ -275,7 +307,8 @@ void bind_alert()
         ;
 
     class_<scrape_failed_alert, bases<tracker_alert>, noncopyable>(
-        "scrape_failed_alert", no_init);
+        "scrape_failed_alert", no_init)
+        ;
 
     class_<udp_error_alert, bases<alert>, noncopyable>(
         "udp_error_alert", no_init)
@@ -311,7 +344,8 @@ void bind_alert()
         ;
 
     class_<torrent_resumed_alert, bases<torrent_alert>, noncopyable>(
-        "torrent_resumed_alert", no_init);
+        "torrent_resumed_alert", no_init
+    );
 
     class_<state_changed_alert, bases<torrent_alert>, noncopyable>(
         "state_changed_alert", no_init)
@@ -319,6 +353,11 @@ void bind_alert()
         .def_readonly("prev_state", &state_changed_alert::prev_state)
         ;
 
+    class_<state_update_alert, bases<alert>, noncopyable>(
+        "state_update_alert", no_init)
+        .add_property("status", &get_status_from_update_alert)
+        ;
+
     class_<dht_reply_alert, bases<tracker_alert>, noncopyable>(
         "dht_reply_alert", no_init)
         .def_readonly("num_peers", &dht_reply_alert::num_peers)
@@ -332,18 +371,22 @@ void bind_alert()
     ;
 
     class_<dht_get_peers_alert, bases<alert>, noncopyable>(
-        "dht_get_peers_alert", no_init)
+        "dht_get_peers_alert", no_init
+    )
         .def_readonly("info_hash", &dht_get_peers_alert::info_hash)
     ;
 
     class_<peer_unsnubbed_alert, bases<peer_alert>, noncopyable>(
-        "peer_unsnubbed_alert", no_init);
+        "peer_unsnubbed_alert", no_init
+    );
 
     class_<peer_snubbed_alert, bases<peer_alert>, noncopyable>(
-        "peer_snubbed_alert", no_init);
+        "peer_snubbed_alert", no_init
+    );
 
     class_<peer_connect_alert, bases<peer_alert>, noncopyable>(
-        "peer_connect_alert", no_init);
+        "peer_connect_alert", no_init
+    );
 
     class_<peer_disconnected_alert, bases<peer_alert>, noncopyable>(
         "peer_disconnected_alert", no_init)
@@ -402,7 +445,7 @@ void bind_alert()
         "stats_alert", no_init)
         	.add_property("transferred", &stats_alert_transferred)
         .def_readonly("interval", &stats_alert::interval)
-        ;
+    ;
 
     enum_<stats_alert::stats_channel>("stats_channel")
         .value("upload_payload", stats_alert::upload_payload)
@@ -417,4 +460,23 @@ void bind_alert()
         .value("download_tracker_protocol", stats_alert::download_tracker_protocol)
     ;
 
+    class_<anonymous_mode_alert, bases<torrent_alert>, noncopyable>(
+        "anonymous_mode_alert", no_init)
+        .def_readonly("kind", &anonymous_mode_alert::kind)
+        .def_readonly("str", &anonymous_mode_alert::str)
+    ;
+
+    enum_<anonymous_mode_alert::kind_t>("kind")
+        .value("tracker_no_anonymous", anonymous_mode_alert::tracker_not_anonymous)
+    ;
+
+    class_<incoming_connection_alert, bases<alert>, noncopyable>(
+        "incoming_connection_alert", no_init)
+        .def_readonly("socket_type", &incoming_connection_alert::socket_type)
+        .add_property("ip", &incoming_connection_alert_ip)
+        ;
+    class_<torrent_need_cert_alert, bases<torrent_alert>, noncopyable>(
+        "torrent_need_cert_alert", no_init)
+        .def_readonly("error", &torrent_need_cert_alert::error) 
+        ;
 }
