@@ -22,41 +22,77 @@ namespace
        obj(i);
     }
 
-    void set_piece_hashes_callback(create_torrent& c, boost::filesystem::path const& p
+#ifndef BOOST_NO_EXCEPTIONS
+    void set_piece_hashes_callback(create_torrent& c, std::string const& p
         , boost::python::object cb)
     {
         set_piece_hashes(c, p, boost::bind(call_python_object, cb, _1));
     }
+#else
+    void set_piece_hashes_callback(create_torrent& c, std::string const& p
+        , boost::python::object cb)
+    {
+        error_code ec;
+        set_piece_hashes(c, p, boost::bind(call_python_object, cb, _1), ec);
+    }
+
+    void set_piece_hashes0(create_torrent& c, std::string const & s)
+    {
+        error_code ec;
+        set_piece_hashes(c, s, ec);
+    }
+#endif
 
     void add_node(create_torrent& ct, std::string const& addr, int port)
     {
         ct.add_node(std::make_pair(addr, port));
     }
+
+    void add_file(file_storage& ct, file_entry const& fe)
+    {
+       ct.add_file(fe);
+    }
+
+    char const* filestorage_name(file_storage const& fs)
+    { return fs.name().c_str(); }
 }
 
 void bind_create_torrent()
 {
-    void (file_storage::*add_file0)(file_entry const&) = &file_storage::add_file;
-    void (file_storage::*add_file1)(fs::path const&, size_type, int, std::time_t, fs::path const&) = &file_storage::add_file;
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
-    void (file_storage::*add_file2)(fs::wpath const&, size_type, int, std::time_t, fs::path const&) = &file_storage::add_file;
+    void (file_storage::*add_file0)(std::string const&, size_type, int, std::time_t, std::string const&) = &file_storage::add_file;
+#if TORRENT_USE_WSTRING
+    void (file_storage::*add_file1)(std::wstring const&, size_type, int, std::time_t, std::string const&) = &file_storage::add_file;
 #endif
 
     void (file_storage::*set_name0)(std::string const&) = &file_storage::set_name;
+    void (file_storage::*rename_file0)(int, std::string const&) = &file_storage::rename_file;
+#if TORRENT_USE_WSTRING
     void (file_storage::*set_name1)(std::wstring const&) = &file_storage::set_name;
+    void (file_storage::*rename_file1)(int, std::wstring const&) = &file_storage::rename_file;
+#endif
 
-    void (*set_piece_hashes0)(create_torrent&, boost::filesystem::path const&) = &set_piece_hashes;
-    void (*add_files0)(file_storage&, boost::filesystem::path const&, boost::uint32_t) = add_files;
+#ifndef BOOST_NO_EXCEPTIONS
+    void (*set_piece_hashes0)(create_torrent&, std::string const&) = &set_piece_hashes;
+#endif
+    void (*add_files0)(file_storage&, std::string const&, boost::uint32_t) = add_files;
+
+    file_entry (file_storage::*at)(int) const = &file_storage::at;
 
     class_<file_storage>("file_storage")
         .def("is_valid", &file_storage::is_valid)
-        .def("add_file", add_file0)
+        .def("add_file", add_file, arg("entry"))
+        .def("add_file", add_file0, (arg("path"), arg("size"), arg("flags") = 0, arg("mtime") = 0, arg("linkpath") = ""))
+#if TORRENT_USE_WSTRING
         .def("add_file", add_file1, (arg("path"), arg("size"), arg("flags") = 0, arg("mtime") = 0, arg("linkpath") = ""))
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
-        .def("add_file", add_file2, (arg("path"), arg("size"), arg("flags") = 0, arg("mtime") = 0, arg("linkpath") = ""))
 #endif
         .def("num_files", &file_storage::num_files)
-        .def("at", &file_storage::at, return_internal_reference<>())
+        .def("at", at)
+//        .def("hash", &file_storage::hash)
+//        .def("symlink", &file_storage::symlink, return_internal_reference<>())
+//        .def("file_index", &file_storage::file_index)
+//        .def("file_base", &file_storage::file_base)
+//        .def("set_file_base", &file_storage::set_file_base)
+//        .def("file_path", &file_storage::file_path)
         .def("total_size", &file_storage::total_size)
         .def("set_num_pieces", &file_storage::set_num_pieces)
         .def("num_pieces", &file_storage::num_pieces)
@@ -64,12 +100,17 @@ void bind_create_torrent()
         .def("piece_length", &file_storage::piece_length)
         .def("piece_size", &file_storage::piece_size)
         .def("set_name", set_name0)
+        .def("rename_file", rename_file0)
+#if TORRENT_USE_WSTRING
         .def("set_name", set_name1)
-        .def("name", &file_storage::name, return_internal_reference<>())
+        .def("rename_file", rename_file1)
+#endif
+        .def("name", &filestorage_name)
         ;
 
     class_<create_torrent>("create_torrent", no_init)
         .def(init<file_storage&>())
+        .def(init<torrent_info const&>(arg("ti")))
         .def(init<file_storage&, int, int, int>((arg("storage"), arg("piece_size") = 0
             , arg("pad_file_limit") = -1, arg("flags") = int(libtorrent::create_torrent::optimize))))
 
@@ -87,6 +128,7 @@ void bind_create_torrent()
         .def("piece_length", &create_torrent::piece_length)
         .def("piece_size", &create_torrent::piece_size)
         .def("priv", &create_torrent::priv)
+        .def("set_root_cert", &create_torrent::set_root_cert, (arg("pem")))
         ;
 
     enum_<create_torrent::flags_t>("create_torrent_flags_t")
