@@ -2,6 +2,10 @@
 // subject to the Boost Software License, Version 1.0. (See accompanying
 // file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 
+#ifdef __GNUC__
+#define BOOST_PYTHON_USE_GCC_SYMBOL_VISIBILITY 1
+#endif
+
 #include <boost/python/module.hpp>
 
 void bind_utility();
@@ -10,7 +14,6 @@ void bind_big_number();
 void bind_session();
 void bind_entry();
 void bind_torrent_info();
-void bind_filesystem();
 void bind_torrent_handle();
 void bind_torrent_status();
 void bind_session_settings();
@@ -39,7 +42,6 @@ BOOST_PYTHON_MODULE(libtorrent)
     bind_entry();
     bind_session();
     bind_torrent_info();
-    bind_filesystem();
     bind_torrent_handle();
     bind_torrent_status();
     bind_session_settings();
