@@ -9,21 +9,29 @@
 using namespace boost::python;
 using namespace libtorrent;
 
-object pieces(torrent_status const& s)
+object bitfield_to_list(bitfield const& bf)
 {
-    list result;
+	list ret;
 
-    for (bitfield::const_iterator i(s.pieces.begin()), e(s.pieces.end()); i != e; ++i)
-        result.append(*i);
-
-    return result;
+	for (bitfield::const_iterator i(bf.begin()), e(bf.end()); i != e; ++i)
+		ret.append(*i);
+	return ret;
 }
 
+object pieces(torrent_status const& s) { return bitfield_to_list(s.pieces); }
+object verified_pieces(torrent_status const& s) { return bitfield_to_list(s.verified_pieces); }
+
 void bind_torrent_status()
 {
     scope status = class_<torrent_status>("torrent_status")
+        .def_readonly("info_hash", &torrent_status::info_hash)
         .def_readonly("state", &torrent_status::state)
         .def_readonly("paused", &torrent_status::paused)
+        .def_readonly("auto_managed", &torrent_status::auto_managed)
+        .def_readonly("sequential_download", &torrent_status::sequential_download)
+        .def_readonly("is_seeding", &torrent_status::is_seeding)
+        .def_readonly("is_finished", &torrent_status::is_finished)
+        .def_readonly("has_metadata", &torrent_status::has_metadata)
         .def_readonly("progress", &torrent_status::progress)
         .def_readonly("progress_ppm", &torrent_status::progress_ppm)
         .add_property(
@@ -55,7 +63,8 @@ void bind_torrent_status()
         .def_readonly("num_incomplete", &torrent_status::num_incomplete)
         .def_readonly("list_seeds", &torrent_status::list_seeds)
         .def_readonly("list_peers", &torrent_status::list_peers)
-        .add_property("pieces", pieces)
+        .add_property("pieces", &pieces)
+        .add_property("verified_pieces", &verified_pieces)
         .def_readonly("num_pieces", &torrent_status::num_pieces)
         .def_readonly("total_done", &torrent_status::total_done)
         .def_readonly("total_wanted_done", &torrent_status::total_wanted_done)
@@ -78,8 +87,20 @@ void bind_torrent_status()
         .def_readonly("seeding_time", &torrent_status::seeding_time)
         .def_readonly("seed_rank", &torrent_status::seed_rank)
         .def_readonly("last_scrape", &torrent_status::last_scrape)
+        .def_readonly("has_incoming", &torrent_status::has_incoming)
+        .def_readonly("sparse_regions", &torrent_status::sparse_regions)
+        .def_readonly("seed_mode", &torrent_status::seed_mode)
+        .def_readonly("upload_mode", &torrent_status::upload_mode)
+        .def_readonly("share_mode", &torrent_status::share_mode)
         .def_readonly("error", &torrent_status::error)
         .def_readonly("priority", &torrent_status::priority)
+        .def_readonly("added_time", &torrent_status::added_time)
+        .def_readonly("completed_time", &torrent_status::completed_time)
+        .def_readonly("last_seen_complete", &torrent_status::last_seen_complete)
+        .def_readonly("time_since_upload", &torrent_status::time_since_upload)
+        .def_readonly("time_since_download", &torrent_status::time_since_download)
+        .def_readonly("queue_position", &torrent_status::queue_position)
+        .def_readonly("need_save_resume", &torrent_status::need_save_resume)
         ;
 
     enum_<torrent_status::state_t>("states")
