@@ -37,20 +37,19 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/storage.hpp"
 #include "libtorrent/hasher.hpp"
 #include "libtorrent/create_torrent.hpp"
+#include "libtorrent/file.hpp"
+#include "libtorrent/file_pool.hpp"
 
-#include <boost/filesystem/operations.hpp>
-#include <boost/filesystem/path.hpp>
 #include <boost/bind.hpp>
 
-using namespace boost::filesystem;
 using namespace libtorrent;
 
 // do not include files and folders whose
 // name starts with a .
-bool file_filter(boost::filesystem::path const& filename)
+bool file_filter(std::string const& f)
 {
-	if (filename.leaf()[0] == '.') return false;
-	fprintf(stderr, "%s\n", filename.string().c_str());
+	if (filename(f)[0] == '.') return false;
+	fprintf(stderr, "%s\n", f.c_str());
 	return true;
 }
 
@@ -70,6 +69,9 @@ void print_usage()
 		"            merkle torrents require client support\n"
 		"            the resulting full merkle tree is written to\n"
 		"            the specified file\n"
+		"-f          include sha-1 file hashes in the torrent\n"
+		"            this helps supporting mixing sources from\n"
+		"            other networks\n"
 		"-w url      adds a web seed to the torrent with\n"
 		"            the specified url\n"
 		"-t url      adds the specified tracker to the\n"
@@ -84,13 +86,14 @@ void print_usage()
 		"            If this is not specified, the torrent file is\n"
 		"            printed to the standard out, except on windows\n"
 		"            where the filename defaults to a.torrent\n"
+		"-c file     add root certificate to the torrent, to verify\n"
+		"            the HTTPS tracker\n"
 		, stderr);
 }
 
 int main(int argc, char* argv[])
 {
 	using namespace libtorrent;
-	using namespace boost::filesystem;
 
 	char const* creator_str = "libtorrent";
 
@@ -109,6 +112,7 @@ int main(int argc, char* argv[])
 		int pad_file_limit = -1;
 		int piece_size = 0;
 		int flags = 0;
+		std::string root_cert;
 
 		std::string outfile;
 		std::string merklefile;
@@ -154,9 +158,16 @@ int main(int argc, char* argv[])
 					++i;
 					outfile = argv[i];
 					break;
+				case 'f':
+					flags |= create_torrent::calculate_file_hashes;
+					break;
 				case 'l':
 					flags |= create_torrent::symlinks;
 					break;
+				case 'c':
+					++i;
+					root_cert = argv[i];
+					break;
 				default:
 					print_usage();
 					return 1;
@@ -165,7 +176,7 @@ int main(int argc, char* argv[])
 
 		file_storage fs;
 		file_pool fp;
-		path full_path = complete(path(argv[1]));
+		std::string full_path = libtorrent::complete(argv[1]);
 
 		add_files(fs, full_path, file_filter, flags);
 		if (fs.num_files() == 0)
@@ -184,7 +195,7 @@ int main(int argc, char* argv[])
 			t.add_url_seed(*i);
 
 		error_code ec;
-		set_piece_hashes(t, full_path.branch_path()
+		set_piece_hashes(t, parent_path(full_path)
 			, boost::bind(&print_progress, _1, t.num_pieces()), ec);
 		if (ec)
 		{
@@ -195,6 +206,20 @@ int main(int argc, char* argv[])
 		fprintf(stderr, "\n");
 		t.set_creator(creator_str);
 
+		if (!root_cert.empty())
+		{
+			std::vector<char> pem;
+			load_file(root_cert, pem, ec, 10000);
+			if (ec)
+			{
+				fprintf(stderr, "failed to load root certificate for tracker: %s\n", ec.message().c_str());
+			}
+			else
+			{
+				t.set_root_cert(std::string(&pem[0], pem.size()));
+			}
+		}
+
 		// create the torrent and print it to stdout
 		std::vector<char> torrent;
 		bencode(back_inserter(torrent), t.generate());
