@@ -34,15 +34,19 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #include "libtorrent/config.hpp"
 
-#ifdef TORRENT_WINDOWS
+#if defined TORRENT_WINDOWS
 
 #include <malloc.h>
 #define TORRENT_ALLOCA(t, n) static_cast<t*>(_alloca(sizeof(t) * (n)))
 
+#elif defined TORRENT_BSD
+
+#include <stdlib.h>
+#define TORRENT_ALLOCA(t, n) static_cast<t*>(alloca(sizeof(t) * (n)))
+
 #else
 
 #include <alloca.h>
-
 #define TORRENT_ALLOCA(t, n) static_cast<t*>(alloca(sizeof(t) * (n)))
 
 #endif
