@@ -34,8 +34,7 @@ POSSIBILITY OF SUCH DAMAGE.
 #define TORRENT_ESCAPE_STRING_HPP_INCLUDED
 
 #include <string>
-#include <limits>
-#include <boost/optional.hpp>
+#include <boost/limits.hpp>
 #include <boost/array.hpp>
 #include "libtorrent/config.hpp"
 #include "libtorrent/size_type.hpp"
@@ -43,54 +42,61 @@ POSSIBILITY OF SUCH DAMAGE.
 
 namespace libtorrent
 {
-	TORRENT_EXPORT boost::array<char, 3 + std::numeric_limits<size_type>::digits10> to_string(size_type n);
+	TORRENT_EXTRA_EXPORT boost::array<char, 4 + std::numeric_limits<size_type>::digits10> to_string(size_type n);
+	TORRENT_EXTRA_EXPORT bool is_alpha(char c);
 	TORRENT_EXPORT bool is_digit(char c);
-	TORRENT_EXPORT bool is_print(char c);
-	TORRENT_EXPORT bool is_space(char c);
-	TORRENT_EXPORT char to_lower(char c);
+	TORRENT_EXTRA_EXPORT bool is_print(char c);
+	TORRENT_EXTRA_EXPORT bool is_space(char c);
+	TORRENT_EXTRA_EXPORT char to_lower(char c);
 
-	TORRENT_EXPORT bool string_begins_no_case(char const* s1, char const* s2);
-	TORRENT_EXPORT bool string_equal_no_case(char const* s1, char const* s2);
+	TORRENT_EXTRA_EXPORT int split_string(char const** tags, int buf_size, char* in);
+	TORRENT_EXTRA_EXPORT bool string_begins_no_case(char const* s1, char const* s2);
+	TORRENT_EXTRA_EXPORT bool string_equal_no_case(char const* s1, char const* s2);
 
-	TORRENT_EXPORT std::string unescape_string(std::string const& s, error_code& ec);
+	TORRENT_EXTRA_EXPORT void url_random(char* begin, char* end);
+
+	TORRENT_EXTRA_EXPORT std::string unescape_string(std::string const& s, error_code& ec);
 	// replaces all disallowed URL characters by their %-encoding
-	TORRENT_EXPORT std::string escape_string(const char* str, int len);
+	TORRENT_EXTRA_EXPORT std::string escape_string(const char* str, int len);
 	// same as escape_string but does not encode '/'
-	TORRENT_EXPORT std::string escape_path(const char* str, int len);
+	TORRENT_EXTRA_EXPORT std::string escape_path(const char* str, int len);
 	// if the url does not appear to be encoded, and it contains illegal url characters
 	// it will be encoded
-	TORRENT_EXPORT std::string maybe_url_encode(std::string const& url);
+	TORRENT_EXTRA_EXPORT std::string maybe_url_encode(std::string const& url);
 
-	TORRENT_EXPORT bool need_encoding(char const* str, int len);
+	TORRENT_EXTRA_EXPORT bool need_encoding(char const* str, int len);
 
 	// encodes a string using the base64 scheme
-	TORRENT_EXPORT std::string base64encode(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string base64encode(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string base64decode(std::string const& s);
 	// encodes a string using the base32 scheme
-	TORRENT_EXPORT std::string base32encode(std::string const& s);
-	TORRENT_EXPORT std::string base32decode(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string base32encode(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string base32decode(std::string const& s);
 
-	TORRENT_EXPORT boost::optional<std::string> url_has_argument(
-		std::string const& url, std::string argument, size_t* out_pos = 0);
+	TORRENT_EXTRA_EXPORT std::string url_has_argument(
+		std::string const& url, std::string argument, std::string::size_type* out_pos = 0);
 
 	// replaces \ with /
-	TORRENT_EXPORT void convert_path_to_posix(std::string& path);
+	TORRENT_EXTRA_EXPORT void convert_path_to_posix(std::string& path);
 
-	TORRENT_EXPORT std::string read_until(char const*& str, char delim, char const* end);
+	TORRENT_EXTRA_EXPORT std::string read_until(char const*& str, char delim, char const* end);
 	TORRENT_EXPORT std::string to_hex(std::string const& s);
 	TORRENT_EXPORT bool is_hex(char const *in, int len);
 	TORRENT_EXPORT void to_hex(char const *in, int len, char* out);
 	TORRENT_EXPORT bool from_hex(char const *in, int len, char* out);
 
-#if TORRENT_USE_WPATH
-	TORRENT_EXPORT std::wstring convert_to_wstring(std::string const& s);
+#if defined TORRENT_WINDOWS && TORRENT_USE_WSTRING
+	TORRENT_EXTRA_EXPORT std::wstring convert_to_wstring(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string convert_from_wstring(std::wstring const& s);
 #endif
 	
-#if defined TORRENT_WINDOWS || TORRENT_USE_LOCALE_FILENAMES
-	TORRENT_EXPORT std::string convert_to_native(std::string const& s);
+#if TORRENT_USE_ICONV || TORRENT_USE_LOCALE
+	TORRENT_EXTRA_EXPORT std::string convert_to_native(std::string const& s);
+	TORRENT_EXTRA_EXPORT std::string convert_from_native(std::string const& s);
 #else
 	inline std::string const& convert_to_native(std::string const& s) { return s; }
+	inline std::string const& convert_from_native(std::string const& s) { return s; }
 #endif		
-	
 }
 
 #endif // TORRENT_ESCAPE_STRING_HPP_INCLUDED
