@@ -35,14 +35,12 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #include <memory>
 #include <string>
-#include <vector>
 
 #ifdef _MSC_VER
 #pragma warning(push, 1)
 #endif
 
 #include <boost/noncopyable.hpp>
-#include <boost/filesystem/path.hpp>
 
 #ifdef _MSC_VER
 #pragma warning(pop)
@@ -51,6 +49,7 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/error_code.hpp"
 #include "libtorrent/size_type.hpp"
 #include "libtorrent/config.hpp"
+#include "libtorrent/intrusive_ptr_base.hpp"
 
 #ifdef TORRENT_WINDOWS
 // windows part
@@ -59,6 +58,8 @@ POSSIBILITY OF SUCH DAMAGE.
 #endif
 #include <windows.h>
 #include <winioctl.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #else
 // posix part
 #define _FILE_OFFSET_BITS 64
@@ -74,18 +75,105 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <unistd.h>
 #include <sys/uio.h>
 #include <fcntl.h>
-#include <sys/stat.h>
 #include <sys/types.h>
+#include <dirent.h> // for DIR
+
 #undef _FILE_OFFSET_BITS
+
 #endif
+
 namespace libtorrent
 {
-	namespace fs = boost::filesystem;
+	struct file_status
+	{
+		size_type file_size;
+		time_t atime;
+		time_t mtime;
+		time_t ctime;
+		enum {
+#if defined TORRENT_WINDOWS
+			directory = _S_IFDIR,
+			regular_file = _S_IFREG
+#else
+			fifo = S_IFIFO,
+			character_special = S_IFCHR,
+			directory = S_IFDIR,
+			block_special = S_IFBLK,
+			regular_file = S_IFREG,
+			link = S_IFLNK,
+			socket = S_IFSOCK
+#endif
+		} modes_t;
+		int mode;
+	};
 
-	class TORRENT_EXPORT file: public boost::noncopyable
+	enum stat_flags_t { dont_follow_links = 1 };
+	TORRENT_EXPORT void stat_file(std::string f, file_status* s
+		, error_code& ec, int flags = 0);
+	TORRENT_EXPORT void rename(std::string const& f
+		, std::string const& newf, error_code& ec);
+	TORRENT_EXPORT void create_directories(std::string const& f
+		, error_code& ec);
+	TORRENT_EXPORT void create_directory(std::string const& f
+		, error_code& ec);
+	TORRENT_EXPORT void remove_all(std::string const& f
+		, error_code& ec);
+	TORRENT_EXPORT void remove(std::string const& f, error_code& ec);
+	TORRENT_EXPORT bool exists(std::string const& f);
+	TORRENT_EXPORT size_type file_size(std::string const& f);
+	TORRENT_EXPORT bool is_directory(std::string const& f
+		, error_code& ec);
+	TORRENT_EXPORT void recursive_copy(std::string const& old_path
+		, std::string const& new_path, error_code& ec);
+	TORRENT_EXPORT void copy_file(std::string const& f
+		, std::string const& newf, error_code& ec);
+
+	TORRENT_EXPORT std::string split_path(std::string const& f);
+	TORRENT_EXPORT char const* next_path_element(char const* p);
+	TORRENT_EXPORT std::string extension(std::string const& f);
+	TORRENT_EXPORT void replace_extension(std::string& f, std::string const& ext);
+	TORRENT_EXPORT bool is_root_path(std::string const& f);
+	TORRENT_EXPORT std::string parent_path(std::string const& f);
+	TORRENT_EXPORT bool has_parent_path(std::string const& f);
+	TORRENT_EXPORT std::string filename(std::string const& f);
+	TORRENT_EXPORT std::string combine_path(std::string const& lhs
+		, std::string const& rhs);
+	TORRENT_EXPORT std::string complete(std::string const& f);
+	TORRENT_EXPORT bool is_complete(std::string const& f);
+	TORRENT_EXPORT std::string current_working_directory();
+#if TORRENT_USE_UNC_PATHS
+	TORRENT_EXTRA_EXPORT std::string canonicalize_path(std::string const& f);
+#endif
+
+	class TORRENT_EXPORT directory : public boost::noncopyable
 	{
 	public:
+		directory(std::string const& path, error_code& ec);
+		~directory();
+		void next(error_code& ec);
+		std::string file() const;
+		bool done() const { return m_done; }
+	private:
+#ifdef TORRENT_WINDOWS
+		HANDLE m_handle;
+#if TORRENT_USE_WSTRING
+		WIN32_FIND_DATAW m_fd;
+#else
+		WIN32_FIND_DATAA m_fd;
+#endif
+#else
+		DIR* m_handle;
+		// the dirent struct contains a zero-sized
+		// array at the end, it will end up referring
+		// to the m_name field
+		struct dirent m_dirent;
+		char m_name[TORRENT_MAX_PATH + 1]; // +1 to make room for null
+#endif
+		bool m_done;
+	};
 
+	struct TORRENT_EXPORT file: boost::noncopyable, intrusive_ptr_base<file>
+	{
 		enum
 		{
 			// when a file is opened with no_buffer
@@ -98,9 +186,10 @@ namespace libtorrent
 			read_write = 2,
 			rw_mask = read_only | write_only | read_write,
 			no_buffer = 4,
-			mode_mask = rw_mask | no_buffer,
 			sparse = 8,
+			no_atime = 16,
 			random_access = 32,
+			lock_file = 64,
 
 			attribute_hidden = 0x1000,
 			attribute_executable = 0x2000,
@@ -126,14 +215,18 @@ namespace libtorrent
 #endif
 
 		file();
-		file(fs::path const& p, int m, error_code& ec);
+		file(std::string const& p, int m, error_code& ec);
 		~file();
 
-		bool open(fs::path const& p, int m, error_code& ec);
+		bool open(std::string const& p, int m, error_code& ec);
 		bool is_open() const;
 		void close();
 		bool set_size(size_type size, error_code& ec);
 
+		// called when we're done writing to the file.
+		// On windows this will clear the sparse bit
+		void finalize();
+
 		int open_mode() const { return m_open_mode; }
 
 		// when opened in unbuffered mode, this is the
@@ -152,6 +245,7 @@ namespace libtorrent
 
 		size_type writev(size_type file_offset, iovec_t const* bufs, int num_bufs, error_code& ec);
 		size_type readv(size_type file_offset, iovec_t const* bufs, int num_bufs, error_code& ec);
+		void hint_read(size_type file_offset, int len);
 
 		size_type get_size(error_code& ec) const;
 
@@ -171,14 +265,15 @@ namespace libtorrent
 
 #ifdef TORRENT_WINDOWS
 		HANDLE m_file_handle;
-#if TORRENT_USE_WPATH
+#if TORRENT_USE_WSTRING
 		std::wstring m_path;
 #else
 		std::string m_path;
-#endif
-#else
+#endif // TORRENT_USE_WSTRING
+#else // TORRENT_WINDOWS
 		int m_fd;
-#endif
+#endif // TORRENT_WINDOWS
+
 #if defined TORRENT_WINDOWS || defined TORRENT_LINUX || defined TORRENT_DEBUG
 		static void init_file();
 		static int m_page_size;
