@@ -33,7 +33,9 @@ POSSIBILITY OF SUCH DAMAGE.
 #ifndef TORRENT_HTTP_CONNECTION
 #define TORRENT_HTTP_CONNECTION
 
-#include <boost/function.hpp>
+#include <boost/function/function1.hpp>
+#include <boost/function/function2.hpp>
+#include <boost/function/function5.hpp>
 #include <boost/bind.hpp>
 #include <boost/shared_ptr.hpp>
 #include <boost/enable_shared_from_this.hpp>
@@ -43,15 +45,17 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <string>
 
 #include "libtorrent/socket.hpp"
+#include "libtorrent/error_code.hpp"
 #include "libtorrent/http_parser.hpp"
-#include "libtorrent/time.hpp"
+#include "libtorrent/deadline_timer.hpp"
 #include "libtorrent/assert.hpp"
 #include "libtorrent/socket_type.hpp"
 #include "libtorrent/session_settings.hpp"
 
+#include "libtorrent/i2p_stream.hpp"
+
 #ifdef TORRENT_USE_OPENSSL
-#include "libtorrent/ssl_stream.hpp"
-#include "libtorrent/variant_stream.hpp"
+#include <boost/asio/ssl/context.hpp>
 #endif
 
 namespace libtorrent
@@ -69,38 +73,18 @@ typedef boost::function<void(http_connection&, std::list<tcp::endpoint>&)> http_
 
 // when bottled, the last two arguments to the handler
 // will always be 0
-struct TORRENT_EXPORT http_connection : boost::enable_shared_from_this<http_connection>, boost::noncopyable
+struct TORRENT_EXTRA_EXPORT http_connection : boost::enable_shared_from_this<http_connection>, boost::noncopyable
 {
 	http_connection(io_service& ios, connection_queue& cc
 		, http_handler const& handler, bool bottled = true
 		, http_connect_handler const& ch = http_connect_handler()
-		, http_filter_handler const& fh = http_filter_handler())
-		: m_sock(ios)
-#if TORRENT_USE_I2P
-		, m_i2p_conn(0)
+		, http_filter_handler const& fh = http_filter_handler()
+#ifdef TORRENT_USE_OPENSSL
+		, boost::asio::ssl::context* ssl_ctx = 0
 #endif
-		, m_read_pos(0)
-		, m_resolver(ios)
-		, m_handler(handler)
-		, m_connect_handler(ch)
-		, m_filter_handler(fh)
-		, m_timer(ios)
-		, m_last_receive(time_now())
-		, m_bottled(bottled)
-		, m_called(false)
-		, m_rate_limit(0)
-		, m_download_quota(0)
-		, m_limiter_timer_active(false)
-		, m_limiter_timer(ios)
-		, m_redirects(5)
-		, m_connection_ticket(-1)
-		, m_cc(cc)
-		, m_ssl(false)
-		, m_priority(0)
-		, m_abort(false)
-	{
-		TORRENT_ASSERT(!m_handler.empty());
-	}
+		);
+
+	~http_connection();
 
 	void rate_limit(int limit);
 
@@ -111,25 +95,33 @@ struct TORRENT_EXPORT http_connection : boost::enable_shared_from_this<http_conn
 
 	void get(std::string const& url, time_duration timeout = seconds(30)
 		, int prio = 0, proxy_settings const* ps = 0, int handle_redirects = 5
-		, std::string const& user_agent = "", address const& bind_addr = address_v4::any());
+		, std::string const& user_agent = "", address const& bind_addr = address_v4::any()
+#if TORRENT_USE_I2P
+		, i2p_connection* i2p_conn = 0
+#endif
+		);
 
 	void start(std::string const& hostname, std::string const& port
 		, time_duration timeout, int prio = 0, proxy_settings const* ps = 0
 		, bool ssl = false, int handle_redirect = 5
-		, address const& bind_addr = address_v4::any());
+		, address const& bind_addr = address_v4::any()
+#if TORRENT_USE_I2P
+		, i2p_connection* i2p_conn = 0
+#endif
+		);
 
 	void close();
 
-#ifdef TORRENT_USE_OPENSSL
-	variant_stream<socket_type, ssl_stream<socket_type> > const& socket() const { return m_sock; }
-#else
 	socket_type const& socket() const { return m_sock; }
-#endif
 
 	std::list<tcp::endpoint> const& endpoints() const { return m_endpoints; }
 	
 private:
 
+#if TORRENT_USE_I2P
+	void on_i2p_resolve(error_code const& e
+		, char const* destination);
+#endif
 	void on_resolve(error_code const& e
 		, tcp::resolver::iterator i);
 	void queue_connect();
@@ -142,13 +134,12 @@ private:
 		, error_code const& e);
 	void on_assign_bandwidth(error_code const& e);
 
-	void callback(error_code const& e, char const* data = 0, int size = 0);
+	void callback(error_code e, char const* data = 0, int size = 0);
 
 	std::vector<char> m_recvbuffer;
-#ifdef TORRENT_USE_OPENSSL
-	variant_stream<socket_type, ssl_stream<socket_type> > m_sock;
-#else
 	socket_type m_sock;
+#if TORRENT_USE_I2P
+	i2p_connection* m_i2p_conn;
 #endif
 	int m_read_pos;
 	tcp::resolver m_resolver;
@@ -157,8 +148,10 @@ private:
 	http_connect_handler m_connect_handler;
 	http_filter_handler m_filter_handler;
 	deadline_timer m_timer;
-	time_duration m_timeout;
+	time_duration m_read_timeout;
+	time_duration m_completion_timeout;
 	ptime m_last_receive;
+	ptime m_start_time;
 	// bottled means that the handler is called once, when
 	// everything is received (and buffered in memory).
 	// non bottled means that once the headers have been
@@ -169,8 +162,13 @@ private:
 	std::string m_hostname;
 	std::string m_port;
 	std::string m_url;
+	std::string m_user_agent;
 
 	std::list<tcp::endpoint> m_endpoints;
+#ifdef TORRENT_USE_OPENSSL
+	asio::ssl::context* m_ssl_ctx;
+	bool m_own_ssl_context;
+#endif
 
 	// the current download limit, in bytes per second
 	// 0 is unlimited.
