@@ -37,29 +37,26 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <map>
 #include <set>
 
+#include <libtorrent/config.hpp>
 #include <libtorrent/kademlia/routing_table.hpp>
 #include <libtorrent/kademlia/rpc_manager.hpp>
 #include <libtorrent/kademlia/node_id.hpp>
 #include <libtorrent/kademlia/msg.hpp>
+#include <libtorrent/kademlia/find_data.hpp>
 
 #include <libtorrent/io.hpp>
 #include <libtorrent/session_settings.hpp>
 #include <libtorrent/assert.hpp>
+#include <libtorrent/thread.hpp>
+#include <libtorrent/bloom_filter.hpp>
 
 #include <boost/cstdint.hpp>
-#include <boost/optional.hpp>
-#include <boost/iterator/transform_iterator.hpp>
 #include <boost/ref.hpp>
-#include <boost/thread/mutex.hpp>
-#include <boost/optional.hpp>
 
 #include "libtorrent/socket.hpp"
 
 namespace libtorrent {
-	
-	namespace aux { struct session_impl; }
-	struct session_status;
-
+	class alert_manager;
 }
 
 namespace libtorrent { namespace dht
@@ -71,6 +68,34 @@ TORRENT_DECLARE_LOG(node);
 
 struct traversal_algorithm;
 
+struct key_desc_t
+{
+	char const* name;
+	int type;
+	int size;
+	int flags;
+
+	enum {
+		// this argument is optional, parsing will not
+		// fail if it's not present
+		optional = 1,
+		// for dictionaries, the following entries refer
+		// to child nodes to this node, up until and including
+		// the next item that has the last_child flag set.
+		// these flags are nestable
+		parse_children = 2,
+		// this is the last item in a child dictionary
+		last_child = 4,
+		// the size argument refers to that the size
+		// has to be divisible by the number, instead
+		// of having that exact size
+		size_divisible = 8
+	}; 
+};
+
+bool TORRENT_EXTRA_EXPORT verify_message(lazy_entry const* msg, key_desc_t const desc[]
+	, lazy_entry const* ret[], int size , char* error, int error_size);
+
 // this is the entry for every peer
 // the timestamp is there to make it possible
 // to remove stale peers
@@ -78,14 +103,46 @@ struct peer_entry
 {
 	tcp::endpoint addr;
 	ptime added;
+	bool seed;
 };
 
 // this is a group. It contains a set of group members
 struct torrent_entry
 {
+	std::string name;
 	std::set<peer_entry> peers;
 };
 
+struct dht_immutable_item
+{
+	dht_immutable_item() : value(0), num_announcers(0), size(0) {}
+	// malloced space for the actual value
+	char* value;
+	// this counts the number of IPs we have seen
+	// announcing this item, this is used to determine
+	// popularity if we reach the limit of items to store
+	bloom_filter<128> ips;
+	// the last time we heard about this
+	ptime last_seen;
+	// number of IPs in the bloom filter
+	int num_announcers;
+	// size of malloced space pointed to by value
+	int size;
+};
+
+struct dht_mutable_item : dht_immutable_item
+{
+	char sig[256];
+	int seq;
+};
+
+struct rsa_key { char bytes[268]; };
+
+inline bool operator<(rsa_key const& lhs, rsa_key const& rhs)
+{
+	return memcmp(lhs.bytes, rhs.bytes, sizeof(lhs.bytes)) < 0;
+}
+
 inline bool operator<(peer_entry const& lhs, peer_entry const& rhs)
 {
 	return lhs.addr.address() == rhs.addr.address()
@@ -98,71 +155,66 @@ struct null_type {};
 class announce_observer : public observer
 {
 public:
-	announce_observer(boost::pool<>& allocator
-		, sha1_hash const& info_hash
-		, int listen_port
-		, std::string const& write_token)
-		: observer(allocator)
-		, m_info_hash(info_hash)
-		, m_listen_port(listen_port)
-		, m_token(write_token)
+	announce_observer(boost::intrusive_ptr<traversal_algorithm> const& algo
+		, udp::endpoint const& ep, node_id const& id)
+		: observer(algo, ep, id)
 	{}
 
-	void send(msg& m)
-	{
-		m.port = m_listen_port;
-		m.info_hash = m_info_hash;
-		m.write_token = m_token;
-	}
-
-	void timeout() {}
-	void reply(msg const&) {}
-	void abort() {}
-
-private:
-	sha1_hash m_info_hash;
-	int m_listen_port;
-	std::string m_token;
+	void reply(msg const&) { flags |= flag_done; }
 };
 
-class node_impl : boost::noncopyable
+struct count_peers
+{
+	int& count;
+	count_peers(int& c): count(c) {}
+	void operator()(std::pair<libtorrent::dht::node_id
+		, libtorrent::dht::torrent_entry> const& t)
+	{
+		count += t.second.peers.size();
+	}
+};
+	
+class TORRENT_EXTRA_EXPORT node_impl : boost::noncopyable
 {
 typedef std::map<node_id, torrent_entry> table_t;
+typedef std::map<node_id, dht_immutable_item> dht_immutable_table_t;
+typedef std::map<rsa_key, dht_mutable_item> dht_mutable_table_t;
+
 public:
-	node_impl(libtorrent::aux::session_impl& ses, boost::function<void(msg const&)> const& f
-		, dht_settings const& settings, boost::optional<node_id> nid);
+	typedef boost::function3<void, address, int, address> external_ip_fun;
+
+	node_impl(libtorrent::alert_manager& alerts
+		, bool (*f)(void*, entry&, udp::endpoint const&, int)
+		, dht_settings const& settings, node_id nid, address const& external_address
+		, external_ip_fun ext_ip, void* userdata);
 
 	virtual ~node_impl() {}
 
-	void refresh(node_id const& id, boost::function0<void> f);
+	void tick();
+	void refresh(node_id const& id, find_data::nodes_callback const& f);
 	void bootstrap(std::vector<udp::endpoint> const& nodes
-		, boost::function0<void> f);
-	void find_node(node_id const& id, boost::function<
-	void(std::vector<node_entry> const&)> f);
+		, find_data::nodes_callback const& f);
 	void add_router_node(udp::endpoint router);
 		
 	void unreachable(udp::endpoint const& ep);
 	void incoming(msg const& m);
 
-	void refresh();
-	void refresh_bucket(int bucket);
+	int num_torrents() const { return m_map.size(); }
+	int num_peers() const
+	{
+		int ret = 0;
+		std::for_each(m_map.begin(), m_map.end(), count_peers(ret));
+		return ret;
+	}
+
 	int bucket_size(int bucket);
 
-	typedef routing_table::iterator iterator;
-	
-	iterator begin() const { return m_table.begin(); }
-	iterator end() const { return m_table.end(); }
-
-	typedef table_t::iterator data_iterator;
-
 	node_id const& nid() const { return m_id; }
 
 	boost::tuple<int, int> size() const{ return m_table.size(); }
 	size_type num_global_nodes() const
 	{ return m_table.num_global_nodes(); }
 
-	data_iterator begin_data() { return m_map.begin(); }
-	data_iterator end_data() { return m_map.end(); }
 	int data_size() const { return int(m_map.size()); }
 
 #ifdef TORRENT_DHT_VERBOSE_LOGGING
@@ -170,16 +222,17 @@ public:
 	{ m_table.print_state(os); }
 #endif
 
-	void announce(sha1_hash const& info_hash, int listen_port
+	void announce(sha1_hash const& info_hash, int listen_port, bool seed
 		, boost::function<void(std::vector<tcp::endpoint> const&)> f);
 
-	bool verify_token(msg const& m);
-	std::string generate_token(msg const& m);
+	bool verify_token(std::string const& token, char const* info_hash
+		, udp::endpoint const& addr);
+
+	std::string generate_token(udp::endpoint const& addr, char const* info_hash);
 	
 	// the returned time is the delay until connection_timeout()
 	// should be called again the next time
 	time_duration connection_timeout();
-	time_duration refresh_timeout();
 
 	// generates a new secret number used to generate write tokens
 	void new_write_key();
@@ -208,27 +261,26 @@ public:
 
 	void status(libtorrent::session_status& s);
 
-protected:
-	// is called when a find data request is received. Should
-	// return false if the data is not stored on this node. If
-	// the data is stored, it should be serialized into 'data'.
-	bool on_find(msg const& m, std::vector<tcp::endpoint>& peers) const;
+	dht_settings const& settings() const { return m_settings; }
 
-	// this is called when a store request is received. The data
-	// is store-parameters and the data to be stored.
-	void on_announce(msg const& m, msg& reply);
+protected:
+
+	void lookup_peers(sha1_hash const& info_hash, int prefix, entry& reply
+		, bool noseed, bool scrape) const;
+	bool lookup_torrents(sha1_hash const& target, entry& reply
+		, char* tags) const;
 
 	dht_settings const& m_settings;
 	
 private:
-	typedef boost::mutex mutex_t;
+	typedef libtorrent::mutex mutex_t;
 	mutex_t m_mutex;
 
 	// this list must be destructed after the rpc manager
 	// since it might have references to it
 	std::set<traversal_algorithm*> m_running_requests;
 
-	void incoming_request(msg const& h);
+	void incoming_request(msg const& h, entry& e);
 
 	node_id m_id;
 
@@ -238,13 +290,17 @@ public:
 
 private:
 	table_t m_map;
+	dht_immutable_table_t m_immutable_table;
+	dht_mutable_table_t m_mutable_table;
 	
 	ptime m_last_tracker_tick;
 
 	// secret random numbers used to create write tokens
 	int m_secret[2];
 
-	libtorrent::aux::session_impl& m_ses;
+	libtorrent::alert_manager& m_alerts;
+	bool (*m_send)(void*, entry&, udp::endpoint const&, int);
+	void* m_userdata;
 };
 
 
