@@ -41,8 +41,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #endif
 
 #include <boost/limits.hpp>
-#include <boost/filesystem/path.hpp>
-#include <boost/thread.hpp>
 
 #ifdef _MSC_VER
 #pragma warning(pop)
@@ -56,18 +54,27 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/fingerprint.hpp"
 #include "libtorrent/disk_io_thread.hpp"
 #include "libtorrent/peer_id.hpp"
-#include "libtorrent/alert.hpp"
+#include "libtorrent/alert.hpp" // alert::error_notification
+#include "libtorrent/add_torrent_params.hpp"
+#include "libtorrent/rss.hpp"
+#include "libtorrent/build_config.hpp"
 
 #include "libtorrent/storage.hpp"
-#include <boost/preprocessor/cat.hpp>
-
 
 #ifdef _MSC_VER
 #	include <eh.h>
 #endif
 
+#ifdef TORRENT_USE_OPENSSL
+// this is a nasty openssl macro
+#ifdef set_key
+#undef set_key
+#endif
+#endif
+
 namespace libtorrent
 {
+	struct plugin;
 	struct torrent_plugin;
 	class torrent;
 	struct ip_filter;
@@ -77,49 +84,14 @@ namespace libtorrent
 	class upnp;
 	class alert;
 
-	// this is used to create linker errors when trying to link to
-	// a library with a conflicting build configuration than the application
-#ifdef TORRENT_DEBUG
-#define G _release
-#else
-#define G _debug
+	TORRENT_EXPORT session_settings min_memory_usage();
+	TORRENT_EXPORT session_settings high_performance_seed();
+
+#ifndef TORRENT_CFG
+#error TORRENT_CFG is not defined!
 #endif
 
-#ifdef TORRENT_USE_OPENSSL
-#define S _ssl
-#else
-#define S _nossl
-#endif
-
-#ifdef TORRENT_DISABLE_DHT
-#define D _nodht
-#else
-#define D _dht
-#endif
-
-#ifdef TORRENT_DISABLE_POOL_ALLOCATOR
-#define P _nopoolalloc
-#else
-#define P _poolalloc
-#endif
-
-#define TORRENT_LINK_TEST_PREFIX libtorrent_build_config
-#define TORRENT_LINK_TEST_NAME BOOST_PP_CAT(TORRENT_LINK_TEST_PREFIX, BOOST_PP_CAT(P, BOOST_PP_CAT(D, BOOST_PP_CAT(S, G))))
-#undef P
-#undef D
-#undef S
-#undef G
-
-	inline void test_link()
-	{
-		extern void TORRENT_LINK_TEST_NAME();
-		TORRENT_LINK_TEST_NAME();
-	}
-
-	namespace fs = boost::filesystem;
-
-	session_settings min_memory_usage();
-	session_settings high_performance_seed();
+	void TORRENT_EXPORT TORRENT_CFG();
 
 	namespace aux
 	{
@@ -127,13 +99,9 @@ namespace libtorrent
 		// hardware exceptions that makes
 		// it hard to debug stuff
 #ifdef _MSC_VER
-		struct eh_initializer
+		struct TORRENT_EXPORT eh_initializer
 		{
-			eh_initializer()
-			{
-				::_set_se_translator(straight_to_debugger);
-			}
-
+			eh_initializer();
 			static void straight_to_debugger(unsigned int, _EXCEPTION_POINTERS*)
 			{ throw; }
 		};
@@ -141,12 +109,6 @@ namespace libtorrent
 		struct eh_initializer {};
 #endif
 		struct session_impl;
-		
-		struct filesystem_init
-		{
-			filesystem_init();
-		};
-
 	}
 
 	class TORRENT_EXPORT session_proxy
@@ -160,40 +122,16 @@ namespace libtorrent
 		boost::shared_ptr<aux::session_impl> m_impl;
 	};
 
-	struct add_torrent_params
-	{
-		add_torrent_params(storage_constructor_type sc = default_storage_constructor)
-			: tracker_url(0)
-			, name(0)
-			, resume_data(0)
-			, storage_mode(storage_mode_sparse)
-			, paused(true)
-			, auto_managed(true)
-			, duplicate_is_error(false)
-			, storage(sc)
-			, userdata(0)
-			, seed_mode(false)
-			, override_resume_data(false)
-			, upload_mode(false)
-		{}
+#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING || defined TORRENT_ERROR_LOGGING
+#define TORRENT_LOGPATH_ARG_DEFAULT , std::string logpath = "."
+#define TORRENT_LOGPATH_ARG , std::string logpath
+#define TORRENT_LOGPATH , logpath
+#else
+#define TORRENT_LOGPATH_ARG_DEFAULT
+#define TORRENT_LOGPATH_ARG
+#define TORRENT_LOGPATH
+#endif
 
-		boost::intrusive_ptr<torrent_info> ti;
-		char const* tracker_url;
-		sha1_hash info_hash;
-		char const* name;
-		fs::path save_path;
-		std::vector<char>* resume_data;
-		storage_mode_t storage_mode;
-		bool paused;
-		bool auto_managed;
-		bool duplicate_is_error;
-		storage_constructor_type storage;
-		void* userdata;
-		bool seed_mode;
-		bool override_resume_data;
-		bool upload_mode;
-	};
-	
 	class TORRENT_EXPORT session: public boost::noncopyable, aux::eh_initializer
 	{
 	public:
@@ -201,45 +139,58 @@ namespace libtorrent
 		session(fingerprint const& print = fingerprint("LT"
 			, LIBTORRENT_VERSION_MAJOR, LIBTORRENT_VERSION_MINOR, 0, 0)
 			, int flags = start_default_features | add_default_plugins
-			, int alert_mask = alert::error_notification
-#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING || defined TORRENT_ERROR_LOGGING
-			, fs::path logpath = "."
-#endif
-				);
+			, boost::uint32_t alert_mask = alert::error_notification
+			TORRENT_LOGPATH_ARG_DEFAULT)
+		{
+			TORRENT_CFG();
+			init(std::make_pair(0, 0), "0.0.0.0", print, flags, alert_mask TORRENT_LOGPATH);
+		}
+
 		session(
 			fingerprint const& print
 			, std::pair<int, int> listen_port_range
 			, char const* listen_interface = "0.0.0.0"
 			, int flags = start_default_features | add_default_plugins
 			, int alert_mask = alert::error_notification
-#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING || defined TORRENT_ERROR_LOGGING
-			, fs::path logpath = "."
-#endif
-			);
+			TORRENT_LOGPATH_ARG_DEFAULT)
+		{
+			TORRENT_CFG();
+			TORRENT_ASSERT(listen_port_range.first > 0);
+			TORRENT_ASSERT(listen_port_range.first < listen_port_range.second);
+			init(listen_port_range, listen_interface, print, flags, alert_mask TORRENT_LOGPATH);
+		}
 			
 		~session();
 
 		enum save_state_flags_t
 		{
-			save_settings = 0x001,
+			save_settings =     0x001,
 			save_dht_settings = 0x002,
-#ifndef TORRENT_NO_DEPRECATE
-			save_dht_proxy = 0x004,
-#endif
-			save_dht_state = 0x008,
-			save_i2p_proxy = 0x010,
+			save_dht_state =    0x004,
+			save_proxy =        0x008,
+			save_i2p_proxy =    0x010,
 			save_encryption_settings = 0x020,
+			save_as_map =       0x040,
+			save_feeds =        0x080
+
 #ifndef TORRENT_NO_DEPRECATE
-			save_peer_proxy = 0x040,
-			save_web_proxy = 0x080,
-			save_tracker_proxy = 0x100,
+			,
+			save_dht_proxy = save_proxy,
+			save_peer_proxy = save_proxy,
+			save_web_proxy = save_proxy,
+			save_tracker_proxy = save_proxy
 #endif
-			save_as_map = 0x200,
-			save_proxy = 0x1c4
 		};
 		void save_state(entry& e, boost::uint32_t flags = 0xffffffff) const;
 		void load_state(lazy_entry const& e);
 
+		void get_torrent_status(std::vector<torrent_status>* ret
+			, boost::function<bool(torrent_status const&)> const& pred
+			, boost::uint32_t flags = 0) const;
+		void refresh_torrent_status(std::vector<torrent_status>* ret
+			, boost::uint32_t flags = 0) const;
+		void post_torrent_updates();
+
 		// returns a list of all torrents in this session
 		std::vector<torrent_handle> get_torrents() const;
 		
@@ -249,8 +200,11 @@ namespace libtorrent
 		torrent_handle find_torrent(sha1_hash const& info_hash) const;
 
 		// all torrent_handles must be destructed before the session is destructed!
+#ifndef BOOST_NO_EXCEPTIONS
 		torrent_handle add_torrent(add_torrent_params const& params);
+#endif
 		torrent_handle add_torrent(add_torrent_params const& params, error_code& ec);
+		void async_add_torrent(add_torrent_params const& params);
 		
 #ifndef BOOST_NO_EXCEPTIONS
 #ifndef TORRENT_NO_DEPRECATE
@@ -258,7 +212,7 @@ namespace libtorrent
 		TORRENT_DEPRECATED_PREFIX
 		torrent_handle add_torrent(
 			torrent_info const& ti
-			, fs::path const& save_path
+			, std::string const& save_path
 			, entry const& resume_data = entry()
 			, storage_mode_t storage_mode = storage_mode_sparse
 			, bool paused = false
@@ -268,7 +222,7 @@ namespace libtorrent
 		TORRENT_DEPRECATED_PREFIX
 		torrent_handle add_torrent(
 			boost::intrusive_ptr<torrent_info> ti
-			, fs::path const& save_path
+			, std::string const& save_path
 			, entry const& resume_data = entry()
 			, storage_mode_t storage_mode = storage_mode_sparse
 			, bool paused = false
@@ -281,7 +235,7 @@ namespace libtorrent
 			char const* tracker_url
 			, sha1_hash const& info_hash
 			, char const* name
-			, fs::path const& save_path
+			, std::string const& save_path
 			, entry const& resume_data = entry()
 			, storage_mode_t storage_mode = storage_mode_sparse
 			, bool paused = false
@@ -302,6 +256,10 @@ namespace libtorrent
 		void get_cache_info(sha1_hash const& ih
 			, std::vector<cached_piece_info>& ret) const;
 
+		feed_handle add_feed(feed_settings const& feed);
+		void remove_feed(feed_handle h);
+		void get_feeds(std::vector<feed_handle>& f) const;
+
 #ifndef TORRENT_DISABLE_DHT
 		void start_dht();
 		void stop_dht();
@@ -321,20 +279,21 @@ namespace libtorrent
 
 #ifndef TORRENT_DISABLE_ENCRYPTION
 		void set_pe_settings(pe_settings const& settings);
-		pe_settings const& get_pe_settings() const;
+		pe_settings get_pe_settings() const;
 #endif
 
 #ifndef TORRENT_DISABLE_EXTENSIONS
 		void add_extension(boost::function<boost::shared_ptr<torrent_plugin>(torrent*, void*)> ext);
+		void add_extension(boost::shared_ptr<plugin> ext);
 #endif
 
 #ifndef TORRENT_DISABLE_GEO_IP
 		int as_for_ip(address const& addr);
-		bool load_asnum_db(char const* file);
-		bool load_country_db(char const* file);
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
-		bool load_country_db(wchar_t const* file);
-		bool load_asnum_db(wchar_t const* file);
+		void load_asnum_db(char const* file);
+		void load_country_db(char const* file);
+#if TORRENT_USE_WSTRING
+		void load_country_db(wchar_t const* file);
+		void load_asnum_db(wchar_t const* file);
 #endif
 #endif
 
@@ -348,7 +307,7 @@ namespace libtorrent
 #endif
 
 		void set_ip_filter(ip_filter const& f);
-		ip_filter const& get_ip_filter() const;
+		ip_filter get_ip_filter() const;
 		
 		void set_port_filter(port_filter const& f);
 		void set_peer_id(peer_id const& pid);
@@ -368,20 +327,33 @@ namespace libtorrent
 		// this function will return false on failure.
 		// If it fails, it will also generate alerts describing
 		// the error. It will return true on success.
+		enum listen_on_flags_t
+		{
+#ifndef TORRENT_NO_DEPRECATE
+			// this is always on starting with 0.16.2
+			listen_reuse_address = 0x01,
+#endif
+			listen_no_system_port = 0x02
+		};
+
+#ifndef TORRENT_NO_DEPRECATE
+		// deprecated in 0.16
+		TORRENT_DEPRECATED_PREFIX
 		bool listen_on(
 			std::pair<int, int> const& port_range
-			, const char* net_interface = 0);
+			, const char* net_interface = 0
+			, int flags = 0) TORRENT_DEPRECATED;
+#endif
+
+		void listen_on(
+			std::pair<int, int> const& port_range
+			, error_code& ec
+			, const char* net_interface = 0
+			, int flags = 0);
 
 		// returns the port we ended up listening on
 		unsigned short listen_port() const;
 
-		// Get the number of uploads.
-		int num_uploads() const;
-
-		// Get the number of connections. This number also contains the
-		// number of half open connections.
-		int num_connections() const;
-
 		enum options_t
 		{
 			none = 0,
@@ -397,12 +369,26 @@ namespace libtorrent
 		void remove_torrent(const torrent_handle& h, int options = none);
 
 		void set_settings(session_settings const& s);
-		session_settings const& settings();
+		session_settings settings() const;
 
 		void set_proxy(proxy_settings const& s);
-		proxy_settings const& proxy() const;
+		proxy_settings proxy() const;
+
+#ifdef TORRENT_STATS
+		void enable_stats_logging(bool s);
+#endif
 
 #ifndef TORRENT_NO_DEPRECATE
+		// deprecated in 0.16
+		// Get the number of uploads.
+		TORRENT_DEPRECATED_PREFIX
+		int num_uploads() const TORRENT_DEPRECATED;
+
+		// Get the number of connections. This number also contains the
+		// number of half open connections.
+		TORRENT_DEPRECATED_PREFIX
+		int num_connections() const TORRENT_DEPRECATED;
+
 		// deprecated in 0.15.
 		TORRENT_DEPRECATED_PREFIX
 		void set_peer_proxy(proxy_settings const& s) TORRENT_DEPRECATED;
@@ -412,49 +398,84 @@ namespace libtorrent
 		void set_tracker_proxy(proxy_settings const& s) TORRENT_DEPRECATED;
 
 		TORRENT_DEPRECATED_PREFIX
-		proxy_settings const& peer_proxy() const TORRENT_DEPRECATED;
+		proxy_settings peer_proxy() const TORRENT_DEPRECATED;
 		TORRENT_DEPRECATED_PREFIX
-		proxy_settings const& web_seed_proxy() const TORRENT_DEPRECATED;
+		proxy_settings web_seed_proxy() const TORRENT_DEPRECATED;
 		TORRENT_DEPRECATED_PREFIX
-		proxy_settings const& tracker_proxy() const TORRENT_DEPRECATED;
+		proxy_settings tracker_proxy() const TORRENT_DEPRECATED;
 
 #ifndef TORRENT_DISABLE_DHT
 		TORRENT_DEPRECATED_PREFIX
 		void set_dht_proxy(proxy_settings const& s) TORRENT_DEPRECATED;
 		TORRENT_DEPRECATED_PREFIX
-		proxy_settings const& dht_proxy() const TORRENT_DEPRECATED;
+		proxy_settings dht_proxy() const TORRENT_DEPRECATED;
 #endif
 #endif // TORRENT_NO_DEPRECATE
 
-		int upload_rate_limit() const;
-		int download_rate_limit() const;
-		int local_upload_rate_limit() const;
-		int local_download_rate_limit() const;
-		int max_half_open_connections() const;
+#if TORRENT_USE_I2P
+		void set_i2p_proxy(proxy_settings const& s);
+		proxy_settings i2p_proxy() const;
+#endif
 
-		void set_local_upload_rate_limit(int bytes_per_second);
-		void set_local_download_rate_limit(int bytes_per_second);
-		void set_upload_rate_limit(int bytes_per_second);
-		void set_download_rate_limit(int bytes_per_second);
-		void set_max_uploads(int limit);
-		void set_max_connections(int limit);
-		void set_max_half_open_connections(int limit);
+#ifndef TORRENT_NO_DEPRECATE
+		// deprecated in 0.16
+		TORRENT_DEPRECATED_PREFIX
+		int upload_rate_limit() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		int download_rate_limit() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		int local_upload_rate_limit() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		int local_download_rate_limit() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		int max_half_open_connections() const TORRENT_DEPRECATED;
 
-		int max_connections() const;
-		int max_uploads() const;
+		TORRENT_DEPRECATED_PREFIX
+		void set_local_upload_rate_limit(int bytes_per_second) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_local_download_rate_limit(int bytes_per_second) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_upload_rate_limit(int bytes_per_second) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_download_rate_limit(int bytes_per_second) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_max_uploads(int limit) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_max_connections(int limit) TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		void set_max_half_open_connections(int limit) TORRENT_DEPRECATED;
 
+		TORRENT_DEPRECATED_PREFIX
+		int max_connections() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		int max_uploads() const TORRENT_DEPRECATED;
+#endif
+
+		// pop one alert from the alert queue, or do nothing
+		// and return a NULL pointer if there are no alerts
+		// in the queue
 		std::auto_ptr<alert> pop_alert();
+
+		// pop all alerts in the alert queue and returns them
+		// in the supplied dequeue 'alerts'. The passed in
+		// queue must be empty when passed in.
+		// the responsibility of individual alerts returned
+		// in the dequeue is passed on to the caller of this function.
+		// when you're done with reacting to the alerts, you need to
+		// delete them all.
+		void pop_alerts(std::deque<alert*>* alerts);
+
 #ifndef TORRENT_NO_DEPRECATE
 		TORRENT_DEPRECATED_PREFIX
 		void set_severity_level(alert::severity_t s) TORRENT_DEPRECATED;
+
 		TORRENT_DEPRECATED_PREFIX
-		void set_alert_mask(int m) TORRENT_DEPRECATED;
+		size_t set_alert_queue_size_limit(size_t queue_size_limit_) TORRENT_DEPRECATED;
 #endif
 		void set_alert_mask(boost::uint32_t m);
-		size_t set_alert_queue_size_limit(size_t queue_size_limit_);
 
 		alert const* wait_for_alert(time_duration max_wait);
-		void set_alert_dispatch(boost::function<void(alert const&)> const& fun);
+		void set_alert_dispatch(boost::function<void(std::auto_ptr<alert>)> const& fun);
 
 		connection_queue& get_connection_queue();
 
@@ -470,9 +491,8 @@ namespace libtorrent
 		
 	private:
 
-		// just a way to initialize boost.filesystem
-		// before the session_impl is created
-		aux::filesystem_init m_dummy;
+		void init(std::pair<int, int> listen_range, char const* listen_interface
+			, fingerprint const& id, int flags, boost::uint32_t alert_mask TORRENT_LOGPATH_ARG);
 
 		// data shared between the main thread
 		// and the working thread
