@@ -35,13 +35,19 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #include "libtorrent/version.hpp"
 #include "libtorrent/config.hpp"
+#include "libtorrent/version.hpp"
+
+#include <string>
 
 namespace libtorrent
 {
 
 	struct TORRENT_EXPORT proxy_settings
 	{
-		proxy_settings() : port(0), type(none) {}
+		proxy_settings() : port(0), type(none)
+			, proxy_hostnames(true)
+			, proxy_peer_connections(true)
+		{}
 
 		std::string hostname;
 		int port;
@@ -71,121 +77,29 @@ namespace libtorrent
 			http,
 			// http proxy with basic authentication
 			// uses username and password
-			http_pw
+			http_pw,
+			// route through a i2p SAM proxy
+			i2p_proxy
 		};
 		
 		proxy_type type;
-	
+
+		// when set to true, hostname are resolved
+		// through the proxy (if supported)
+		bool proxy_hostnames;
+
+		// if true, use this proxy for peers too
+		bool proxy_peer_connections;
 	};
 
 	struct TORRENT_EXPORT session_settings
 	{
 		session_settings(std::string const& user_agent_ = "libtorrent/"
-			LIBTORRENT_VERSION)
-			: user_agent(user_agent_)
-			, tracker_completion_timeout(60)
-			, tracker_receive_timeout(40)
-			, stop_tracker_timeout(5)
-			, tracker_maximum_response_length(1024*1024)
-			, piece_timeout(20)
-			, request_timeout(50)
-			, request_queue_time(3)
-			, max_allowed_in_request_queue(250)
-			, max_out_request_queue(200)
-			, whole_pieces_threshold(20)
-			, peer_timeout(120)
-			, urlseed_timeout(20)
-			, urlseed_pipeline_size(5)
-			, urlseed_wait_retry(30)
-			, file_pool_size(40)
-			, allow_multiple_connections_per_ip(false)
-			, max_failcount(3)
-			, min_reconnect_time(60)
-			, peer_connect_timeout(7)
-			, ignore_limits_on_local_network(true)
-			, connection_speed(10)
-			, send_redundant_have(false)
-			, lazy_bitfields(true)
-			, inactivity_timeout(600)
-			, unchoke_interval(15)
-			, optimistic_unchoke_interval(30)
-			, num_want(200)
-			, initial_picker_threshold(4)
-			, allowed_fast_set_size(10)
-			, max_queued_disk_bytes(256 * 1024)
-			, handshake_timeout(10)
-#ifndef TORRENT_DISABLE_DHT
-			, use_dht_as_fallback(false)
-#endif
-			, free_torrent_hashes(true)
-			, upnp_ignore_nonrouters(false)
- 			, send_buffer_watermark(100 * 1024)
-			, auto_upload_slots(true)
-			, auto_upload_slots_rate_based(false)
-			, use_parole_mode(true)
-			, cache_size(1024)
-			, cache_buffer_chunk_size(16)
-			, cache_expiry(60)
-			, use_read_cache(true)
-			, disk_io_write_mode(0)
-			, disk_io_read_mode(0)
-			, coalesce_reads(false)
-			, coalesce_writes(false)
-			, outgoing_ports(0,0)
-			, peer_tos(0)
-			, active_downloads(8)
-			, active_seeds(5)
-			, active_limit(15)
-			, auto_manage_prefer_seeds(false)
-			, dont_count_slow_torrents(true)
-			, auto_manage_interval(30)
-			, share_ratio_limit(2.f)
-			, seed_time_ratio_limit(7.f)
-			, seed_time_limit(24 * 60 * 60) // 24 hours
-			, peer_turnover(2 / 50.f)
-			, peer_turnover_cutoff(.9f)
-			, close_redundant_connections(true)
-			, auto_scrape_interval(1800)
-			, auto_scrape_min_interval(300)
-			, max_peerlist_size(4000)
-			, max_paused_peerlist_size(4000)
-			, min_announce_interval(5 * 60)
-			, prioritize_partial_pieces(false)
-			, auto_manage_startup(120)
-			, rate_limit_ip_overhead(true)
-			, announce_to_all_trackers(false)
-			, announce_to_all_tiers(false)
-			, prefer_udp_trackers(true)
-			, strict_super_seeding(false)
-			, seeding_piece_quota(20)
-#ifdef TORRENT_WINDOWS
-			, max_sparse_regions(30000)
-#else
-			, max_sparse_regions(0)
-#endif
-#ifndef TORRENT_DISABLE_MLOCK
-			, lock_disk_cache(false)
-#endif
-			, max_rejects(50)
-			, recv_socket_buffer_size(0)
-			, send_socket_buffer_size(0)
-			, optimize_hashing_for_speed(true)
-			, file_checks_delay_per_block(0)
-			, disk_cache_algorithm(largest_contiguous)
-			, read_cache_line_size(16)
-			, write_cache_line_size(32)
-			, optimistic_disk_retry(10 * 60)
-			, disable_hash_checks(false)
-			, allow_reordered_disk_operations(true)
-			, max_suggest_pieces(10)
-			, drop_skipped_requests(false)
-			, low_prio_disk(true)
-			, local_service_announce_interval(5 * 60)
-			, udp_tracker_token_expiry(60)
-			, report_true_downloaded(false)
-			, strict_end_game_mode(true)
-			, broadcast_lsd(false)
-		{}
+			LIBTORRENT_VERSION);
+		~session_settings();
+
+		// libtorrent version. Used for forward binary compatibility
+		int version;
 
 		// this is the user agent that will be sent to the tracker
 		// when doing requests. It is used to identify the client.
@@ -328,7 +242,7 @@ namespace libtorrent
 
 		// if this is set, this IP will be reported do the
 		// tracker in the ip= parameter.
-		address announce_ip;
+		std::string announce_ip;
 
 		// the num want sent to trackers
 		int num_want;
@@ -341,6 +255,13 @@ namespace libtorrent
 		// that supports the fast extensions
 		int allowed_fast_set_size;
 
+		// this determines which pieces will be suggested to peers
+		// suggest read cache will make libtorrent suggest pieces
+		// that are fresh in the disk read cache, to potentially
+		// lower disk access and increase the cache hit ratio
+		enum { no_piece_suggestions = 0, suggest_read_cache = 1 };
+		int suggest_mode;
+
 		// the maximum number of bytes a connection may have
 		// pending in the disk write queue before its download
 		// rate is being throttled. This prevents fast downloads
@@ -352,6 +273,14 @@ namespace libtorrent
 		// this is a per session setting.
 		int max_queued_disk_bytes;
 
+		// this is the low watermark for the disk buffer queue.
+		// whenever the number of queued bytes exceed the
+		// max_queued_disk_bytes, libtorrent will wait for
+		// it to drop below this value before issuing more
+		// reads from the sockets. If set to 0, the
+		// low watermark will be half of the max queued disk bytes
+		int max_queued_disk_bytes_low_watermark;
+
 		// the number of seconds to wait for a handshake
 		// response from a peer. If no response is received
 		// within this time, the peer is disconnected.
@@ -374,6 +303,14 @@ namespace libtorrent
 		// our currently configured router.
 		bool upnp_ignore_nonrouters;
 
+		// This is the minimum send buffer target size (send buffer
+		// includes bytes pending being read from disk). For good
+		// and snappy seeding performance, set this fairly high, to
+		// at least fit a few blocks. This is essentially the initial
+		// window size which will determine how fast we can ramp up
+		// the send rate
+ 		int send_buffer_low_watermark;
+
  		// if the send buffer has fewer bytes than this, we'll
  		// read another 16kB block onto it. If set too small,
  		// upload rate capacity will suffer. If set too high,
@@ -382,20 +319,41 @@ namespace libtorrent
  		// the upload rate is low, this is the upper limit.
  		int send_buffer_watermark;
 
-		// if auto_upload_slots is true, and a global upload
-		// limit is set and the upload rate is less than 90%
-		// of the upload limit, on new slot is opened up. If
-		// the upload rate is >= upload limit for an extended
-		// period of time, one upload slot is closed. The
-		// upload slots are never automatically decreased below
-		// the manual settings, through max_uploads.
+		// the current upload rate to a peer is multiplied by
+		// this factor to get the send buffer watermark. The
+		// factor is specified as a percentage. i.e. 50 -> 0.5
+		// This product is clamped to the send_buffer_watermark
+		// setting to not exceed the max. For high speed
+		// upload, this should be set to a greater value than
+		// 100. The default is 50.
+		int send_buffer_watermark_factor;
+
+#ifndef TORRENT_NO_DEPRECATE
+		// deprecated in 0.16
 		bool auto_upload_slots;
-
-		// this only affects the auto upload slots mechanism.
-		// if auto_upload_slots is false, this field is not
-		// considered.
 		bool auto_upload_slots_rate_based;
+#endif
 
+		enum choking_algorithm_t
+		{
+			fixed_slots_choker,
+			auto_expand_choker,
+			rate_based_choker,
+			bittyrant_choker
+		};
+
+		int choking_algorithm;
+
+		enum seed_choking_algorithm_t
+		{
+			round_robin,
+			fastest_upload,
+			anti_leech
+		};
+ 
+		// the choking algorithm to use for seeding torrents
+		int seed_choking_algorithm;
+		
 		// if set to true, peers that participate in a failing
 		// piece is put in parole mode. i.e. They will only
 		// download whole pieces until they either fail or pass.
@@ -417,13 +375,22 @@ namespace libtorrent
 
 		// the number of seconds a write cache entry sits
 		// idle in the cache before it's forcefully flushed
-		// to disk. Default is 60 seconds.
+		// to disk. Default is 5 minutes.
 		int cache_expiry;
 
 		// when true, the disk I/O thread uses the disk
 		// cache for caching blocks read from disk too
 		bool use_read_cache;
 
+		// don't implicitly cache pieces in the read cache,
+		// only cache pieces that are explicitly asked to be
+		// cached.
+		bool explicit_read_cache;
+
+		// the number of seconds between refreshes of
+		// explicit caches
+		int explicit_cache_interval;
+
 		enum io_buffer_mode_t
 		{
 			enable_os_cache = 0,
@@ -453,8 +420,17 @@ namespace libtorrent
 		// they are subject to. If there are too many torrents
 		// some of the auto managed ones will be paused until
 		// some slots free up.
+		// active_dht_limit and active_tracker_limit limits the
+		// number of torrents that will be active on the DHT
+		// versus the tracker. If the active limit is set higher
+		// than these numbers, some torrents will be "active" in
+		// the sense that they will accept incoming connections,
+		// but not announce on the DHT or the tracker
 		int active_downloads;
 		int active_seeds;
+		int active_dht_limit;
+		int active_tracker_limit;
+		int active_lsd_limit;
 		int active_limit;
 
 		// prefer seeding torrents when determining which torrents to give 
@@ -483,9 +459,14 @@ namespace libtorrent
 		float seed_time_ratio_limit;
 		int seed_time_limit;
 
+		// the interval (in seconds) between optimistic disconnects
+		// if the disconnects happen and how many peers are disconnected
+		// is controlled by peer_turnover and peer_turnover_cutoff
+		int peer_turnover_interval;
+
 		// the percentage of peers to disconnect every
-		// 90 seconds (if we're at the peer limit)
-		// defaults to 1/50:th
+		// turnoever interval (if we're at the peer limit)
+		// defaults to 2/50:th
 		float peer_turnover;
 
 		// when we are connected to more than
@@ -608,7 +589,7 @@ namespace libtorrent
 		int file_checks_delay_per_block;
 
 		enum disk_cache_algo_t
-		{ lru, largest_contiguous };
+		{ lru, largest_contiguous, avoid_readback };
 
 		disk_cache_algo_t disk_cache_algorithm;
 
@@ -643,6 +624,15 @@ namespace libtorrent
 		// doesn't give you an benefits
 		bool allow_reordered_disk_operations;
 
+		// if this is true, i2p torrents are allowed
+		// to also get peers from other sources than
+		// the tracker, and connect to regular IPs,
+		// not providing any anonymization. This may
+		// be useful if the user is not interested in
+		// the anonymization of i2p, but still wants to
+		// be able to connect to i2p peers.
+		bool allow_i2p_mixed;
+
 		// the max number of pieces that a peer can
 		// suggest to use before we start dropping
 		// previous suggested piece
@@ -664,11 +654,63 @@ namespace libtorrent
 		// torrents. Defaults to 5 minutes
 		int local_service_announce_interval;
 
+		// number of seconds between DHT announces for
+		// torrents. Defaults to 15 minutes
+		int dht_announce_interval;
+
 		// the number of seconds a connection ID received
 		// from a UDP tracker is valid for. This is specified
 		// as 60 seconds
 		int udp_tracker_token_expiry;
 
+		// if this is set to true, any block read from the
+		// disk cache will be dropped from the cache immediately
+		// following. This may be useful if the block is not
+		// expected to be hit again. It would save some memory
+		bool volatile_read_cache;
+
+		// if this is set to true, the size of the cache line
+		// generated by a particular read request depends on the
+		// rate you're sending to that peer. This optimizes the
+		// memory usage of the disk read cache by reading
+		// further ahead for peers that you're uploading at high
+		// rates to
+		bool guided_read_cache;
+
+		// this is the default minimum time any read cache line
+		// is kept in the cache.
+		int default_cache_min_age;
+
+		// the global number of optimistic unchokes
+		// 0 means automatic
+		int num_optimistic_unchoke_slots;
+
+		// if set to true, files won't have their atime updated
+		// on disk reads. This works on linux
+		bool no_atime_storage;
+
+		// === BitTyrant unchoker settings ==
+
+		// when using BitTyrant choker, this is the default
+		// assumed reciprocation rate. This is where each peer starts
+		int default_est_reciprocation_rate;
+
+		// this is the increase of the estimated reciprocation rate
+		// in percent. We increase by this amount once every unchoke
+		// interval that we are choked by the other peer and we have
+		// unchoked them
+		int increase_est_reciprocation_rate;
+
+		// each unchoke interval that we stay unchoked by the other
+		// peer, and we have unchoked this peer as well, we decrease
+		// our estimate of the reciprocation rate, since we might have
+		// over-estimated it
+		int decrease_est_reciprocation_rate;
+
+		// if set to true, an incoming connection to a torrent that's
+		// paused and auto-managed will make the torrent start.
+		bool incoming_starts_queued_torrents;
+
 		// when set to true, the downloaded counter sent to trackers
 		// will include the actual number of payload bytes donwnloaded
 		// including redundant bytes. If set to false, it will not include
@@ -684,6 +726,211 @@ namespace libtorrent
 		// default in order to avoid flooding networks for no good reason. If
 		// a network is known not to support multicast, this can be enabled
 		bool broadcast_lsd;
+
+		// when set to true, libtorrent will try to make outgoing utp connections
+		bool enable_outgoing_utp;
+
+		// if set to false, libtorrent will reject incoming utp connections
+		bool enable_incoming_utp;
+
+		// when set to false, no outgoing TCP connections will be made
+		bool enable_outgoing_tcp;
+
+		// if set to false, libtorrent will reject incoming tcp connections
+		bool enable_incoming_tcp;
+
+		// the max number of peers we accept from pex messages from a single peer.
+		// this limits the number of concurrent peers any of our peers claims to
+		// be connected to. If they clain to be connected to more than this, we'll
+		// ignore any peer that exceeds this limit
+		int max_pex_peers;
+
+		// when set to true, the file modification time is ignored when loading
+		// resume data. The resume data includes the expected timestamp of each
+		// file and is typically compared to make sure the files haven't changed
+		// since the last session
+		bool ignore_resume_timestamps;
+
+		// normally, if a resume file is incomplete (typically there's no
+		// "file sizes" field) the torrent is queued for a full check. If
+		// this settings is set to true, instead libtorrent will assume
+		// we have none of the files and go straight to download
+		bool no_recheck_incomplete_resume;
+
+		// when this is true, libtorrent will take actions to make sure no
+		// privacy sensitive information is leaked out from the client. This
+		// mode is assumed to be combined with using a proxy for all your
+		// traffic. With this option, your true IP address will not be exposed
+		// nor anything that can tie your connection to your true IP
+		bool anonymous_mode;
+
+		// the number of milliseconds between internal ticks. Should be no
+		// more than one second (i.e. 1000).
+		int tick_interval;
+
+		// specifies whether downloads from web seeds is reported to the
+		// tracker or not. Defaults to on
+		bool report_web_seed_downloads;
+
+		// this is the target share ratio for share-mode torrents
+		int share_mode_target;
+
+		// max upload rate in bytes per second for the session
+		int upload_rate_limit;
+
+		// max download rate in bytes per second for the session
+		int download_rate_limit;
+
+		// max upload rate in bytes per second for peers on the local
+		// network, in the session
+		int local_upload_rate_limit;
+
+		// max download rate in bytes per second for peers on the local
+		// network, in the session
+		int local_download_rate_limit;
+
+		// max upload rate used by the DHT in bytes per second
+		int dht_upload_rate_limit;
+
+		// the max number of unchoke slots in the session (might be
+		// overridden by unchoke algorithm)
+		int unchoke_slots_limit;
+
+		// the max number of half-open TCP connections
+		int half_open_limit;
+
+		// the max number of connections in the session
+		int connections_limit;
+
+		// target delay, milliseconds
+		int utp_target_delay;
+
+		// max number of bytes to increase cwnd per rtt in uTP
+		// congestion controller
+		int utp_gain_factor;
+
+		// the shortest allowed uTP connection timeout in milliseconds
+		// defaults to 500 milliseconds. The shorter timeout, the
+		// faster the connection recovers from a loss of an entire window
+		int utp_min_timeout;
+
+		// the number of SYN packets that are sent before giving up
+		int utp_syn_resends;
+		
+		// the number of resent packets sent on a closed socket before giving up
+		int utp_fin_resends;
+
+		// the number of times to send a packet before giving up
+		int utp_num_resends;
+
+		// initial timeout for uTP SYN packets
+		int utp_connect_timeout;
+
+		// number of milliseconds of delaying ACKing packets the most
+		int utp_delayed_ack;
+
+		// set to true if the uTP socket buffer size is allowed to increase
+		// dynamically based on the NIC MTU setting. This is true by default
+		// and improves uTP performance for networks with larger frame sizes
+		// including loopback
+		bool utp_dynamic_sock_buf;
+
+		// what to multiply the congestion window by on packet loss.
+		// it's specified as a percent. The default is 50, i.e. cut
+		// in half
+		int utp_loss_multiplier;
+
+		enum bandwidth_mixed_algo_t
+		{
+			// disables the mixed mode bandwidth balancing
+			prefer_tcp = 0,
+
+			// does not throttle uTP, throttles TCP to the same proportion
+			// of throughput as there are TCP connections
+			peer_proportional = 1
+
+		};
+		// the algorithm to use to balance bandwidth between tcp
+		// connections and uTP connections
+		int mixed_mode_algorithm;
+
+		// set to true if uTP connections should be rate limited
+		// defaults to false
+		bool rate_limit_utp;
+
+		// this is the number passed in to listen(). i.e.
+		// the number of connections to accept while we're
+		// not waiting in an accept() call.
+		int listen_queue_size;
+
+		// if this is true, the &ip= argument in tracker requests
+		// (unless otherwise specified) will be set to the intermediate
+		// IP address if the user is double NATed. If ther user is not
+		// double NATed, this option does not have an affect
+		bool announce_double_nat;
+
+		// the first tracker response after a torrent is started
+		// will cause this many connections to be made immediately.
+		// instead of waiting for the connection scheduler which
+		// triggeres every second
+		int torrent_connect_boost;
+
+		// this controls whether or not seeding (and complete) torrents
+		// attempt to make outgoing connections or not. It defaults to
+		// true, but can be set to zero for specific applications where
+		// making outgoing connections is costly and known to not
+		// add any benefits
+		bool seeding_outgoing_connections;
+
+		// when this is true, libtorrent will not attempt to make outgoing
+		// connections to peers whose port is < 1024. This is a safety
+		// precaution to avoid being part of a DDoS attack
+		bool no_connect_privileged_ports;
+
+		// the max alert queue size
+		int alert_queue_size;
+
+		// the max allowed size for metadata received by the
+		// ut_metadata extension (i.e. magnet links)
+		int max_metadata_size;
+
+		// attempt to smooth out connects to avoid getting spikes in
+		// opening connections and timing out connections
+		bool smooth_connects;
+
+		// always send user-agent
+		bool always_send_user_agent;
+
+		// if true, trackers will also be filtered by the IP
+		// filter, otherwise they are exempt
+		bool apply_ip_filter_to_trackers;
+
+		// to avoid write jobs starving read jobs, if this many
+		// write jobs have been taking priority in a row, service
+		// one read job
+		int read_job_every;
+
+		// issue posix_fadvise() or fcntl(F_RDADVISE) for disk reads
+		// ahead of time
+		bool use_disk_read_ahead;
+
+		// if set to true, files will be locked when opened.
+		// preventing any other process from modifying them
+		bool lock_files;
+
+		// open an ssl listen socket for ssl torrents on this port
+		int ssl_listen;
+
+		// this is the factor X in the formula to calculate the
+		// next tracker timeout:
+		// delay = 5 + X/100 * fails^2
+		// so, it's an exponential back-off, and this factor
+		// determines how fast the back-off happens. Default
+		// is 250
+		int tracker_backoff;
+
+		// when true, web seeds sending bad data will be banned
+		bool ban_web_seeds;
 	};
 
 #ifndef TORRENT_DISABLE_DHT
@@ -692,8 +939,15 @@ namespace libtorrent
 		dht_settings()
 			: max_peers_reply(100)
 			, search_branching(5)
+#ifndef TORRENT_NO_DEPRECATE
 			, service_port(0)
+#endif
 			, max_fail_count(20)
+			, max_torrents(2000)
+			, max_dht_items(700)
+			, max_torrent_search_reply(20)
+			, restrict_routing_ips(true)
+			, restrict_search_ips(true)
 		{}
 		
 		// the maximum number of peers to send in a
@@ -704,13 +958,37 @@ namespace libtorrent
 		// searching the DHT.
 		int search_branching;
 		
+#ifndef TORRENT_NO_DEPRECATE
 		// the listen port for the dht. This is a UDP port.
 		// zero means use the same as the tcp interface
 		int service_port;
+#endif
 		
 		// the maximum number of times a node can fail
 		// in a row before it is removed from the table.
 		int max_fail_count;
+
+		// this is the max number of torrents the DHT will track
+		int max_torrents;
+
+		// max number of items the DHT will store
+		int max_dht_items;
+
+		// the max number of torrents to return in a
+		// torrent search query to the DHT
+		int max_torrent_search_reply;
+
+		// when set, nodes whose IP address that's in
+		// the same /24 (or /64 for IPv6) range in the
+		// same routing table bucket. This is an attempt
+		// to mitigate node ID spoofing attacks
+		// also restrict any IP to only have a single
+		// entry in the whole routing table
+		bool restrict_routing_ips;
+
+		// applies the same IP restrictions on nodes
+		// received during a DHT search (traversal algorithm)
+		bool restrict_search_ips;
 	};
 #endif
 
@@ -734,9 +1012,9 @@ namespace libtorrent
 
 		enum enc_level
 		{
-			plaintext, // use only plaintext encryption
-			rc4, // use only rc4 encryption 
-			both // allow both
+			plaintext = 1, // use only plaintext encryption
+			rc4 = 2, // use only rc4 encryption 
+			both = 3 // allow both
 		};
 
 		enc_policy out_enc_policy;
