@@ -35,41 +35,40 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #include <string>
 #include <vector>
-//#include <iosfwd>
 
 #ifdef _MSC_VER
 #pragma warning(push, 1)
 #endif
 
 #include <boost/optional.hpp>
-#include <boost/filesystem/path.hpp>
 #include <boost/shared_array.hpp>
 
 #ifdef _MSC_VER
 #pragma warning(pop)
 #endif
 
+#include "libtorrent/config.hpp"
 #include "libtorrent/entry.hpp"
 #include "libtorrent/lazy_entry.hpp"
-#include "libtorrent/socket.hpp"
 #include "libtorrent/peer_id.hpp"
 #include "libtorrent/size_type.hpp"
-#include "libtorrent/config.hpp"
-#include "libtorrent/time.hpp"
+#include "libtorrent/ptime.hpp"
 #include "libtorrent/intrusive_ptr_base.hpp"
 #include "libtorrent/assert.hpp"
 #include "libtorrent/file_storage.hpp"
+#include "libtorrent/copy_ptr.hpp"
+#include "libtorrent/socket.hpp"
+#include "libtorrent/policy.hpp" // for policy::peer
 
 namespace libtorrent
 {
-	namespace pt = boost::posix_time;
-	namespace gr = boost::gregorian;
-	namespace fs = boost::filesystem;
+	class peer_connection;
+	struct session_settings;
 
 	enum
 	{
-		// wait 60 seconds before retrying a failed tracker
-		tracker_retry_delay_min = 10
+		// wait at least 5 seconds before retrying a failed tracker
+		tracker_retry_delay_min = 5
 		// when tracker_failed_max trackers
 		// has failed, wait 60 minutes instead
 		, tracker_retry_delay_max = 60 * 60
@@ -77,22 +76,20 @@ namespace libtorrent
 
 	struct TORRENT_EXPORT announce_entry
 	{
-		announce_entry(std::string const& u)
-			: url(u)
-			, next_announce(min_time())
-			, min_announce(min_time())
-			, tier(0)
-			, fail_limit(0)
-			, fails(0)
-			, source(0)
-			, verified(false)
-			, updating(false)
-			, start_sent(false)
-			, complete_sent(false)
-			, send_stats(true)
-		{}
+		announce_entry(std::string const& u);
+		~announce_entry();
 
+		// tracker URL as it appeared in the torrent file
 		std::string url;
+		std::string trackerid;
+
+		// if this tracker has returned an error or warning message
+		// that message is stored here
+		std::string message;
+
+		// if this tracker failed the last time it was contacted
+		// this error code specifies what error occurred
+		error_code last_error;
 
 		int next_announce_in() const;
 		int min_announce_in() const;
@@ -103,13 +100,19 @@ namespace libtorrent
 		// no announces before this time
 		ptime min_announce;
 
+		// the tier this tracker belongs to
 		boost::uint8_t tier;
+
 		// the number of times this tracker can fail
 		// in a row before it's removed. 0 means unlimited
 		boost::uint8_t fail_limit;
 
 		// the number of times in a row this tracker has failed
-		boost::uint8_t fails;
+		boost::uint8_t fails:7;
+
+		// true if we're currently trying to announce with 
+		// this tracker
+		bool updating:1;
 
 		enum tracker_source
 		{
@@ -118,17 +121,14 @@ namespace libtorrent
 			source_magnet_link = 4,
 			source_tex = 8
 		};
+
 		// where did we get this tracker from
-		boost::uint8_t source;
+		boost::uint8_t source:4;
 
 		// is set to true if we have ever received a response from
 		// this tracker
 		bool verified:1;
 
-		// true if we're currently trying to announce with 
-		// this tracker
-		bool updating:1;
-
 		// this is true if event start has been sent to the tracker
 		bool start_sent:1;
 
@@ -145,15 +145,7 @@ namespace libtorrent
 			min_announce = min_time();
 		}
 
-		void failed(int retry_interval = 0)
-		{
-			++fails;
-			int delay = (std::min)(tracker_retry_delay_min + int(fails) * int(fails)
-				* tracker_retry_delay_min, int(tracker_retry_delay_max));
-			delay = (std::max)(delay, retry_interval);
-			next_announce = time_now() + seconds(delay);
-			updating = false;
-		}
+		void failed(session_settings const& sett, int retry_interval = 0);
 
 		bool will_announce(ptime now) const
 		{
@@ -165,15 +157,59 @@ namespace libtorrent
 		bool can_announce(ptime now, bool is_seed) const;
 
 		bool is_working() const
+		{ return fails == 0; }
+
+		void trim();
+	};
+
+	struct web_seed_entry
+	{
+		// http seeds are different from url seeds in the
+		// protocol they use. http seeds follows the original
+		// http seed spec. by John Hoffman
+		enum type_t { url_seed, http_seed };
+
+		typedef std::vector<std::pair<std::string, std::string> > headers_t;
+
+		web_seed_entry(std::string const& url_, type_t type_
+			, std::string const& auth_ = std::string()
+			, headers_t const& extra_headers_ = headers_t());
+
+		bool operator==(web_seed_entry const& e) const
+		{ return url == e.url && type == e.type; }
+
+		bool operator<(web_seed_entry const& e) const
 		{
-			return fails == 0;
+			if (url < e.url) return true;
+			if (url > e.url) return false;
+		  	return type < e.type;
 		}
 
-		void trim()
-		{
-			while (!url.empty() && is_space(url[0]))
-				url.erase(url.begin());
-		}
+		std::string url;
+		type_t type;
+		std::string auth;
+		headers_t extra_headers;
+
+		// if this is > now, we can't reconnect yet
+		ptime retry;
+
+		// this indicates whether or not we're resolving the
+		// hostname of this URL
+		bool resolving;
+
+		// if the user wanted to remove this while
+		// we were resolving it. In this case, we set
+		// the removed flag to true, to make the resolver
+		// callback remove it
+		bool removed;
+
+		tcp::endpoint endpoint;
+
+		// this is the peer_info field used for the
+		// connection, just to count hash failures
+		// it's also used to hold the peer_connection
+		// pointer, when the web seed is connected
+		policy::peer peer_info;
 	};
 
 #ifndef BOOST_NO_EXCEPTIONS
@@ -181,29 +217,35 @@ namespace libtorrent
 	typedef libtorrent_exception invalid_torrent_file;
 #endif
 
-	int TORRENT_EXPORT load_file(fs::path const& filename, std::vector<char>& v, error_code& ec);
+	int TORRENT_EXPORT load_file(std::string const& filename
+		, std::vector<char>& v, error_code& ec, int limit = 8000000);
 
 	class TORRENT_EXPORT torrent_info : public intrusive_ptr_base<torrent_info>
 	{
 	public:
-		torrent_info(torrent_info const& t);
-#ifndef BOOST_NO_EXCEPTIONS
-		torrent_info(lazy_entry const& torrent_file);
-		torrent_info(char const* buffer, int size);
-		torrent_info(fs::path const& filename);
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
-		torrent_info(fs::wpath const& filename);
-#endif
+
+#ifdef TORRENT_DEBUG
+		void check_invariant() const;
 #endif
 
-		torrent_info(sha1_hash const& info_hash);
-		torrent_info(lazy_entry const& torrent_file, error_code& ec);
-		torrent_info(char const* buffer, int size, error_code& ec);
-		torrent_info(fs::path const& filename, error_code& ec);
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
-		torrent_info(fs::wpath const& filename, error_code& ec);
+#ifndef BOOST_NO_EXCEPTIONS
+		torrent_info(lazy_entry const& torrent_file, int flags = 0);
+		torrent_info(char const* buffer, int size, int flags = 0);
+		torrent_info(std::string const& filename, int flags = 0);
+#if TORRENT_USE_WSTRING
+		torrent_info(std::wstring const& filename, int flags = 0);
+#endif // TORRENT_USE_WSTRING
 #endif
 
+		torrent_info(torrent_info const& t, int flags = 0);
+		torrent_info(sha1_hash const& info_hash, int flags = 0);
+		torrent_info(lazy_entry const& torrent_file, error_code& ec, int flags = 0);
+		torrent_info(char const* buffer, int size, error_code& ec, int flags = 0);
+		torrent_info(std::string const& filename, error_code& ec, int flags = 0);
+#if TORRENT_USE_WSTRING
+		torrent_info(std::wstring const& filename, error_code& ec, int flags = 0);
+#endif // TORRENT_USE_WSTRING
+
 		~torrent_info();
 
 		file_storage const& files() const { return m_files; }
@@ -215,28 +257,37 @@ namespace libtorrent
 			m_files.rename_file(index, new_filename);
 		}
 
-#ifndef BOOST_FILESYSTEM_NARROW_ONLY
+#if TORRENT_USE_WSTRING
 		void rename_file(int index, std::wstring const& new_filename)
 		{
 			copy_on_write();
 			m_files.rename_file(index, new_filename);
 		}
-#endif
+#endif // TORRENT_USE_WSTRING
 
 		void remap_files(file_storage const& f);
 
 		void add_tracker(std::string const& url, int tier = 0);
 		std::vector<announce_entry> const& trackers() const { return m_urls; }
 
-		std::vector<std::string> const& url_seeds() const
-		{ return m_url_seeds; }
-		void add_url_seed(std::string const& url)
-		{ m_url_seeds.push_back(url); }
+#ifndef TORRENT_NO_DEPRECATE
+		// deprecated in 0.16. Use web_seeds() instead
+		TORRENT_DEPRECATED_PREFIX
+		std::vector<std::string> url_seeds() const TORRENT_DEPRECATED;
+		TORRENT_DEPRECATED_PREFIX
+		std::vector<std::string> http_seeds() const TORRENT_DEPRECATED;
+#endif // TORRENT_NO_DEPRECATE
 
-		std::vector<std::string> const& http_seeds() const
-		{ return m_http_seeds; }
-		void add_http_seed(std::string const& url)
-		{ m_http_seeds.push_back(url); }
+		void add_url_seed(std::string const& url
+			, std::string const& extern_auth = std::string()
+			, web_seed_entry::headers_t const& extra_headers = web_seed_entry::headers_t());
+
+		void add_http_seed(std::string const& url
+			, std::string const& extern_auth = std::string()
+			, web_seed_entry::headers_t const& extra_headers = web_seed_entry::headers_t());
+
+		std::vector<web_seed_entry> const& web_seeds() const
+		{ return m_web_seeds; }
 
 		size_type total_size() const { return m_files.total_size(); }
 		int piece_length() const { return m_files.piece_length(); }
@@ -252,7 +303,7 @@ namespace libtorrent
 		reverse_file_iterator rbegin_files() const { return m_files.rbegin(); }
 		reverse_file_iterator rend_files() const { return m_files.rend(); }
 		int num_files() const { return m_files.num_files(); }
-		file_entry const& file_at(int index) const { return m_files.at(index); }
+		file_entry file_at(int index) const { return m_files.at(index); }
 
 		file_iterator file_at_offset(size_type offset) const
 		{ return m_files.file_at_offset(offset); }
@@ -271,10 +322,16 @@ namespace libtorrent
 // ------- end deprecation -------
 #endif
 
+#ifdef TORRENT_USE_OPENSSL
+		std::string const& ssl_cert() const { return m_ssl_root_cert; }
+#endif
+
 		bool is_valid() const { return m_files.is_valid(); }
 
 		bool priv() const { return m_private; }
 
+		bool is_i2p() const { return m_i2p; }
+
 		int piece_size(int index) const { return m_files.piece_size(index); }
 
 		sha1_hash hash_for_piece(int index) const
@@ -290,7 +347,7 @@ namespace libtorrent
 			TORRENT_ASSERT(index < m_files.num_pieces());
 			if (is_merkle_torrent())
 			{
-				TORRENT_ASSERT(index < int(m_merkle_tree.size()) - m_merkle_first_leaf);
+				TORRENT_ASSERT(index < int(m_merkle_tree.size() - m_merkle_first_leaf));
 				return (const char*)&m_merkle_tree[m_merkle_first_leaf + index][0];
 			}
 			else
@@ -298,12 +355,12 @@ namespace libtorrent
 				TORRENT_ASSERT(m_piece_hashes);
 				TORRENT_ASSERT(m_piece_hashes >= m_info_section.get());
 				TORRENT_ASSERT(m_piece_hashes < m_info_section.get() + m_info_section_size);
-				TORRENT_ASSERT(index < m_info_section_size / 20);
+				TORRENT_ASSERT(index < int(m_info_section_size / 20));
 				return &m_piece_hashes[index*20];
 			}
 		}
 
-		boost::optional<pt::ptime> creation_date() const;
+		boost::optional<time_t> creation_date() const;
 
 		const std::string& creator() const
 		{ return m_created_by; }
@@ -319,13 +376,16 @@ namespace libtorrent
 		void add_node(std::pair<std::string, int> const& node)
 		{ m_nodes.push_back(node); }
 		
-		bool parse_info_section(lazy_entry const& e, error_code& ex);
+		bool parse_info_section(lazy_entry const& e, error_code& ec, int flags);
 
 		lazy_entry const* info(char const* key) const
 		{
 			if (m_info_dict.type() == lazy_entry::none_t)
+			{
+				error_code ec;
 				lazy_bdecode(m_info_section.get(), m_info_section.get()
-					+ m_info_section_size, m_info_dict);
+					+ m_info_section_size, m_info_dict, ec);
+			}
 			return m_info_dict.dict_find(key);
 		}
 
@@ -341,34 +401,51 @@ namespace libtorrent
 		std::map<int, sha1_hash> build_merkle_list(int piece) const;
 		bool is_merkle_torrent() const { return !m_merkle_tree.empty(); }
 
+		// if we're logging member offsets, we need access to them
+#if defined TORRENT_DEBUG \
+		&& !defined TORRENT_LOGGING \
+		&& !defined TORRENT_VERBOSE_LOGGING \
+		&& !defined TORRENT_ERROR_LOGGING
 	private:
+#endif
 
 		// not assignable
 		torrent_info const& operator=(torrent_info const&);
 
 		void copy_on_write();
-		bool parse_torrent_file(lazy_entry const& libtorrent, error_code& ec);
+		bool parse_torrent_file(lazy_entry const& libtorrent, error_code& ec, int flags);
+
+		// the index to the first leaf. This is where the hash for the
+		// first piece is stored
+		boost::uint32_t m_merkle_first_leaf;
 
 		file_storage m_files;
 
 		// if m_files is modified, it is first copied into
 		// m_orig_files so that the original name and
 		// filenames are preserved.
-		boost::shared_ptr<const file_storage> m_orig_files;
+		copy_ptr<const file_storage> m_orig_files;
 
 		// the urls to the trackers
 		std::vector<announce_entry> m_urls;
-		std::vector<std::string> m_url_seeds;
-		std::vector<std::string> m_http_seeds;
+		std::vector<web_seed_entry> m_web_seeds;
 		nodes_t m_nodes;
 
-		// the hash that identifies this torrent
-		sha1_hash m_info_hash;
+		// if this is a merkle torrent, this is the merkle
+		// tree. It has space for merkle_num_nodes(merkle_num_leafs(num_pieces))
+		// hashes
+		std::vector<sha1_hash> m_merkle_tree;
 
-		// if a creation date is found in the torrent file
-		// this will be set to that, otherwise it'll be
-		// 1970, Jan 1
-		pt::ptime m_creation_date;
+		// this is a copy of the info section from the torrent.
+		// it use maintained in this flat format in order to
+		// make it available through the metadata extension
+		boost::shared_array<char> m_info_section;
+
+		// this is a pointer into the m_info_section buffer
+		// pointing to the first byte of the first sha-1 hash
+		char const* m_piece_hashes;
+
+		// TODO: these strings could be lazy_entry* to save memory
 
 		// if a comment is found in the torrent file
 		// this will be set to that comment
@@ -378,38 +455,44 @@ namespace libtorrent
 		// to create the torrent file
 		std::string m_created_by;
 
+#ifdef TORRENT_USE_OPENSSL
+		// for ssl-torrens, this contains the root
+		// certificate, in .pem format (i.e. ascii
+		// base64 encoded with head and tails)
+		std::string m_ssl_root_cert;
+#endif
+
+		// the info section parsed. points into m_info_section
+		// parsed lazily
+		mutable lazy_entry m_info_dict;
+
+		// if a creation date is found in the torrent file
+		// this will be set to that, otherwise it'll be
+		// 1970, Jan 1
+		time_t m_creation_date;
+
+		// the hash that identifies this torrent
+		sha1_hash m_info_hash;
+
+		// the number of bytes in m_info_section
+		boost::uint32_t m_info_section_size:24;
+
 		// this is used when creating a torrent. If there's
 		// only one file there are cases where it's impossible
 		// to know if it should be written as a multifile torrent
 		// or not. e.g. test/test  there's one file and one directory
 		// and they have the same name.
-		bool m_multifile;
+		bool m_multifile:1;
 		
 		// this is true if the torrent is private. i.e., is should not
 		// be announced on the dht
-		bool m_private;
+		bool m_private:1;
 
-		// this is a copy of the info section from the torrent.
-		// it use maintained in this flat format in order to
-		// make it available through the metadata extension
-		boost::shared_array<char> m_info_section;
-		int m_info_section_size;
-
-		// this is a pointer into the m_info_section buffer
-		// pointing to the first byte of the first sha-1 hash
-		char const* m_piece_hashes;
-
-		// if this is a merkle torrent, this is the merkle
-		// tree. It has space for merkle_num_nodes(merkle_num_leafs(num_pieces))
-		// hashes
-		std::vector<sha1_hash> m_merkle_tree;
-		// the index to the first leaf. This is where the hash for the
-		// first piece is stored
-		int m_merkle_first_leaf;
-
-		// the info section parsed. points into m_info_section
-		// parsed lazily
-		mutable lazy_entry m_info_dict;
+		// this is true if one of the trackers has an .i2p top
+		// domain in its hostname. This means the DHT and LSD
+		// features are disabled for this torrent (unless the
+		// settings allows mixing i2p peers with regular peers)
+		bool m_i2p:1;
 	};
 
 }
