@@ -35,8 +35,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/storage.hpp"
 #include "libtorrent/escape_string.hpp"
 
-#include <boost/date_time/posix_time/posix_time_types.hpp>
-#include <boost/date_time/gregorian/greg_date.hpp>
 #include <boost/bind.hpp>
 #include <boost/next_prior.hpp>
 
@@ -45,8 +43,6 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #define MAX_SYMLINK_PATH 200
 
-namespace gr = boost::gregorian;
-
 namespace libtorrent
 {
 	// defined in torrent_info.cpp
@@ -57,48 +53,23 @@ namespace libtorrent
 
 	namespace detail
 	{
-		int TORRENT_EXPORT get_file_attributes(boost::filesystem::path const& p)
+		int get_file_attributes(std::string const& p)
 		{
 #ifdef TORRENT_WINDOWS
 
-#if TORRENT_USE_WPATH
-			std::wstring path = convert_to_wstring(p.external_file_string());
+#if TORRENT_USE_WSTRING
+			std::wstring path = convert_to_wstring(p);
 			DWORD attr = GetFileAttributesW(path.c_str());
 #else
-			std::string path = convert_to_native(p.external_file_string());
+			std::string path = convert_to_native(p);
 			DWORD attr = GetFileAttributesA(path.c_str());
-#endif
+#endif // TORRENT_USE_WSTRING
 			if (attr == INVALID_FILE_ATTRIBUTES) return 0;
 			if (attr & FILE_ATTRIBUTE_HIDDEN) return file_storage::attribute_hidden;
 			return 0;
 #else
 			struct stat s;
-			if (lstat(convert_to_native(p.external_file_string()).c_str(), &s) < 0) return 0;
-			int file_attr = 0;
-			if (s.st_mode & S_IXUSR) 
-				file_attr += file_storage::attribute_executable;
-			if(S_ISLNK(s.st_mode))
-				file_attr += file_storage::attribute_symlink;
-			return file_attr;
-#endif
-		}
-	
-#if TORRENT_USE_WPATH
-		int TORRENT_EXPORT get_file_attributes(boost::filesystem::wpath const& p)
-		{
-#ifdef TORRENT_WINDOWS
-			std::wstring const& path = p.external_file_string();
-			DWORD attr = GetFileAttributesW(path.c_str());
-			if (attr == INVALID_FILE_ATTRIBUTES) return 0;
-			if (attr & FILE_ATTRIBUTE_HIDDEN) return file_storage::attribute_hidden;
-			return 0;
-#else
-			std::string native;
-			wchar_utf8(p.string(), native);
-			native = convert_to_native(native);
-
-			struct stat s;
-			if (lstat(native.c_str(), &s) < 0) return 0;
+			if (lstat(convert_to_native(p).c_str(), &s) < 0) return 0;
 			int file_attr = 0;
 			if (s.st_mode & S_IXUSR) 
 				file_attr += file_storage::attribute_executable;
@@ -107,112 +78,211 @@ namespace libtorrent
 			return file_attr;
 #endif
 		}
-#endif // TORRENT_USE_WPATH
-
-		std::time_t get_file_mtime(char const* path)
-		{
-#ifdef TORRENT_WINDOWS
-			struct _stat s;
-			if (::_stat(path, &s) < 0) return 0;
-#else
-			struct stat s;
-			if (lstat(path, &s) < 0) return 0;
-#endif
-			return s.st_mtime;
-		}
-
-		std::time_t TORRENT_EXPORT get_file_mtime(boost::filesystem::path const& p)
-		{
-#if defined TORRENT_WINDOWS && TORRENT_USE_WPATH
-			std::wstring path = convert_to_wstring(p.external_file_string());
-			struct _stat s;
-			if (::_wstat(path.c_str(), &s) < 0) return 0;
-			return s.st_mtime;
-#else
-			std::string path = convert_to_native(p.external_file_string());
-			return get_file_mtime(p.string().c_str());
-#endif
-		}
-
-#if TORRENT_USE_WPATH
-		std::time_t TORRENT_EXPORT get_file_mtime(boost::filesystem::wpath const& p)
-		{
-#ifdef TORRENT_WINDOWS
-			struct _stat s;
-			if (::_wstat(p.string().c_str(), &s) < 0) return 0;
-			return s.st_mtime;
-#else
-			std::string utf8;
-			wchar_utf8(p.string(), utf8);
-			utf8 = convert_to_native(utf8);
-			return get_file_mtime(utf8.c_str());
-#endif
-		}
-#endif // TORRENT_USE_WPATH
-
+	
 #ifndef TORRENT_WINDOWS
-		boost::filesystem::path get_symlink_path_impl(char const* path)
+		std::string get_symlink_path_impl(char const* path)
 		{
 			char buf[MAX_SYMLINK_PATH];
-			int char_read = readlink(path,buf,MAX_SYMLINK_PATH);
+			std::string f = convert_to_native(path);
+			int char_read = readlink(f.c_str(),buf,MAX_SYMLINK_PATH);
 			if (char_read < 0) return "";
 			if (char_read < MAX_SYMLINK_PATH) buf[char_read] = 0;
 			else buf[0] = 0;
-			return buf;
+			return convert_from_native(buf);
 		}
 #endif
 
-		boost::filesystem::path TORRENT_EXPORT get_symlink_path(boost::filesystem::path const& p)
+		std::string get_symlink_path(std::string const& p)
 		{
 #if defined TORRENT_WINDOWS
 			return "";
 #else
-			std::string path = convert_to_native(p.external_file_string());
-			return get_symlink_path_impl(p.string().c_str());
+			std::string path = convert_to_native(p);
+			return get_symlink_path_impl(p.c_str());
 #endif
 		}
 
-#if TORRENT_USE_WPATH
-		boost::filesystem::path TORRENT_EXPORT get_symlink_path(boost::filesystem::wpath const& p)
+		void add_files_impl(file_storage& fs, std::string const& p
+			, std::string const& l, boost::function<bool(std::string)> pred, boost::uint32_t flags)
 		{
-#ifdef TORRENT_WINDOWS
-			return "";
-#else
-			std::string utf8;
-			wchar_utf8(p.string(), utf8);
-			utf8 = convert_to_native(utf8);
-			return get_symlink_path_impl(utf8.c_str());
-#endif
-		}
-#endif // TORRENT_USE_WPATH
+			std::string f = combine_path(p, l);
+			if (!pred(f)) return;
+			error_code ec;
+			file_status s;
+			stat_file(f, &s, ec, (flags & create_torrent::symlinks) ? dont_follow_links : 0);
+			if (ec) return;
 
+			// recurse into directories
+			bool recurse = (s.mode & file_status::directory) != 0;
+
+			// if the file is not a link or we're following links, and it's a directory
+			// only then should we recurse
+#ifndef TORRENT_WINDOWS
+			if ((s.mode & file_status::link) && (flags & create_torrent::symlinks))
+				recurse = false;
+#endif
+
+			if (recurse)
+			{
+				for (directory i(f, ec); !i.done(); i.next(ec))
+				{
+					std::string leaf = i.file();
+					if (ignore_subdir(leaf)) continue;
+					add_files_impl(fs, p, combine_path(l, leaf), pred, flags);
+				}
+			}
+			else
+			{
+				// #error use the fields from s
+				int file_flags = get_file_attributes(f);
+
+				// mask all bits to check if the file is a symlink
+				if ((file_flags & file_storage::attribute_symlink)
+					&& (flags & create_torrent::symlinks)) 
+				{
+					std::string sym_path = get_symlink_path(f);
+					fs.add_file(l, 0, file_flags, s.mtime, sym_path);
+				}
+				else
+				{
+					fs.add_file(l, s.file_size, file_flags, s.mtime);
+				}
+			}
+		}
 	}
 
+	struct piece_holder
+	{
+		piece_holder(int bytes): m_piece(page_aligned_allocator::malloc(bytes)) {}
+		~piece_holder() { page_aligned_allocator::free(m_piece); }
+		char* bytes() { return m_piece; }
+	private:
+		char* m_piece;
+	};
+
+#if TORRENT_USE_WSTRING
+	void set_piece_hashes(create_torrent& t, std::wstring const& p
+		, boost::function<void(int)> const& f, error_code& ec)
+	{
+		file_pool fp;
+		std::string utf8;
+		wchar_utf8(p, utf8);
+#if TORRENT_USE_UNC_PATHS
+		utf8 = canonicalize_path(utf8);
+#endif
+		boost::scoped_ptr<storage_interface> st(
+			default_storage_constructor(const_cast<file_storage&>(t.files()), 0, utf8, fp
+			, std::vector<boost::uint8_t>()));
+
+		// calculate the hash for all pieces
+		int num = t.num_pieces();
+		std::vector<char> buf(t.piece_length());
+		for (int i = 0; i < num; ++i)
+		{
+			// read hits the disk and will block. Progress should
+			// be updated in between reads
+			st->read(&buf[0], i, 0, t.piece_size(i));
+			if (st->error())
+			{
+				ec = st->error();
+				return;
+			}
+			hasher h(&buf[0], t.piece_size(i));
+			t.set_hash(i, h.final());
+			f(i);
+		}
+	}
+#endif
+
+	void set_piece_hashes(create_torrent& t, std::string const& p
+		, boost::function<void(int)> f, error_code& ec)
+	{
+		file_pool fp;
+#if TORRENT_USE_UNC_PATHS
+		std::string path = canonicalize_path(p);
+#else
+		std::string const& path = p;
+#endif
+		boost::scoped_ptr<storage_interface> st(
+			default_storage_constructor(const_cast<file_storage&>(t.files()), 0, path, fp
+			, std::vector<boost::uint8_t>()));
+
+		// if we're calculating file hashes as well, use this hasher
+		hasher filehash;
+		int file_idx = 0;
+		size_type left_in_file = t.files().at(0).size;
+
+		// calculate the hash for all pieces
+		int num = t.num_pieces();
+		piece_holder buf(t.piece_length());
+		for (int i = 0; i < num; ++i)
+		{
+			// read hits the disk and will block. Progress should
+			// be updated in between reads
+			st->read(buf.bytes(), i, 0, t.piece_size(i));
+			if (st->error())
+			{
+				ec = st->error();
+				return;
+			}
+			
+			if (t.should_add_file_hashes())
+			{
+				int left_in_piece = t.piece_size(i);
+				int this_piece_size = left_in_piece;
+				// the number of bytes from this file we just read
+				while (left_in_piece > 0)
+				{
+					int to_hash_for_file = int((std::min)(size_type(left_in_piece), left_in_file));
+					if (to_hash_for_file > 0)
+					{
+						int offset = this_piece_size - left_in_piece;
+						filehash.update(buf.bytes() + offset, to_hash_for_file);
+					}
+					left_in_file -= to_hash_for_file;
+					left_in_piece -= to_hash_for_file;
+					if (left_in_file == 0)
+					{
+						if (!t.files().at(file_idx).pad_file)
+							t.set_file_hash(file_idx, filehash.final());
+						filehash.reset();
+						file_idx++;
+						if (file_idx >= t.files().num_files()) break;
+						left_in_file = t.files().at(file_idx).size;
+					}
+				}
+			}
+
+			hasher h(buf.bytes(), t.piece_size(i));
+			t.set_hash(i, h.final());
+			f(i);
+		}
+	}
+
+	create_torrent::~create_torrent() {}
+
 	create_torrent::create_torrent(file_storage& fs, int piece_size, int pad_file_limit, int flags)
 		: m_files(fs)
-		, m_creation_date(pt::second_clock::universal_time())
+		, m_creation_date(time(0))
 		, m_multifile(fs.num_files() > 1)
 		, m_private(false)
-		, m_merkle_torrent(flags & merkle)
-		, m_include_mtime(flags & modification_time)
-		, m_include_symlinks(flags & symlinks)
+		, m_merkle_torrent((flags & merkle) != 0)
+		, m_include_mtime((flags & modification_time) != 0)
+		, m_include_symlinks((flags & symlinks) != 0)
+		, m_calculate_file_hashes((flags & calculate_file_hashes) != 0)
 	{
 		TORRENT_ASSERT(fs.num_files() > 0);
 
 		// return instead of crash in release mode
 		if (fs.num_files() == 0) return;
 
-#if BOOST_VERSION < 103600
-		if (!m_multifile && m_files.at(0).path.has_branch_path()) m_multifile = true;
-#else
-		if (!m_multifile && m_files.at(0).path.has_parent_path()) m_multifile = true;
-#endif
+		if (!m_multifile && has_parent_path(m_files.file_path(*m_files.begin()))) m_multifile = true;
 
 		// a piece_size of 0 means automatic
 		if (piece_size == 0 && !m_merkle_torrent)
 		{
 			const int target_size = 40 * 1024;
-			piece_size = fs.total_size() / (target_size / 20);
+			piece_size = int(fs.total_size() / (target_size / 20));
 	
 			int i = 16*1024;
 			for (; i < 2*1024*1024; i *= 2)
@@ -248,12 +318,13 @@ namespace libtorrent
 
 	create_torrent::create_torrent(torrent_info const& ti)
 		: m_files(const_cast<file_storage&>(ti.files()))
-		, m_creation_date(pt::second_clock::universal_time())
+		, m_creation_date(time(0))
 		, m_multifile(ti.num_files() > 1)
 		, m_private(ti.priv())
 		, m_merkle_torrent(ti.is_merkle_torrent())
 		, m_include_mtime(false)
 		, m_include_symlinks(false)
+		, m_calculate_file_hashes(false)
 	{
 		TORRENT_ASSERT(ti.is_valid());
 		if (ti.creation_date()) m_creation_date = *ti.creation_date();
@@ -271,10 +342,15 @@ namespace libtorrent
 			, end(trackers.end()); i != end; ++i)
 			add_tracker(i->url, i->tier);
 
-		std::vector<std::string> const& web_seeds = ti.url_seeds();
-		for (std::vector<std::string>::const_iterator i = web_seeds.begin()
+		std::vector<web_seed_entry> const& web_seeds = ti.web_seeds();
+		for (std::vector<web_seed_entry>::const_iterator i = web_seeds.begin()
 			, end(web_seeds.end()); i != end; ++i)
-			add_url_seed(*i);
+		{
+			if (i->type == web_seed_entry::url_seed)
+				add_url_seed(i->url);
+			else if (i->type == web_seed_entry::http_seed)
+				add_http_seed(i->url);
+		}
 
 		m_piece_hash.resize(m_files.num_pieces());
 		for (int i = 0; i < num_pieces(); ++i) set_hash(i, ti.hash_for_piece(i));
@@ -331,8 +407,7 @@ namespace libtorrent
 		if (!m_comment.empty())
 			dict["comment"] = m_comment;
 
-		dict["creation date"] =
-			(m_creation_date - pt::ptime(gr::date(1970, gr::Jan, 1))).total_seconds();
+		dict["creation date"] = m_creation_date;
 
 		if (!m_created_by.empty())
 			dict["created by"] = m_created_by;
@@ -354,10 +429,6 @@ namespace libtorrent
 			}
 		}
 
-		// if this function is used with an older version of the library
-		// this member won't exist
-		// TODO: for distributions that needs to be binary compatible with
-		// previos 0.15.x releases, this if-block should be removed
 		if (!m_http_seeds.empty())
 		{
 			if (m_http_seeds.size() == 1)
@@ -384,29 +455,39 @@ namespace libtorrent
 
 		info["name"] = m_files.name();
 
+		if (!m_root_cert.empty())
+			info["ssl-cert"] = m_root_cert;
+
 		if (m_private) info["private"] = 1;
 
 		if (!m_multifile)
 		{
-			if (m_include_mtime) info["mtime"] = m_files.at(0).mtime;
-			info["length"] = m_files.at(0).size;
-			if (m_files.at(0).pad_file || m_files.at(0).hidden_attribute || m_files.at(0).executable_attribute || m_files.at(0).symlink_attribute)
+			file_entry e = m_files.at(0);
+			if (m_include_mtime) info["mtime"] = e.mtime;
+			info["length"] = e.size;
+			if (e.pad_file
+				|| e.hidden_attribute
+				|| e.executable_attribute
+				|| e.symlink_attribute)
 			{
 				std::string& attr = info["attr"].string();
-				if (m_files.at(0).pad_file) attr += 'p';
-				if (m_files.at(0).hidden_attribute) attr += 'h';
-				if (m_files.at(0).executable_attribute) attr += 'x';
-				if (m_include_symlinks && m_files.at(0).symlink_attribute) attr += 'l';
+				if (e.pad_file) attr += 'p';
+				if (e.hidden_attribute) attr += 'h';
+				if (e.executable_attribute) attr += 'x';
+				if (m_include_symlinks && e.symlink_attribute) attr += 'l';
 			}
-			if (m_include_symlinks && m_files.at(0).symlink_attribute)
+			if (m_include_symlinks
+				&& e.symlink_attribute)
 			{
 				entry& sympath_e = info["symlink path"];
 				
-				for (fs::path::iterator j = (m_files.at(0).symlink_path.begin());
-					j != m_files.at(0).symlink_path.end(); ++j)
-				{
-					sympath_e.list().push_back(entry(*j));
-				}
+				std::string split = split_path(e.symlink_path);
+				for (char const* e = split.c_str(); e != 0; e = next_path_element(e))
+					sympath_e.list().push_back(entry(e));
+			}
+			if (!m_filehashes.empty())
+			{
+				info["sha1"] = m_filehashes[0].to_string();
 			}
 		}
 		else
@@ -420,23 +501,23 @@ namespace libtorrent
 				{
 					files.list().push_back(entry());
 					entry& file_e = files.list().back();
-					if (m_include_mtime) file_e["mtime"] = i->mtime; 
+					if (m_include_mtime && m_files.mtime(*i)) file_e["mtime"] = m_files.mtime(*i); 
 					file_e["length"] = i->size;
 					entry& path_e = file_e["path"];
 
-#if BOOST_VERSION < 103600
-					TORRENT_ASSERT(i->path.has_branch_path());
-#else
-					TORRENT_ASSERT(i->path.has_parent_path());
-#endif
-					TORRENT_ASSERT(*i->path.begin() == m_files.name());
+					TORRENT_ASSERT(has_parent_path(m_files.file_path(*i)));
 
-					for (fs::path::iterator j = boost::next(i->path.begin());
-						j != i->path.end(); ++j)
-					{
-						path_e.list().push_back(entry(*j));
-					}
-					if (i->pad_file || i->hidden_attribute || i->executable_attribute || i->symlink_attribute)
+					std::string split = split_path(m_files.file_path(*i));
+					TORRENT_ASSERT(split.c_str() == m_files.name());
+
+					for (char const* e = next_path_element(split.c_str());
+						e != 0; e = next_path_element(e))
+						path_e.list().push_back(entry(e));
+
+					if (i->pad_file
+						|| i->hidden_attribute
+						|| i->executable_attribute
+						|| i->symlink_attribute)
 					{
 						std::string& attr = file_e["attr"].string();
 						if (i->pad_file) attr += 'p';
@@ -444,15 +525,20 @@ namespace libtorrent
 						if (i->executable_attribute) attr += 'x';
 						if (m_include_symlinks && i->symlink_attribute) attr += 'l';
 					}
-					if (m_include_symlinks && i->symlink_attribute)
+					if (m_include_symlinks
+						&& i->symlink_attribute
+						&& i->symlink_index != -1)
 					{
 						entry& sympath_e = file_e["symlink path"];
 
-						for (fs::path::iterator j = (i->symlink_path.begin());
-							j != i->symlink_path.end(); ++j)
-						{
-							sympath_e.list().push_back(entry(*j));
-						}
+						std::string split = split_path(m_files.symlink(*i));
+						for (char const* e = split.c_str(); e != 0; e = next_path_element(e))
+							sympath_e.list().push_back(entry(e));
+					}
+					int file_index = i - m_files.begin();
+					if (!m_filehashes.empty() && m_filehashes[file_index] != sha1_hash())
+					{
+						file_e["sha1"] = m_filehashes[file_index].to_string();
 					}
 				}
 			}
@@ -520,6 +606,11 @@ namespace libtorrent
 			, boost::bind(&announce_entry::second, _1) < boost::bind(&announce_entry::second, _2));
 	}
 
+	void create_torrent::set_root_cert(std::string const& cert)
+	{
+		m_root_cert = cert;
+	}
+
 	void create_torrent::set_hash(int index, sha1_hash const& h)
 	{
 		TORRENT_ASSERT(index >= 0);
@@ -527,6 +618,14 @@ namespace libtorrent
 		m_piece_hash[index] = h;
 	}
 
+	void create_torrent::set_file_hash(int index, sha1_hash const& h)
+	{
+		TORRENT_ASSERT(index >= 0);
+		TORRENT_ASSERT(index < (int)m_files.num_files());
+		if (m_filehashes.empty()) m_filehashes.resize(m_files.num_files());
+		m_filehashes[index] = h;
+	}
+
 	void create_torrent::add_node(std::pair<std::string, int> const& node)
 	{
 		m_nodes.push_back(node);
