@@ -34,6 +34,7 @@ POSSIBILITY OF SUCH DAMAGE.
 
 #include "libtorrent/config.hpp"
 #include "libtorrent/error_code.hpp"
+#include "libtorrent/escape_string.hpp" // for to_string()
 
 namespace libtorrent
 {
@@ -156,11 +157,11 @@ namespace libtorrent
 			"pex message too large",
 			"invalid pex message",
 			"invalid lt_tracker message",
-			"",
-			"",
+			"pex messages sent too frequent (possible attack)",
+			"torrent has no metadata",
 			"invalid dont-have message",
-			"",
-			"",
+			"SSL connection required",
+			"invalid SSL certificate",
 			"",
 			"",
 			"",
@@ -214,12 +215,87 @@ namespace libtorrent
 			"",
 			"",
 			"",
+
+// i2p errors
+			"no i2p router is set up",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+
+// tracker errors
+			"scrape not available on tracker",
+			"invalid tracker response",
+			"invalid peer dictionary entry",
+			"tracker sent a failure message",
+			"missing or invalid 'files' entry",
+			"missing or invalid 'hash' entry",
+			"missing or invalid 'peers' and 'peers6' entry",
+			"udp tracker response packet has invalid size",
+			"invalid transaction id in udp tracker response",
+			"invalid action field in udp tracker response",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+			"",
+
+// bdecode errors
+			"expected string in bencoded string",
+			"expected colon in bencoded string",
+			"unexpected end of file in bencoded string",
+			"expected value (list, dict, int or string) in bencoded string",
+			"bencoded nesting depth exceeded",
+			"bencoded item count limit exceeded",
 		};
 		if (ev < 0 || ev >= int(sizeof(msgs)/sizeof(msgs[0])))
 			return "Unknown error";
 		return msgs[ev];
 	}
 
+	const char* http_error_category::name() const
+	{
+		return "http error";
+	}
+
+	std::string http_error_category::message(int ev) const
+	{
+		std::string ret;
+		ret += to_string(ev).elems;
+		ret += " ";
+		switch (ev)
+		{
+			case errors::cont: ret += "Continue"; break;
+			case errors::ok: ret += "OK"; break;
+			case errors::created: ret += "Created"; break;
+			case errors::accepted: ret += "Accepted"; break;
+			case errors::no_content: ret += "No Content"; break;
+			case errors::multiple_choices: ret += "Multiple Choices"; break;
+			case errors::moved_permanently: ret += "Moved Permanently"; break;
+			case errors::moved_temporarily: ret += "Moved Temporarily"; break;
+			case errors::not_modified: ret += "Not Modified"; break;
+			case errors::bad_request: ret += "Bad Request"; break;
+			case errors::unauthorized: ret += "Unauthorized"; break;
+			case errors::forbidden: ret += "Forbidden"; break;
+			case errors::not_found: ret += "Not Found"; break;
+			case errors::internal_server_error: ret += "Internal Server Error"; break;
+			case errors::not_implemented: ret += "Not Implemented"; break;
+			case errors::bad_gateway: ret += "Bad Gateway"; break;
+			case errors::service_unavailable: ret += "Service Unavailable"; break;
+			default: ret += "(unknown HTTP error)"; break;
+		}
+		return ret;
+	}
 #endif
 
 }
