@@ -37,6 +37,11 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/parse_url.hpp"
 #include "libtorrent/socket.hpp"
 #include "libtorrent/connection_queue.hpp"
+#include "libtorrent/socket_type.hpp" // for async_shutdown
+
+#if defined TORRENT_ASIO_DEBUGGING
+#include "libtorrent/debug.hpp"
+#endif
 
 #include <boost/bind.hpp>
 #include <string>
@@ -46,10 +51,64 @@ namespace libtorrent {
 
 enum { max_bottled_buffer = 1024 * 1024 };
 
+http_connection::http_connection(io_service& ios, connection_queue& cc
+	, http_handler const& handler, bool bottled
+	, http_connect_handler const& ch
+	, http_filter_handler const& fh
+#ifdef TORRENT_USE_OPENSSL
+	, boost::asio::ssl::context* ssl_ctx
+#endif
+	)
+	: m_sock(ios)
+#if TORRENT_USE_I2P
+	, m_i2p_conn(0)
+#endif
+	, m_read_pos(0)
+	, m_resolver(ios)
+	, m_handler(handler)
+	, m_connect_handler(ch)
+	, m_filter_handler(fh)
+	, m_timer(ios)
+	, m_last_receive(time_now())
+	, m_start_time(time_now())
+	, m_bottled(bottled)
+	, m_called(false)
+#ifdef TORRENT_USE_OPENSSL
+	, m_ssl_ctx(ssl_ctx)
+	, m_own_ssl_context(false)
+#endif
+	, m_rate_limit(0)
+	, m_download_quota(0)
+	, m_limiter_timer_active(false)
+	, m_limiter_timer(ios)
+	, m_redirects(5)
+	, m_connection_ticket(-1)
+	, m_cc(cc)
+	, m_ssl(false)
+	, m_priority(0)
+	, m_abort(false)
+{
+	TORRENT_ASSERT(!m_handler.empty());
+}
+
+http_connection::~http_connection()
+{
+	TORRENT_ASSERT(m_connection_ticket == -1);
+#ifdef TORRENT_USE_OPENSSL
+	if (m_own_ssl_context) delete m_ssl_ctx;
+#endif
+}
+
 void http_connection::get(std::string const& url, time_duration timeout, int prio
 	, proxy_settings const* ps, int handle_redirects, std::string const& user_agent
-	, address const& bind_addr)
+	, address const& bind_addr
+#if TORRENT_USE_I2P
+	, i2p_connection* i2p_conn
+#endif
+	)
 {
+	m_user_agent = user_agent;
+
 	std::string protocol;
 	std::string auth;
 	std::string hostname;
@@ -98,13 +157,15 @@ void http_connection::get(std::string const& url, time_duration timeout, int pri
 #define APPEND_FMT1(fmt, arg) ptr += snprintf(ptr, end - ptr, fmt, arg)
 #define APPEND_FMT2(fmt, arg1, arg2) ptr += snprintf(ptr, end - ptr, fmt, arg1, arg2)
 
+	// exclude ssl here, because SSL assumes CONNECT support in the
+	// proxy and is handled at the lower layer
 	if (ps && (ps->type == proxy_settings::http
 		|| ps->type == proxy_settings::http_pw)
 		&& !ssl)
 	{
 		// if we're using an http proxy and not an ssl
 		// connection, just do a regular http proxy request
-		APPEND_FMT1("GET %s HTTP/1.0\r\n", url.c_str());
+		APPEND_FMT1("GET %s HTTP/1.1\r\n", url.c_str());
 		if (ps->type == proxy_settings::http_pw)
 			APPEND_FMT1("Proxy-Authorization: Basic %s\r\n", base64encode(
 				ps->username + ":" + ps->password).c_str());
@@ -113,17 +174,19 @@ void http_connection::get(std::string const& url, time_duration timeout, int pri
 	}
 	else
 	{
-		APPEND_FMT2("GET %s HTTP/1.0\r\n"
+		APPEND_FMT2("GET %s HTTP/1.1\r\n"
 			"Host: %s", path.c_str(), hostname.c_str());
 		if (port != default_port) APPEND_FMT1(":%d\r\n", port);
 		else APPEND_FMT("\r\n");
 	}
 
+//	APPEND_FMT("Accept: */*\r\n");
+
 	if (!auth.empty())
 		APPEND_FMT1("Authorization: Basic %s\r\n", base64encode(auth).c_str());
 
-	if (!user_agent.empty())
-		APPEND_FMT1("User-Agent: %s\r\n", user_agent.c_str());
+	if (!m_user_agent.empty())
+		APPEND_FMT1("User-Agent: %s\r\n", m_user_agent.c_str());
 	
 	if (m_bottled)
 		APPEND_FMT("Accept-Encoding: gzip\r\n");
@@ -133,12 +196,20 @@ void http_connection::get(std::string const& url, time_duration timeout, int pri
 	sendbuffer.assign(request);
 	m_url = url;
 	start(hostname, to_string(port).elems, timeout, prio
-		, ps, ssl, handle_redirects, bind_addr);
+		, ps, ssl, handle_redirects, bind_addr
+#if TORRENT_USE_I2P
+		, i2p_conn
+#endif
+		);
 }
 
 void http_connection::start(std::string const& hostname, std::string const& port
 	, time_duration timeout, int prio, proxy_settings const* ps, bool ssl, int handle_redirects
-	, address const& bind_addr)
+	, address const& bind_addr
+#if TORRENT_USE_I2P
+	, i2p_connection* i2p_conn
+#endif
+	)
 {
 	TORRENT_ASSERT(prio >= 0 && prio < 3);
 
@@ -149,9 +220,13 @@ void http_connection::start(std::string const& hostname, std::string const& port
 	// deletes this object
 	boost::shared_ptr<http_connection> me(shared_from_this());
 
-	m_timeout = timeout;
+	m_completion_timeout = timeout;
+	m_read_timeout = (std::max)(seconds(5), timeout / 5);
 	error_code ec;
-	m_timer.expires_from_now(m_timeout, ec);
+	m_timer.expires_from_now(m_completion_timeout, ec);
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_timeout");
+#endif
 	m_timer.async_wait(boost::bind(&http_connection::on_timeout
 		, boost::weak_ptr<http_connection>(me), _1));
 	m_called = false;
@@ -170,6 +245,9 @@ void http_connection::start(std::string const& hostname, std::string const& port
 	if (m_sock.is_open() && m_hostname == hostname && m_port == port
 		&& m_ssl == ssl && m_bind_addr == bind_addr)
 	{
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("http_connection::on_write");
+#endif
 		async_write(m_sock, asio::buffer(sendbuffer)
 			, boost::bind(&http_connection::on_write, me, _1));
 	}
@@ -178,41 +256,71 @@ void http_connection::start(std::string const& hostname, std::string const& port
 		m_ssl = ssl;
 		m_bind_addr = bind_addr;
 		error_code ec;
-		m_sock.close(ec);
+		if (m_sock.is_open()) m_sock.close(ec);
+
+#if TORRENT_USE_I2P
+		bool is_i2p = false;
+		char const* top_domain = strrchr(hostname.c_str(), '.');
+		if (top_domain && strcmp(top_domain, ".i2p") == 0 && i2p_conn)
+		{
+			// this is an i2p name, we need to use the sam connection
+			// to do the name lookup
+			is_i2p = true;
+			m_i2p_conn = i2p_conn;
+			// quadruple the timeout for i2p destinations
+			// because i2p is sloooooow
+			m_completion_timeout *= 4;
+			m_read_timeout *= 4;
+		}
+#endif
+
+#if TORRENT_USE_I2P
+		if (is_i2p && i2p_conn->proxy().type != proxy_settings::i2p_proxy)
+		{
+			m_resolver.get_io_service().post(boost::bind(&http_connection::callback
+				, me, error_code(errors::no_i2p_router, get_libtorrent_category()), (char*)0, 0));
+			return;
+		}
+#endif
+
+		proxy_settings const* proxy = ps;
+#if TORRENT_USE_I2P
+		if (is_i2p) proxy = &i2p_conn->proxy();
+#endif
 
 		// in this case, the upper layer is assumed to have taken
 		// care of the proxying already. Don't instantiate the socket
 		// with this proxy
-		if (ps && (ps->type == proxy_settings::http
-			|| ps->type == proxy_settings::http_pw)
+		if (proxy && (proxy->type == proxy_settings::http
+			|| proxy->type == proxy_settings::http_pw)
 			&& !ssl)
 		{
-			ps = 0;
+			proxy = 0;
 		}
 		proxy_settings null_proxy;
 
+		void* userdata = 0;
 #ifdef TORRENT_USE_OPENSSL
 		if (m_ssl)
 		{
-			m_sock.instantiate<ssl_stream<socket_type> >(m_resolver.get_io_service());
-			ssl_stream<socket_type>* s = m_sock.get<ssl_stream<socket_type> >();
-			TORRENT_ASSERT(s);
-			bool ret = instantiate_connection(m_resolver.get_io_service()
-				, ps ? *ps : null_proxy, s->next_layer());
-			TORRENT_ASSERT(ret);
+			if (m_ssl_ctx == 0)
+			{
+				m_ssl_ctx = new (std::nothrow) boost::asio::ssl::context(
+					m_resolver.get_io_service(), asio::ssl::context::sslv23_client);
+				if (m_ssl_ctx)
+				{
+					m_own_ssl_context = true;
+					error_code ec;
+					m_ssl_ctx->set_verify_mode(asio::ssl::context::verify_none, ec);
+					TORRENT_ASSERT(!ec);
+				}
+			}
+			userdata = m_ssl_ctx;
 		}
-		else
-		{
-			m_sock.instantiate<socket_type>(m_resolver.get_io_service());
-			bool ret = instantiate_connection(m_resolver.get_io_service()
-				, ps ? *ps : null_proxy, *m_sock.get<socket_type>());
-			TORRENT_ASSERT(ret);
-		}
-#else
-		bool ret = instantiate_connection(m_resolver.get_io_service()
-			, ps ? *ps : null_proxy, m_sock);
-		TORRENT_ASSERT(ret);
 #endif
+		instantiate_connection(m_resolver.get_io_service()
+			, proxy ? *proxy : null_proxy, m_sock, userdata);
+
 		if (m_bind_addr != address_v4::any())
 		{
 			error_code ec;
@@ -226,10 +334,44 @@ void http_connection::start(std::string const& hostname, std::string const& port
 			}
 		}
 
-		m_endpoints.clear();
-		tcp::resolver::query query(hostname, port);
-		m_resolver.async_resolve(query, boost::bind(&http_connection::on_resolve
-			, me, _1, _2));
+		setup_ssl_hostname(m_sock, hostname, ec);
+		if (ec)
+		{
+			m_resolver.get_io_service().post(boost::bind(&http_connection::callback
+				, me, ec, (char*)0, 0));
+			return;
+		}
+
+#if TORRENT_USE_I2P
+		if (is_i2p)
+		{
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("http_connection::on_i2p_resolve");
+#endif
+			i2p_conn->async_name_lookup(hostname.c_str(), boost::bind(&http_connection::on_i2p_resolve
+				, me, _1, _2));
+		}
+		else
+#endif
+		if (ps && ps->proxy_hostnames
+			&& (ps->type == proxy_settings::socks5
+				|| ps->type == proxy_settings::socks5_pw))
+		{
+			m_hostname = hostname;
+			m_port = port;
+			m_endpoints.push_back(tcp::endpoint(address(), atoi(port.c_str())));
+			queue_connect();
+		}
+		else
+		{
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("http_connection::on_resolve");
+#endif
+			m_endpoints.clear();
+			tcp::resolver::query query(hostname, port);
+			m_resolver.async_resolve(query, boost::bind(&http_connection::on_resolve
+				, me, _1, _2));
+		}
 		m_hostname = hostname;
 		m_port = port;
 	}
@@ -237,40 +379,42 @@ void http_connection::start(std::string const& hostname, std::string const& port
 
 void http_connection::on_connect_timeout()
 {
-	if (m_connection_ticket > -1) m_cc.done(m_connection_ticket);
-	m_connection_ticket = -1;
+	TORRENT_ASSERT(m_connection_ticket > -1);
 
 	// keep ourselves alive even if the callback function
 	// deletes this object
 	boost::shared_ptr<http_connection> me(shared_from_this());
 
-	if (!m_endpoints.empty())
-	{
-		error_code ec;
-		m_sock.close(ec);
-	} 
-	else
-	{ 
-		callback(asio::error::timed_out);
-		close();
-	}
+	error_code ec;
+	m_sock.close(ec);
 }
 
 void http_connection::on_timeout(boost::weak_ptr<http_connection> p
 	, error_code const& e)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_timeout");
+#endif
 	boost::shared_ptr<http_connection> c = p.lock();
 	if (!c) return;
 
 	if (e == asio::error::operation_aborted) return;
 
-	if (c->m_last_receive + c->m_timeout < time_now_hires())
+	ptime now = time_now_hires();
+
+	if (c->m_start_time + c->m_completion_timeout < now
+		|| c->m_last_receive + c->m_read_timeout < now)
 	{
 		if (c->m_connection_ticket > -1 && !c->m_endpoints.empty())
 		{
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("http_connection::on_timeout");
+#endif
 			error_code ec;
-			c->m_sock.close(ec);
-			c->m_timer.expires_at(c->m_last_receive + c->m_timeout, ec);
+			async_shutdown(c->m_sock, c);
+			c->m_timer.expires_at((std::min)(
+				c->m_last_receive + c->m_read_timeout
+				, c->m_start_time + c->m_completion_timeout), ec);
 			c->m_timer.async_wait(boost::bind(&http_connection::on_timeout, p, _1));
 		}
 		else
@@ -282,27 +426,73 @@ void http_connection::on_timeout(boost::weak_ptr<http_connection> p
 	}
 
 	if (!c->m_sock.is_open()) return;
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_timeout");
+#endif
 	error_code ec;
-	c->m_timer.expires_at(c->m_last_receive + c->m_timeout, ec);
+	c->m_timer.expires_at((std::min)(
+		c->m_last_receive + c->m_read_timeout
+		, c->m_start_time + c->m_completion_timeout), ec);
 	c->m_timer.async_wait(boost::bind(&http_connection::on_timeout, p, _1));
 }
 
 void http_connection::close()
 {
+	if (m_abort) return;
+
 	error_code ec;
 	m_timer.cancel(ec);
 	m_resolver.cancel();
 	m_limiter_timer.cancel(ec);
-	m_sock.close(ec);
+
+	async_shutdown(m_sock, shared_from_this());
+
 	m_hostname.clear();
 	m_port.clear();
 	m_handler.clear();
 	m_abort = true;
 }
 
+#if TORRENT_USE_I2P
+void http_connection::on_i2p_resolve(error_code const& e
+	, char const* destination)
+{
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_i2p_resolve");
+#endif
+	if (e)
+	{
+		callback(e);
+		close();
+		return;
+	}
+
+#ifdef TORRENT_USE_OPENSSL
+	TORRENT_ASSERT(m_ssl == false);
+	TORRENT_ASSERT(m_sock.get<socket_type>());
+	TORRENT_ASSERT(m_sock.get<socket_type>()->get<i2p_stream>());
+	m_sock.get<socket_type>()->get<i2p_stream>()->set_destination(destination);
+	m_sock.get<socket_type>()->get<i2p_stream>()->set_command(i2p_stream::cmd_connect);
+	m_sock.get<socket_type>()->get<i2p_stream>()->set_session_id(m_i2p_conn->session_id());
+#else
+	m_sock.get<i2p_stream>()->set_destination(destination);
+	m_sock.get<i2p_stream>()->set_command(i2p_stream::cmd_connect);
+	m_sock.get<i2p_stream>()->set_session_id(m_i2p_conn->session_id());
+#endif
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_connect");
+#endif
+	m_sock.async_connect(tcp::endpoint(), boost::bind(&http_connection::on_connect
+		, shared_from_this(), _1));
+}
+#endif
+
 void http_connection::on_resolve(error_code const& e
 	, tcp::resolver::iterator i)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_resolve");
+#endif
 	if (e)
 	{
 		boost::shared_ptr<http_connection> me(shared_from_this());
@@ -347,18 +537,49 @@ void http_connection::queue_connect()
 
 	m_cc.enqueue(boost::bind(&http_connection::connect, shared_from_this(), _1, target)
 		, boost::bind(&http_connection::on_connect_timeout, shared_from_this())
-		, m_timeout, m_priority);
+		, m_read_timeout, m_priority);
 }
 
 void http_connection::connect(int ticket, tcp::endpoint target_address)
 {
+	if (ticket == -1)
+	{
+		close();
+		return;
+	}
+
 	m_connection_ticket = ticket;
+	if (m_proxy.proxy_hostnames
+		&& (m_proxy.type == proxy_settings::socks5
+			|| m_proxy.type == proxy_settings::socks5_pw))
+	{
+		// we're using a socks proxy and we're resolving
+		// hostnames through it
+#ifdef TORRENT_USE_OPENSSL
+		if (m_ssl)
+		{
+			TORRENT_ASSERT(m_sock.get<ssl_stream<socks5_stream> >());
+			m_sock.get<ssl_stream<socks5_stream> >()->next_layer().set_dst_name(m_hostname);
+		}
+		else
+#endif
+		{
+			TORRENT_ASSERT(m_sock.get<socks5_stream>());
+			m_sock.get<socks5_stream>()->set_dst_name(m_hostname);
+		}
+	}
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_connect");
+#endif
 	m_sock.async_connect(target_address, boost::bind(&http_connection::on_connect
 		, shared_from_this(), _1));
 }
 
 void http_connection::on_connect(error_code const& e)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_connect");
+#endif
 	if (m_connection_ticket >= 0)
 	{
 		m_cc.done(m_connection_ticket);
@@ -366,9 +587,13 @@ void http_connection::on_connect(error_code const& e)
 	}
 
 	m_last_receive = time_now_hires();
+	m_start_time = m_last_receive;
 	if (!e)
 	{ 
 		if (m_connect_handler) m_connect_handler(*this);
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("http_connection::on_write");
+#endif
 		async_write(m_sock, asio::buffer(sendbuffer)
 			, boost::bind(&http_connection::on_write, shared_from_this(), _1));
 	}
@@ -387,13 +612,15 @@ void http_connection::on_connect(error_code const& e)
 	}
 }
 
-void http_connection::callback(error_code const& e, char const* data, int size)
+void http_connection::callback(error_code e, char const* data, int size)
 {
 	if (m_bottled && m_called) return;
 
 	std::vector<char> buf;
 	if (data && m_bottled && m_parser.header_finished())
 	{
+		size = m_parser.collapse_chunk_headers((char*)data, size);
+
 		std::string const& encoding = m_parser.header("content-encoding");
 		if ((encoding == "gzip" || encoding == "x-gzip") && size > 0 && data)
 		{
@@ -407,6 +634,11 @@ void http_connection::callback(error_code const& e, char const* data, int size)
 			size = int(buf.size());
 			data = size == 0 ? 0 : &buf[0];
 		}
+
+		// if we completed the whole response, no need
+		// to tell the user that the connection was closed by
+		// the server or by us. Just clear any error
+		if (m_parser.finished()) e.clear();
 	}
 	m_called = true;
 	error_code ec;
@@ -416,6 +648,12 @@ void http_connection::callback(error_code const& e, char const* data, int size)
 
 void http_connection::on_write(error_code const& e)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_write");
+#endif
+
+	if (e == asio::error::operation_aborted) return;
+
 	if (e)
 	{
 		boost::shared_ptr<http_connection> me(shared_from_this());
@@ -434,10 +672,18 @@ void http_connection::on_write(error_code const& e)
 		if (m_download_quota == 0)
 		{
 			if (!m_limiter_timer_active)
+			{
+#if defined TORRENT_ASIO_DEBUGGING
+				add_outstanding_async("http_connection::on_assign_bandwidth");
+#endif
 				on_assign_bandwidth(error_code());
+			}
 			return;
 		}
 	}
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_read");
+#endif
 	m_sock.async_read_some(asio::buffer(&m_recvbuffer[0] + m_read_pos
 		, amount_to_read)
 		, boost::bind(&http_connection::on_read
@@ -447,12 +693,17 @@ void http_connection::on_write(error_code const& e)
 void http_connection::on_read(error_code const& e
 	, std::size_t bytes_transferred)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_read");
+#endif
 	if (m_rate_limit)
 	{
 		m_download_quota -= bytes_transferred;
 		TORRENT_ASSERT(m_download_quota >= 0);
 	}
 
+	if (e == asio::error::operation_aborted) return;
+
 	// keep ourselves alive even if the callback function
 	// deletes this object
 	boost::shared_ptr<http_connection> me(shared_from_this());
@@ -518,13 +769,22 @@ void http_connection::on_read(error_code const& e
 				}
 
 				error_code ec;
+				// it would be nice to gracefully shut down SSL here
+				// but then we'd have to do all the reconnect logic
+				// in its handler. For now, just kill the connection.
+//				async_shutdown(m_sock, shared_from_this());
 				m_sock.close(ec);
 				using boost::tuples::ignore;
 				boost::tie(ignore, ignore, ignore, ignore, ignore)
 					= parse_url_components(location, ec);
 				if (!ec)
 				{
-					get(location, m_timeout, m_priority, &m_proxy, m_redirects - 1);
+					get(location, m_completion_timeout, m_priority, &m_proxy, m_redirects - 1
+						, m_user_agent, m_bind_addr
+#if TORRENT_USE_I2P
+						, m_i2p_conn
+#endif
+						);
 				}
 				else
 				{
@@ -540,7 +800,12 @@ void http_connection::on_read(error_code const& e
 						url += '/';
 					url += location;
 
-					get(url, m_timeout, m_priority, &m_proxy, m_redirects - 1);
+					get(url, m_completion_timeout, m_priority, &m_proxy, m_redirects - 1
+						, m_user_agent, m_bind_addr
+#if TORRENT_USE_I2P
+						, m_i2p_conn
+#endif
+						);
 				}
 				return;
 			}
@@ -586,10 +851,18 @@ void http_connection::on_read(error_code const& e
 		if (m_download_quota == 0)
 		{
 			if (!m_limiter_timer_active)
+			{
+#if defined TORRENT_ASIO_DEBUGGING
+				add_outstanding_async("http_connection::on_assign_bandwidth");
+#endif
 				on_assign_bandwidth(error_code());
+			}
 			return;
 		}
 	}
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_read");
+#endif
 	m_sock.async_read_some(asio::buffer(&m_recvbuffer[0] + m_read_pos
 		, amount_to_read)
 		, boost::bind(&http_connection::on_read
@@ -598,6 +871,9 @@ void http_connection::on_read(error_code const& e
 
 void http_connection::on_assign_bandwidth(error_code const& e)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("http_connection::on_assign_bandwidth");
+#endif
 	if ((e == asio::error::operation_aborted
 		&& m_limiter_timer_active)
 		|| !m_sock.is_open())
@@ -618,6 +894,9 @@ void http_connection::on_assign_bandwidth(error_code const& e)
 
 	if (!m_sock.is_open()) return;
 
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_read");
+#endif
 	m_sock.async_read_some(asio::buffer(&m_recvbuffer[0] + m_read_pos
 		, amount_to_read)
 		, boost::bind(&http_connection::on_read
@@ -626,6 +905,9 @@ void http_connection::on_assign_bandwidth(error_code const& e)
 	error_code ec;
 	m_limiter_timer_active = true;
 	m_limiter_timer.expires_from_now(milliseconds(250), ec);
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("http_connection::on_assign_bandwidth");
+#endif
 	m_limiter_timer.async_wait(boost::bind(&http_connection::on_assign_bandwidth
 		, shared_from_this(), _1));
 }
@@ -639,6 +921,9 @@ void http_connection::rate_limit(int limit)
 		error_code ec;
 		m_limiter_timer_active = true;
 		m_limiter_timer.expires_from_now(milliseconds(250), ec);
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("http_connection::on_assign_bandwidth");
+#endif
 		m_limiter_timer.async_wait(boost::bind(&http_connection::on_assign_bandwidth
 			, shared_from_this(), _1));
 	}
