@@ -33,46 +33,18 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/pch.hpp"
 
 #include <algorithm>
-#include <iomanip>
 #include <ctime>
 
 #include "libtorrent/kademlia/node_id.hpp"
 #include "libtorrent/hasher.hpp"
 #include "libtorrent/assert.hpp"
-#include "libtorrent/broadcast_socket.hpp"
+#include "libtorrent/broadcast_socket.hpp" // for is_local et.al
+#include "libtorrent/socket_io.hpp" // for hash_address
+#include "libtorrent/random.hpp"
 
 namespace libtorrent { namespace dht
 {
 
-void hash_address(address const& ip, sha1_hash& h)
-{
-#if TORRENT_USE_IPV6
-	if (ip.is_v6())
-	{
-		address_v6::bytes_type b = ip.to_v6().to_bytes();
-		h = hasher((char*)&b[0], b.size()).final();
-	}
-	else
-#endif
-	{
-		address_v4::bytes_type b = ip.to_v4().to_bytes();
-		h = hasher((char*)&b[0], b.size()).final();
-	}
-}
-
-// verifies whether a node-id matches the IP it's used from
-// returns true if the node-id is OK coming from this source
-// and false otherwise.
-bool verify_id(node_id const& nid, address const& source_ip)
-{
-	// no need to verify local IPs, they would be incorrect anyway
-	if (is_local(source_ip)) return true;
-
-	node_id h;
-	hash_address(source_ip, h);
-	return memcmp(&nid[0], &h[0], 4) == 0;
-}
-	
 // returns the distance between the two nodes
 // using the kademlia XOR-metric
 node_id distance(node_id const& n1, node_id const& n2)
@@ -124,20 +96,72 @@ int distance_exp(node_id const& n1, node_id const& n2)
 	return 0;
 }
 
-struct static_ { static_() { std::srand(std::time(0)); } } static__;
-	
-node_id generate_id()
+struct static_ { static_() { std::srand((unsigned int)std::time(0)); } } static__;
+
+node_id generate_id_impl(address const& ip_, boost::uint32_t r)
 {
-	char random[20];
-#ifdef _MSC_VER
-	std::generate(random, random + 20, &rand);
-#else
-	std::generate(random, random + 20, &std::rand);
+	boost::uint8_t* ip = 0;
+	
+	const static boost::uint8_t v4mask[] = { 0x01, 0x07, 0x1f, 0x7f };
+	const static boost::uint8_t v6mask[] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f };
+	boost::uint8_t const* mask = 0;
+	int num_octets = 0;
+
+	address_v4::bytes_type b4;
+#if TORRENT_USE_IPV6
+	address_v6::bytes_type b6;
+	if (ip_.is_v6())
+	{
+		b6 = ip_.to_v6().to_bytes();
+		ip = &b6[0];
+		num_octets = 8;
+		mask = v6mask;
+	}
+	else
 #endif
+	{
+		b4 = ip_.to_v4().to_bytes();
+		ip = &b4[0];
+		num_octets = 4;
+		mask = v4mask;
+	}
+
+	for (int i = 0; i < num_octets; ++i)
+		ip[i] &= mask[i];
 
 	hasher h;
-	h.update(random, 20);
-	return h.final();
+	h.update((char*)ip, num_octets);
+	boost::uint8_t rand = r & 0x7;
+	h.update((char*)&rand, 1);
+	node_id id = h.final();
+	for (int i = 4; i < 19; ++i) id[i] = random();
+	id[19] = r;
+
+	return id;
+}
+
+node_id generate_random_id()
+{
+	char r[20];
+	for (int i = 0; i < 20; ++i) r[i] = random();
+	return hasher(r, 20).final();
+}
+
+// verifies whether a node-id matches the IP it's used from
+// returns true if the node-id is OK coming from this source
+// and false otherwise.
+bool verify_id(node_id const& nid, address const& source_ip)
+{
+	// no need to verify local IPs, they would be incorrect anyway
+	if (is_local(source_ip)) return true;
+
+	node_id h = generate_id_impl(source_ip, nid[19]);
+	return memcmp(&nid[0], &h[0], 4) == 0;
+}
+
+node_id generate_id(address const& ip)
+{
+	return generate_id_impl(ip, rand());
 }
 
 } }  // namespace libtorrent::dht
