@@ -38,6 +38,11 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/buffer.hpp"
 #include "libtorrent/http_parser.hpp"
 #include "libtorrent/escape_string.hpp"
+#include "libtorrent/socket_io.hpp" // for print_address
+
+#if defined TORRENT_ASIO_DEBUGGING
+#include "libtorrent/debug.hpp"
+#endif
 
 #include <boost/bind.hpp>
 #include <boost/ref.hpp>
@@ -48,7 +53,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <boost/asio/ip/host_name.hpp>
 #include <boost/asio/ip/multicast.hpp>
 #endif
-#include <boost/thread/mutex.hpp>
 #include <cstdlib>
 #include <boost/config.hpp>
 
@@ -78,7 +82,7 @@ lsd::lsd(io_service& ios, address const& listen_interface
 
 lsd::~lsd() {}
 
-void lsd::announce(sha1_hash const& ih, int listen_port)
+void lsd::announce(sha1_hash const& ih, int listen_port, bool broadcast)
 {
 	if (m_disabled) return;
 
@@ -94,7 +98,7 @@ void lsd::announce(sha1_hash const& ih, int listen_port)
 
 	m_retry_count = 1;
 	error_code ec;
-	m_socket.send(msg, msg_len, ec);
+	m_socket.send(msg, msg_len, ec, broadcast ? broadcast_socket::broadcast : 0);
 	if (ec)
 	{
 		m_disabled = true;
@@ -110,23 +114,31 @@ void lsd::announce(sha1_hash const& ih, int listen_port)
 	}
 #endif
 
-	m_broadcast_timer.expires_from_now(milliseconds(250 * m_retry_count), ec);
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("lsd::resend_announce");
+#endif
+	m_broadcast_timer.expires_from_now(seconds(2 * m_retry_count), ec);
 	m_broadcast_timer.async_wait(boost::bind(&lsd::resend_announce, self(), _1
 		, std::string(msg)));
 }
 
 void lsd::resend_announce(error_code const& e, std::string msg)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("lsd::resend_announce");
+#endif
 	if (e) return;
 
 	error_code ec;
 	m_socket.send(msg.c_str(), int(msg.size()), ec);
 
 	++m_retry_count;
-	if (m_retry_count >= 5)
-		return;
+	if (m_retry_count >= 3) return;
 
-	m_broadcast_timer.expires_from_now(milliseconds(250 * m_retry_count), ec);
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("lsd::resend_announce");
+#endif
+	m_broadcast_timer.expires_from_now(seconds(2 * m_retry_count), ec);
 	m_broadcast_timer.async_wait(boost::bind(&lsd::resend_announce, self(), _1, msg));
 }
 
@@ -163,43 +175,48 @@ void lsd::on_announce(udp::endpoint const& from, char* buffer
 	if (port_str.empty())
 	{
 #if defined(TORRENT_LOGGING) || defined(TORRENT_VERBOSE_LOGGING)
-	m_log << time_now_string()
-		<< " <== announce: invalid BT-SEARCH, missing port" << std::endl;
+		m_log << time_now_string()
+			<< " <== announce: invalid BT-SEARCH, missing port" << std::endl;
 #endif
 		return;
 	}
 
-	std::string const& ih_str = p.header("infohash");
-	if (ih_str.empty())
-	{
-#if defined(TORRENT_LOGGING) || defined(TORRENT_VERBOSE_LOGGING)
-	m_log << time_now_string()
-		<< " <== announce: invalid BT-SEARCH, missing infohash" << std::endl;
-#endif
-		return;
-	}
-
-	sha1_hash ih(0);
-	from_hex(ih_str.c_str(), 40, (char*)&ih[0]);
 	int port = std::atoi(port_str.c_str());
 
-	if (!ih.is_all_zeros() && port != 0)
+	typedef std::multimap<std::string, std::string> headers_t;
+	headers_t const& headers = p.headers();
+	std::pair<headers_t::const_iterator, headers_t::const_iterator> ihs
+		= headers.equal_range("infohash");
+
+	for (headers_t::const_iterator i = ihs.first; i != ihs.second; ++i)
 	{
+		std::string const& ih_str = i->second;
+		if (ih_str.size() != 40)
+		{
 #if defined(TORRENT_LOGGING) || defined(TORRENT_VERBOSE_LOGGING)
-		char msg[200];
-		snprintf(msg, 200, "%s *** incoming local announce %s:%d ih: %s\n"
-			, time_now_string(), print_address(from.address()).c_str()
-			, port, ih_str.c_str());
+			m_log << time_now_string()
+				<< " <== announce: invalid BT-SEARCH, invalid infohash: "
+				<< ih_str << std::endl;
 #endif
-		// we got an announce, pass it on through the callback
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
-			m_callback(tcp::endpoint(from.address(), port), ih);
-#ifndef BOOST_NO_EXCEPTIONS
+			continue;
 		}
-		catch (std::exception&) {}
+
+		sha1_hash ih(0);
+		from_hex(ih_str.c_str(), 40, (char*)&ih[0]);
+
+		if (!ih.is_all_zeros() && port != 0)
+		{
+#if defined(TORRENT_LOGGING) || defined(TORRENT_VERBOSE_LOGGING)
+			char msg[200];
+			snprintf(msg, 200, "%s *** incoming local announce %s:%d ih: %s\n"
+				, time_now_string(), print_address(from.address()).c_str()
+				, port, ih_str.c_str());
 #endif
+			// we got an announce, pass it on through the callback
+			TORRENT_TRY {
+				m_callback(tcp::endpoint(from.address(), port), ih);
+			} TORRENT_CATCH(std::exception&) {}
+		}
 	}
 }
 
@@ -212,8 +229,3 @@ void lsd::close()
 	m_callback.clear();
 }
 
-void lsd::use_broadcast(bool b)
-{
-	m_socket.enable_ip_broadcast(b);
-}
-
