@@ -34,6 +34,7 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/session.hpp"
 #include "libtorrent/torrent_handle.hpp"
 #include "libtorrent/escape_string.hpp"
+#include "libtorrent/error_code.hpp"
 
 #include <string>
 
@@ -95,10 +96,10 @@ namespace libtorrent
 		return ret;
 	}
 
-#ifndef BOOST_NO_EXCEPTIONS
 #ifndef TORRENT_NO_DEPRECATE
+#ifndef BOOST_NO_EXCEPTIONS
 	torrent_handle add_magnet_uri(session& ses, std::string const& uri
-		, fs::path const& save_path
+		, std::string const& save_path
 		, storage_mode_t storage_mode
 		, bool paused
 		, storage_constructor_type sc
@@ -108,25 +109,24 @@ namespace libtorrent
 		std::string tracker;
 
 		error_code ec;
-		boost::optional<std::string> display_name = url_has_argument(uri, "dn");
-		if (display_name) name = unescape_string(display_name->c_str(), ec);
-		boost::optional<std::string> tracker_string = url_has_argument(uri, "tr");
-		if (tracker_string) tracker = unescape_string(tracker_string->c_str(), ec);
+		std::string display_name = url_has_argument(uri, "dn");
+		if (!display_name.empty()) name = unescape_string(display_name.c_str(), ec);
+		std::string tracker_string = url_has_argument(uri, "tr");
+		if (!tracker_string.empty()) tracker = unescape_string(tracker_string.c_str(), ec);
 	
-		boost::optional<std::string> btih = url_has_argument(uri, "xt");
-		if (!btih) return torrent_handle();
+		std::string btih = url_has_argument(uri, "xt");
+		if (btih.empty()) return torrent_handle();
 
-		if (btih->compare(0, 9, "urn:btih:") != 0) return torrent_handle();
+		if (btih.compare(0, 9, "urn:btih:") != 0) return torrent_handle();
 
 		sha1_hash info_hash;
-		if (btih->size() == 40 + 9) from_hex(&(*btih)[9], 40, (char*)&info_hash[0]);
-		else info_hash.assign(base32decode(btih->substr(9)));
+		if (btih.size() == 40 + 9) from_hex(&btih[9], 40, (char*)&info_hash[0]);
+		else info_hash.assign(base32decode(btih.substr(9)));
 
 		return ses.add_torrent(tracker.empty() ? 0 : tracker.c_str(), info_hash
 			, name.empty() ? 0 : name.c_str(), save_path, entry()
 			, storage_mode, paused, sc, userdata);
 	}
-#endif
 
 	torrent_handle add_magnet_uri(session& ses, std::string const& uri
 		, add_torrent_params p)
@@ -136,38 +136,58 @@ namespace libtorrent
 		if (ec) throw libtorrent_exception(ec);
 		return ret;
 	}
-#endif
+#endif // BOOST_NO_EXCEPTIONS
+
 	torrent_handle add_magnet_uri(session& ses, std::string const& uri
 		, add_torrent_params p, error_code& ec)
+	{
+		parse_magnet_uri(uri, p, ec);
+		if (ec) return torrent_handle();
+		return ses.add_torrent(p, ec);
+	}
+
+#endif // TORRENT_NO_DEPRECATE
+
+	void parse_magnet_uri(std::string const& uri, add_torrent_params& p, error_code& ec)
 	{
 		std::string name;
 		std::string tracker;
 
 		error_code e;
-		boost::optional<std::string> display_name = url_has_argument(uri, "dn");
-		if (display_name) name = unescape_string(display_name->c_str(), e);
-		size_t pos = std::string::npos;
-		boost::optional<std::string> tracker_string = url_has_argument(uri, "tr", &pos);
-		if (tracker_string) tracker = unescape_string(tracker_string->c_str(), e);
+		std::string display_name = url_has_argument(uri, "dn");
+		if (!display_name.empty()) name = unescape_string(display_name.c_str(), e);
+
+		// parse trackers out of the magnet link
+		std::string::size_type pos = std::string::npos;
+		std::string url = url_has_argument(uri, "tr", &pos);
+		while (pos != std::string::npos)
+		{
+			error_code e;
+			url = unescape_string(url, e);
+			if (e) continue;
+			p.trackers.push_back(url);
+			pos = uri.find("&tr=", pos);
+			if (pos == std::string::npos) break;
+			pos += 4;
+			url = uri.substr(pos, uri.find('&', pos) - pos);
+		}
 	
-		boost::optional<std::string> btih = url_has_argument(uri, "xt");
-		if (!btih)
+		std::string btih = url_has_argument(uri, "xt");
+		if (btih.empty())
 		{
 			ec = errors::missing_info_hash_in_uri;
-			return torrent_handle();
+			return;
 		}
 
-		if (btih->compare(0, 9, "urn:btih:") != 0)
+		if (btih.compare(0, 9, "urn:btih:") != 0)
 		{
 			ec = errors::missing_info_hash_in_uri;
-			return torrent_handle();
+			return;
 		}
 
 #ifndef TORRENT_DISABLE_DHT
 		std::string::size_type node_pos = std::string::npos;
-		boost::optional<std::string> node_opt = url_has_argument(uri, "dht", &node_pos);
-		std::string node;
-		if (node_opt) node = *node_opt;
+		std::string node = url_has_argument(uri, "dht", &node_pos);
 		while (!node.empty())
 		{
 			std::string::size_type divider = node.find_last_of(':');
@@ -175,7 +195,7 @@ namespace libtorrent
 			{
 				int port = atoi(node.c_str()+divider+1);
 				if (port != 0)
-					ses.add_dht_node(std::make_pair(node.substr(0, divider), port));
+					p.dht_nodes.push_back(std::make_pair(node.substr(0, divider), port));
 			}
 			
 			node_pos = uri.find("&dht=", node_pos);
@@ -186,29 +206,11 @@ namespace libtorrent
 #endif
 
 		sha1_hash info_hash;
-		if (btih->size() == 40 + 9) from_hex(&(*btih)[9], 40, (char*)&info_hash[0]);
-		else info_hash.assign(base32decode(btih->substr(9)));
+		if (btih.size() == 40 + 9) from_hex(&btih[9], 40, (char*)&info_hash[0]);
+		else info_hash.assign(base32decode(btih.substr(9)));
 
-		if (!tracker.empty()) p.tracker_url = tracker.c_str();
 		p.info_hash = info_hash;
-		if (!name.empty()) p.name = name.c_str();
-		torrent_handle ret = ses.add_torrent(p, ec);
-
-		int tier = 1;
-		// there might be more trackers in the url
-		while (pos != std::string::npos)
-		{
-			pos = uri.find("&tr=", pos);
-			if (pos == std::string::npos) break;
-			pos += 4;
-			error_code ec;
-			std::string url = unescape_string(uri.substr(pos, uri.find('&', pos) - pos), ec);
-			if (ec) continue;
-			announce_entry ae(url);
-			ae.tier = tier++;
-			ret.add_tracker(ae);
-		}
-		return ret;
+		if (!name.empty()) p.name = name;
 	}
 }
 
