@@ -54,16 +54,14 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/aux_/session_impl.hpp"
 #include "libtorrent/piece_picker.hpp"
 #include "libtorrent/broadcast_socket.hpp"
+#include "libtorrent/peer_info.hpp"
+#include "libtorrent/random.hpp"
+#include "libtorrent/extensions.hpp"
 
 #ifdef TORRENT_DEBUG
 #include "libtorrent/bt_peer_connection.hpp"
 #endif
 
-namespace libtorrent
-{
-	class peer_connection;
-}
-
 namespace
 {
 	using namespace libtorrent;
@@ -75,18 +73,24 @@ namespace
 		{}
 
 		bool operator()(policy::peer const* p) const
-		{ return p->address() == m_ep.address() && p->port == m_ep.port(); }
+		{
+			TORRENT_ASSERT(p->in_use);
+			return p->address() == m_ep.address() && p->port == m_ep.port();
+		}
 
 		tcp::endpoint const& m_ep;
 	};
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 	struct match_peer_connection
 	{
 		match_peer_connection(peer_connection const& c) : m_conn(c) {}
 
 		bool operator()(policy::peer const* p) const
-		{ return p->connection == &m_conn; }
+		{
+			TORRENT_ASSERT(p->in_use);
+			return p->connection == &m_conn;
+		}
 
 		peer_connection const& m_conn;
 	};
@@ -97,6 +101,7 @@ namespace
 
 		bool operator()(policy::peer const* p) const
 		{
+			TORRENT_ASSERT(p->in_use);
 			return p->connection == &m_conn
 				|| (p->ip() == m_conn.remote()
 					&& p->connectable);
@@ -112,7 +117,7 @@ namespace libtorrent
 {
 	// returns the rank of a peer's source. We have an affinity
 	// to connecting to peers with higher rank. This is to avoid
-	// problems when out peer list is diluted by stale peers from
+	// problems when our peer list is diluted by stale peers from
 	// the resume data for instance
 	int source_rank(int source_bitmask)
 	{
@@ -143,14 +148,13 @@ namespace libtorrent
 		if (!t.are_files_checked()) return;
 
 		TORRENT_ASSERT(t.valid_metadata());
-		TORRENT_ASSERT(c.peer_info_struct() != 0 || !dynamic_cast<bt_peer_connection*>(&c));
+		TORRENT_ASSERT(c.peer_info_struct() != 0 || c.type() != peer_connection::bittorrent_connection);
 		int num_requests = c.desired_queue_size()
 			- (int)c.download_queue().size()
 			- (int)c.request_queue().size();
 
 #ifdef TORRENT_VERBOSE_LOGGING
-		(*c.m_logger) << time_now_string() << " PIECE_PICKER [ req: " << num_requests
-			<< " endgame: " << c.endgame() << "]\n";
+		c.peer_log("*** PIECE_PICKER [ req: %d engame: %d ]", num_requests, c.endgame());
 #endif
 		TORRENT_ASSERT(c.desired_queue_size() > 0);
 		// if our request queue is already full, we
@@ -219,11 +223,11 @@ namespace libtorrent
 		// then use this mode.
 		p.pick_pieces(*bits, interesting_pieces
 			, num_requests, prefer_whole_pieces, c.peer_info_struct()
-			, state, c.picker_options(), suggested);
+			, state, c.picker_options(), suggested, t.num_peers());
 
 #ifdef TORRENT_VERBOSE_LOGGING
-		(*c.m_logger) << time_now_string() << " PIECE_PICKER [ php: " << prefer_whole_pieces
-			<< " picked: " << interesting_pieces.size() << " ]\n";
+		c.peer_log("*** PIECE_PICKER [ prefer_whole: %d picked: %d ]"
+			, prefer_whole_pieces, int(interesting_pieces.size()));
 #endif
 
 		// if the number of pieces we have + the number of pieces
@@ -233,8 +237,7 @@ namespace libtorrent
 		// also, if we already have at least one outstanding
 		// request, we shouldn't pick any busy pieces either
 		bool dont_pick_busy_blocks = (ses.m_settings.strict_end_game_mode
-			&& p.num_have() + int(p.get_download_queue().size())
-				< t.torrent_file().num_pieces())
+			&& p.num_downloading_pieces() < p.num_want_left())
 			|| dq.size() + rq.size() > 0;
 
 		// this is filled with an interesting piece
@@ -244,6 +247,10 @@ namespace libtorrent
 		for (std::vector<piece_block>::iterator i = interesting_pieces.begin();
 			i != interesting_pieces.end(); ++i)
 		{
+#ifdef TORRENT_STATS
+			++ses.m_piece_picker_blocks;
+#endif
+
 			if (prefer_whole_pieces == 0 && num_requests <= 0) break;
 
 			int num_block_requests = p.num_peers(*i);
@@ -304,7 +311,10 @@ namespace libtorrent
 		// there aren't enough. This means we're in end-game mode
 		// as long as we have at least one request outstanding,
 		// we shouldn't pick another piece
-		c.set_endgame(true);
+		// if we are attempting to download 'allowed' pieces
+		// and can't find any, that doesn't count as end-game
+		if (!c.has_peer_choked())
+			c.set_endgame(true);
 	
 		// if we don't have any potential busy blocks to request
 		// or if we already have outstanding requests, don't
@@ -315,6 +325,10 @@ namespace libtorrent
 			return;
 		}
 
+#ifdef TORRENT_STATS
+		++ses.m_end_game_piece_picker_blocks;
+#endif
+
 #ifdef TORRENT_DEBUG
 		piece_picker::downloading_piece st;
 		p.piece_info(busy_block.piece_index, st);
@@ -330,8 +344,8 @@ namespace libtorrent
 	}
 
 	policy::policy(torrent* t)
-		: m_round_robin(0)
-		, m_torrent(t)
+		: m_torrent(t)
+		, m_round_robin(0)
 		, m_num_connect_candidates(0)
 		, m_num_seeds(0)
 		, m_finished(false)
@@ -343,8 +357,7 @@ namespace libtorrent
 		INVARIANT_CHECK;
 
 		aux::session_impl& ses = m_torrent->session();
-		piece_picker* p = 0;
-		if (m_torrent->has_picker()) p = &m_torrent->picker();
+		if (!m_torrent->apply_ip_filter()) return;
 
 		for (iterator i = m_peers.begin(); i != m_peers.end();)
 		{
@@ -354,20 +367,32 @@ namespace libtorrent
 				continue;
 			}
 		
+			if (ses.m_alerts.should_post<peer_blocked_alert>())
+				ses.m_alerts.post_alert(peer_blocked_alert(m_torrent->get_handle(), (*i)->address()));
+
+			int current = i - m_peers.begin();
+			TORRENT_ASSERT(current >= 0);
+			TORRENT_ASSERT(m_peers.size() > 0);
+			TORRENT_ASSERT(i != m_peers.end());
+
 			if ((*i)->connection)
 			{
-				(*i)->connection->disconnect(errors::banned_by_ip_filter);
-				if (ses.m_alerts.should_post<peer_blocked_alert>())
-					ses.m_alerts.post_alert(peer_blocked_alert((*i)->address()));
+				// disconnecting the peer here may also delete the
+				// peer_info_struct. If that is the case, just continue
+				int count = m_peers.size();
+				peer_connection* p = (*i)->connection;
+				
+				p->disconnect(errors::banned_by_ip_filter);
+				// what *i refers to has changed, i.e. cur was deleted
+				if (m_peers.size() < count)
+				{
+					i = m_peers.begin() + current;
+					continue;
+				}
 				TORRENT_ASSERT((*i)->connection == 0
 					|| (*i)->connection->peer_info_struct() == 0);
 			}
-			else
-			{
-				if (ses.m_alerts.should_post<peer_blocked_alert>())
-					ses.m_alerts.post_alert(peer_blocked_alert((*i)->address()));
-			}
-			int current = i - m_peers.begin();
+
 			erase_peer(i);
 			i = m_peers.begin() + current;
 		}
@@ -377,6 +402,8 @@ namespace libtorrent
 	{
 		INVARIANT_CHECK;
 
+		TORRENT_ASSERT(p->in_use);
+
 		std::pair<iterator, iterator> range = find_peers(p->address());
 		iterator iter = std::find_if(range.first, range.second, match_peer_endpoint(p->ip()));
 		if (iter == range.second) return;
@@ -404,7 +431,7 @@ namespace libtorrent
 		if (m_round_robin > i - m_peers.begin()) --m_round_robin;
 		if (m_round_robin >= int(m_peers.size())) m_round_robin = 0;
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 		TORRENT_ASSERT((*i)->in_use);
 		(*i)->in_use = false;
 #endif
@@ -418,6 +445,16 @@ namespace libtorrent
 				static_cast<ipv6_peer*>(*i));
 		}
 		else
+#endif
+#if TORRENT_USE_I2P
+		if ((*i)->is_i2p_addr)
+		{
+			TORRENT_ASSERT(m_torrent->session().m_i2p_peer_pool.is_from(
+				static_cast<i2p_peer*>(*i)));
+			m_torrent->session().m_i2p_peer_pool.destroy(
+				static_cast<i2p_peer*>(*i));
+		}
+		else
 #endif
 		{
 			TORRENT_ASSERT(m_torrent->session().m_ipv4_peer_pool.is_from(
@@ -425,26 +462,32 @@ namespace libtorrent
 			m_torrent->session().m_ipv4_peer_pool.destroy(
 				static_cast<ipv4_peer*>(*i));
 		}
-
 		m_peers.erase(i);
 	}
 
 	bool policy::should_erase_immediately(peer const& p) const
 	{
-		return p.source == peer_info::resume_data
-			&& p.failcount > 0
-			&& !p.banned;
+		TORRENT_ASSERT(p.in_use);
+		return p.source == peer_info::resume_data;
 	}
 
 	bool policy::is_erase_candidate(peer const& pe, bool finished) const
 	{
-		return pe.connection == 0
-			&& pe.last_connected != 0
-			&& !pe.banned
-			&& !is_connect_candidate(pe, m_finished);
+		TORRENT_ASSERT(pe.in_use);
+		if (pe.connection) return false;
+		if (is_connect_candidate(pe, finished)) return false;
+
+		return (pe.failcount > 0)
+			|| (pe.source == peer_info::resume_data);
 	}
 
-	void policy::erase_peers()
+	bool policy::is_force_erase_candidate(peer const& pe) const
+	{
+		TORRENT_ASSERT(pe.in_use);
+		return pe.connection == 0;
+	}
+
+	void policy::erase_peers(int flags)
 	{
 		INVARIANT_CHECK;
 
@@ -455,39 +498,49 @@ namespace libtorrent
 		if (max_peerlist_size == 0 || m_peers.empty()) return;
 
 		int erase_candidate = -1;
+		int force_erase_candidate = -1;
 
 		TORRENT_ASSERT(m_finished == m_torrent->is_finished());
 
-		int round_robin = rand() % m_peers.size();
+		int round_robin = random() % m_peers.size();
+
+		int low_watermark = max_peerlist_size * 95 / 100;
+		if (low_watermark == max_peerlist_size) --low_watermark;
 
 		for (int iterations = (std::min)(int(m_peers.size()), 300);
 			iterations > 0; --iterations)
 		{
-			if (m_peers.size() < max_peerlist_size * 0.95)
+			if (int(m_peers.size()) < low_watermark)
 				break;
 
 			if (round_robin == int(m_peers.size())) round_robin = 0;
 
 			peer& pe = *m_peers[round_robin];
+			TORRENT_ASSERT(pe.in_use);
 			int current = round_robin;
 
+			if (is_erase_candidate(pe, m_finished)
+				&& (erase_candidate == -1
+					|| !compare_peer_erase(*m_peers[erase_candidate], pe)))
 			{
-				if (is_erase_candidate(pe, m_finished)
-					&& (erase_candidate == -1
-						|| !compare_peer_erase(*m_peers[erase_candidate], pe)))
+				if (should_erase_immediately(pe))
 				{
-					if (should_erase_immediately(pe))
-					{
-						if (erase_candidate > current) --erase_candidate;
-						TORRENT_ASSERT(current >= 0 && current < int(m_peers.size()));
-						--round_robin;
-						erase_peer(m_peers.begin() + current);
-					}
-					else
-					{
-						erase_candidate = current;
-					}
+					if (erase_candidate > current) --erase_candidate;
+					if (force_erase_candidate > current) --force_erase_candidate;
+					TORRENT_ASSERT(current >= 0 && current < int(m_peers.size()));
+					erase_peer(m_peers.begin() + current);
+					continue;
 				}
+				else
+				{
+					erase_candidate = current;
+				}
+			}
+			if (is_force_erase_candidate(pe)
+				&& (force_erase_candidate == -1
+					|| !compare_peer_erase(*m_peers[force_erase_candidate], pe)))
+			{
+				force_erase_candidate = current;
 			}
 
 			++round_robin;
@@ -498,15 +551,30 @@ namespace libtorrent
 			TORRENT_ASSERT(erase_candidate >= 0 && erase_candidate < int(m_peers.size()));
 			erase_peer(m_peers.begin() + erase_candidate);
 		}
+		else if ((flags & force_erase) && force_erase_candidate > -1)
+		{
+			TORRENT_ASSERT(force_erase_candidate >= 0 && force_erase_candidate < int(m_peers.size()));
+			erase_peer(m_peers.begin() + force_erase_candidate);
+		}
 	}
 
 	void policy::ban_peer(policy::peer* p)
 	{
 		INVARIANT_CHECK;
 
+		TORRENT_ASSERT(p->in_use);
+
+		if (!m_torrent->settings().ban_web_seeds && p->web_seed)
+			return;
+
 		if (is_connect_candidate(*p, m_finished))
 			--m_num_connect_candidates;
 
+#ifdef TORRENT_STATS
+		aux::session_impl& ses = m_torrent->session();
+		++ses.m_num_banned_peers;
+#endif
+
 		p->banned = true;
 		TORRENT_ASSERT(!is_connect_candidate(*p, m_finished));
 	}
@@ -515,6 +583,7 @@ namespace libtorrent
 	{
 		INVARIANT_CHECK;
 
+		TORRENT_ASSERT(p->in_use);
 		TORRENT_ASSERT(c);
 
 		const bool was_conn_cand = is_connect_candidate(*p, m_finished);
@@ -526,6 +595,7 @@ namespace libtorrent
 	{
 		INVARIANT_CHECK;
 
+		TORRENT_ASSERT(p->in_use);
 		const bool was_conn_cand = is_connect_candidate(*p, m_finished);
 		p->failcount = f;
 		if (was_conn_cand != is_connect_candidate(*p, m_finished))
@@ -537,8 +607,10 @@ namespace libtorrent
 
 	bool policy::is_connect_candidate(peer const& p, bool finished) const
 	{
+		TORRENT_ASSERT(p.in_use);
 		if (p.connection
 			|| p.banned
+			|| p.web_seed
 			|| !p.connectable
 			|| (p.seed && finished)
 			|| int(p.failcount) >= m_torrent->settings().max_failcount)
@@ -548,7 +620,11 @@ namespace libtorrent
 		if (ses.m_port_filter.access(p.port) & port_filter::blocked)
 			return false;
 
-		if (p.port < 1024 && p.source == peer_info::dht)
+		// only apply this to peers we've only heard
+		// about from the DHT
+		if (ses.m_settings.no_connect_privileged_ports
+			&& p.port < 1024
+			&& p.source == peer_info::dht)
 			return false;
 
 		return true;
@@ -572,7 +648,7 @@ namespace libtorrent
 			// set external_ip to a random value, to
 			// radomize which peers we prefer
 			address_v4::bytes_type bytes;
-			std::generate(bytes.begin(), bytes.end(), &std::rand);
+			std::generate(bytes.begin(), bytes.end(), &random);
 			external_ip = address_v4(bytes);
 		}
 
@@ -592,6 +668,7 @@ namespace libtorrent
 			if (m_round_robin >= int(m_peers.size())) m_round_robin = 0;
 
 			peer& pe = *m_peers[m_round_robin];
+			TORRENT_ASSERT(pe.in_use);
 			int current = m_round_robin;
 
 #ifndef TORRENT_DISABLE_DHT
@@ -610,7 +687,7 @@ namespace libtorrent
 			// if the number of peers is growing large
 			// we need to start weeding.
 
-			if (m_peers.size() >= max_peerlist_size * 0.95
+			if (int(m_peers.size()) >= max_peerlist_size * 0.95
 				&& max_peerlist_size > 0)
 			{
 				if (is_erase_candidate(pe, m_finished)
@@ -621,8 +698,8 @@ namespace libtorrent
 					{
 						if (erase_candidate > current) --erase_candidate;
 						if (candidate > current) --candidate;
-						--m_round_robin;
 						erase_peer(m_peers.begin() + current);
+						continue;
 					}
 					else
 					{
@@ -673,16 +750,9 @@ namespace libtorrent
 		return m_peers.begin() + candidate;
 	}
 
-	void policy::pulse()
-	{
-		INVARIANT_CHECK;
-
-		erase_peers();
-	}
-
 	bool policy::new_connection(peer_connection& c, int session_time)
 	{
-		TORRENT_ASSERT(!c.is_local());
+		TORRENT_ASSERT(!c.is_outgoing());
 
 		INVARIANT_CHECK;
 
@@ -699,9 +769,21 @@ namespace libtorrent
 		aux::session_impl& ses = m_torrent->session();
 		
 		if (m_torrent->num_peers() >= m_torrent->max_connections()
-			&& ses.num_connections() >= ses.max_connections()
+			&& ses.num_connections() >= ses.settings().connections_limit
 			&& c.remote().address() != m_torrent->current_tracker().address())
 		{
+#if defined TORRENT_LOGGING || defined TORRENT_VERBOSE_LOGGING
+			(*m_torrent->session().m_logger) << time_now_string()
+				<< " *** TOO MANY CONNECTIONS ["
+				" torrent: " << m_torrent->name() <<
+				" torrent peers: " << m_torrent->num_peers() <<
+				" torrent limit: " << m_torrent->max_connections() <<
+				" global peers: " << ses.num_connections() <<
+				" global limit: " << ses.settings().connections_limit <<
+				" global list peers " << int(m_peers.size()) <<
+				" global list limit: " << m_torrent->settings().max_peerlist_size <<
+				" ]\n";
+#endif
 			c.disconnect(errors::too_many_connections);
 			return false;
 		}
@@ -735,10 +817,19 @@ namespace libtorrent
 			if (iter != m_peers.end() && (*iter)->address() == c.remote().address()) found = true;
 		}
 
+		// make sure the iterator we got is properly sorted relative
+		// to the connection's address
+//		TORRENT_ASSERT(m_peers.empty()
+//			|| (iter == m_peers.end() && (*(iter-1))->address() < c.remote().address())
+//			|| (iter != m_peers.end() && c.remote().address() < (*iter)->address())
+//			|| (iter != m_peers.end() && iter != m_peers.begin() && (*(iter-1))->address() < c.remote().address()));
+
 		if (found)
 		{
 			i = *iter;
+			TORRENT_ASSERT(i->connection != &c);
 
+			TORRENT_ASSERT(i->in_use);
 			if (i->banned)
 			{
 				c.disconnect(errors::peer_banned);
@@ -780,7 +871,7 @@ namespace libtorrent
 					i->connection->disconnect(ec2);
 					TORRENT_ASSERT(i->connection == 0);
 				}
-				else if (!i->connection->is_connecting() || c.is_local())
+				else if (!i->connection->is_connecting() || c.is_outgoing())
 				{
 					c.disconnect(errors::duplicate_peer_id);
 					return false;
@@ -813,8 +904,30 @@ namespace libtorrent
 
 			if (int(m_peers.size()) >= m_torrent->settings().max_peerlist_size)
 			{
-				c.disconnect(errors::too_many_connections);
-				return false;
+				// this may invalidate our iterator!
+				erase_peers(force_erase);
+				if (int(m_peers.size()) >= m_torrent->settings().max_peerlist_size)
+				{
+#if defined TORRENT_LOGGING || defined TORRENT_VERBOSE_LOGGING
+					(*m_torrent->session().m_logger) << time_now_string()
+						<< " *** TOO MANY CONNECTIONS ["
+						" torrent: " << m_torrent->name() <<
+						" torrent peers: " << m_torrent->num_peers() <<
+						" torrent limit: " << m_torrent->max_connections() <<
+						" global peers: " << ses.num_connections() <<
+						" global limit: " << ses.settings().connections_limit <<
+						" global list peers " << int(m_peers.size()) <<
+						" global list limit: " << m_torrent->settings().max_peerlist_size <<
+						" ]\n";
+#endif
+					c.disconnect(errors::too_many_connections);
+					return false;
+				}
+				// restore it
+				iter = std::lower_bound(
+					m_peers.begin(), m_peers.end()
+					, c.remote().address(), peer_address_compare()
+				);
 			}
 
 #if TORRENT_USE_IPV6
@@ -827,8 +940,6 @@ namespace libtorrent
 				(peer*)m_torrent->session().m_ipv4_peer_pool.malloc();
 			if (p == 0) return false;
 
-//			TORRENT_ASSERT(p->in_use == false);
-
 #if TORRENT_USE_IPV6
 			if (is_v6)
 				m_torrent->session().m_ipv6_peer_pool.set_next_size(500);
@@ -843,7 +954,7 @@ namespace libtorrent
 #endif
 				new (p) ipv4_peer(c.remote(), false, 0);
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 			p->in_use = true;
 #endif
 
@@ -865,7 +976,7 @@ namespace libtorrent
 		TORRENT_ASSERT(i);
 		c.set_peer_info(i);
 		TORRENT_ASSERT(i->connection == 0);
-		c.add_stat(i->prev_amount_download, i->prev_amount_upload);
+		c.add_stat(size_type(i->prev_amount_download) << 10, size_type(i->prev_amount_upload) << 10);
 
 		// restore transfer rate limits
 		int rate_limit;
@@ -884,6 +995,7 @@ namespace libtorrent
 		// this cannot be a connect candidate anymore, since i->connection is set
 		TORRENT_ASSERT(!is_connect_candidate(*i, m_finished));
 		TORRENT_ASSERT(has_connection(&c));
+		m_torrent->state_updated();
 		return true;
 	}
 
@@ -891,6 +1003,7 @@ namespace libtorrent
 	{
 		TORRENT_ASSERT(p != 0);
 		TORRENT_ASSERT(p->connection);
+		TORRENT_ASSERT(p->in_use);
 
 		INVARIANT_CHECK;
 
@@ -905,6 +1018,7 @@ namespace libtorrent
 			if (i != range.second)
 			{
 				policy::peer& pp = **i;
+				TORRENT_ASSERT(pp.in_use);
 				if (pp.connection)
 				{
 					bool was_conn_cand = is_connect_candidate(pp, m_finished);
@@ -914,6 +1028,14 @@ namespace libtorrent
 					pp.source |= src;
 					if (!was_conn_cand && is_connect_candidate(pp, m_finished))
 						++m_num_connect_candidates;
+					// calling disconnect() on a peer, may actually end
+					// up "garbage collecting" its policy::peer entry
+					// as well, if it's considered useless (which this specific)
+					// case will, since it was an incoming peer that just disconnected
+					// and we allow multiple connections per IP. Because of that,
+					// we need to make sure we don't let it do that, by unlinking
+					// the peer_connection from the policy::peer first.
+					p->connection->set_peer_info(0);
 					p->connection->disconnect(errors::duplicate_peer_id);
 					erase_peer(p);
 					return false;
@@ -948,6 +1070,7 @@ namespace libtorrent
 	// pointer. see smart_ban.cpp
 	bool policy::has_peer(policy::peer const* p) const
 	{
+		TORRENT_ASSERT(p->in_use);
 		// find p in m_peers
 		for (const_iterator i = m_peers.begin()
 			, end(m_peers.end()); i != end; ++i)
@@ -960,6 +1083,7 @@ namespace libtorrent
 	void policy::set_seed(policy::peer* p, bool s)
 	{
 		if (p == 0) return;
+		TORRENT_ASSERT(p->in_use);
 		if (p->seed == s) return;
 		bool was_conn_cand = is_connect_candidate(*p, m_finished);
 		p->seed = s;
@@ -970,12 +1094,186 @@ namespace libtorrent
 			if (m_num_connect_candidates < 0) m_num_connect_candidates = 0;
 		}
 
+		if (p->web_seed) return;
 		if (s) ++m_num_seeds;
 		else --m_num_seeds;
 		TORRENT_ASSERT(m_num_seeds >= 0);
 		TORRENT_ASSERT(m_num_seeds <= int(m_peers.size()));
 	}
 
+	bool policy::insert_peer(policy::peer* p, iterator iter, int flags)
+	{
+		TORRENT_ASSERT(p);
+		TORRENT_ASSERT(p->in_use);
+
+		int max_peerlist_size = m_torrent->is_paused()
+			?m_torrent->settings().max_paused_peerlist_size
+			:m_torrent->settings().max_peerlist_size;
+
+		if (max_peerlist_size
+			&& int(m_peers.size()) >= max_peerlist_size)
+		{
+			if (p->source == peer_info::resume_data) return false;
+
+			erase_peers();
+			if (int(m_peers.size()) >= max_peerlist_size)
+				return 0;
+
+			// since some peers were removed, we need to
+			// update the iterator to make it valid again
+#if TORRENT_USE_I2P
+			if (p->is_i2p_addr)
+			{
+				iter = std::lower_bound(
+					m_peers.begin(), m_peers.end()
+					, p->dest(), peer_address_compare());
+			}
+			else
+#endif
+			iter = std::lower_bound(
+				m_peers.begin(), m_peers.end()
+				, p->address(), peer_address_compare());
+		}
+
+		iter = m_peers.insert(iter, p);
+
+		if (m_round_robin >= iter - m_peers.begin()) ++m_round_robin;
+
+#ifndef TORRENT_DISABLE_ENCRYPTION
+		if (flags & 0x01) p->pe_support = true;
+#endif
+		if (flags & 0x02)
+		{
+			p->seed = true;
+			++m_num_seeds;
+		}
+		if (flags & 0x04)
+			p->supports_utp = true;
+		if (flags & 0x08)
+			p->supports_holepunch = true;
+
+#ifndef TORRENT_DISABLE_GEO_IP
+		int as = m_torrent->session().as_for_ip(p->address());
+#ifdef TORRENT_DEBUG
+		p->inet_as_num = as;
+#endif
+		p->inet_as = m_torrent->session().lookup_as(as);
+#endif
+		if (is_connect_candidate(*p, m_finished))
+			++m_num_connect_candidates;
+
+		m_torrent->state_updated();
+
+		return true;
+	}
+
+	void policy::update_peer(policy::peer* p, int src, int flags
+		, tcp::endpoint const& remote, char const* destination)
+	{
+		bool was_conn_cand = is_connect_candidate(*p, m_finished);
+
+		TORRENT_ASSERT(p->in_use);
+		p->connectable = true;
+
+		TORRENT_ASSERT(p->address() == remote.address());
+		p->port = remote.port();
+		p->source |= src;
+			
+		// if this peer has failed before, decrease the
+		// counter to allow it another try, since somebody
+		// else is appearantly able to connect to it
+		// only trust this if it comes from the tracker
+		if (p->failcount > 0 && src == peer_info::tracker)
+			--p->failcount;
+
+		// if we're connected to this peer
+		// we already know if it's a seed or not
+		// so we don't have to trust this source
+		if ((flags & 0x02) && !p->connection)
+		{
+			if (!p->seed) ++m_num_seeds;
+			p->seed = true;
+		}
+		if (flags & 0x04)
+			p->supports_utp = true;
+		if (flags & 0x08)
+			p->supports_holepunch = true;
+
+#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
+		if (p->connection)
+		{
+			// this means we're already connected
+			// to this peer. don't connect to
+			// it again.
+
+			error_code ec;
+			char hex_pid[41];
+			to_hex((char*)&p->connection->pid()[0], 20, hex_pid);
+			char msg[200];
+			snprintf(msg, 200, "already connected to peer: %s %s"
+				, print_endpoint(remote).c_str(), hex_pid);
+			m_torrent->debug_log(msg);
+
+			TORRENT_ASSERT(p->connection->associated_torrent().lock().get() == m_torrent);
+		}
+#endif
+
+		if (was_conn_cand != is_connect_candidate(*p, m_finished))
+		{
+			m_num_connect_candidates += was_conn_cand ? -1 : 1;
+			if (m_num_connect_candidates < 0) m_num_connect_candidates = 0;
+		}
+	}
+
+#if TORRENT_USE_I2P
+	policy::peer* policy::add_i2p_peer(char const* destination, int src, char flags)
+	{
+		INVARIANT_CHECK;
+	
+		bool found = false;
+		iterator iter = std::lower_bound(
+			m_peers.begin(), m_peers.end()
+			, destination, peer_address_compare()
+		);
+
+		if (iter != m_peers.end() && strcmp((*iter)->dest(), destination) == 0)
+			found = true;
+
+		peer* p = 0;
+
+		if (!found)
+		{
+			// we don't have any info about this peer.
+			// add a new entry
+			p = (peer*)m_torrent->session().m_i2p_peer_pool.malloc();
+			if (p == 0) return 0;
+			m_torrent->session().m_i2p_peer_pool.set_next_size(500);
+			new (p) i2p_peer(destination, true, src);
+
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
+			p->in_use = true;
+#endif
+
+			if (!insert_peer(p, iter, flags))
+			{
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
+				p->in_use = false;
+#endif
+
+				m_torrent->session().m_i2p_peer_pool.destroy((i2p_peer*)p);
+				return 0;
+			}
+		}
+		else
+		{
+			p = *iter;
+			update_peer(p, src, flags, tcp::endpoint(), destination);
+		}
+		m_torrent->state_updated();
+		return p;
+	}
+#endif // TORRENT_USE_I2P
+
 	policy::peer* policy::add_peer(tcp::endpoint const& remote, peer_id const& pid
 		, int src, char flags)
 	{
@@ -987,37 +1285,50 @@ namespace libtorrent
 
 		aux::session_impl& ses = m_torrent->session();
 
+		// if this is an i2p torrent, and we don't allow mixed mode
+		// no regular peers should ever be added!
+		if (!ses.m_settings.allow_i2p_mixed && m_torrent->torrent_file().is_i2p())
+		{
+			if (ses.m_alerts.should_post<peer_blocked_alert>())
+				ses.m_alerts.post_alert(peer_blocked_alert(m_torrent->get_handle(), remote.address()));
+			return 0;
+		}
+
 		port_filter const& pf = ses.m_port_filter;
 		if (pf.access(remote.port()) & port_filter::blocked)
 		{
 			if (ses.m_alerts.should_post<peer_blocked_alert>())
-				ses.m_alerts.post_alert(peer_blocked_alert(remote.address()));
+				ses.m_alerts.post_alert(peer_blocked_alert(m_torrent->get_handle(), remote.address()));
+#ifndef TORRENT_DISABLE_EXTENSIONS
+			m_torrent->notify_extension_add_peer(remote, src, torrent_plugin::filtered);
+#endif
 			return 0;
 		}
 
-		if (remote.port() < 1024)
+		if (ses.m_settings.no_connect_privileged_ports && remote.port() < 1024)
 		{
 			if (ses.m_alerts.should_post<peer_blocked_alert>())
-				ses.m_alerts.post_alert(peer_blocked_alert(remote.address()));
+				ses.m_alerts.post_alert(peer_blocked_alert(m_torrent->get_handle(), remote.address()));
+#ifndef TORRENT_DISABLE_EXTENSIONS
+			m_torrent->notify_extension_add_peer(remote, src, torrent_plugin::filtered);
+#endif
 			return 0;
 		}
 
 		// if the IP is blocked, don't add it
-		if (ses.m_ip_filter.access(remote.address()) & ip_filter::blocked)
+		if (m_torrent->apply_ip_filter()
+			&& (ses.m_ip_filter.access(remote.address()) & ip_filter::blocked))
 		{
 			if (ses.m_alerts.should_post<peer_blocked_alert>())
-			{
-				ses.m_alerts.post_alert(peer_blocked_alert(remote.address()));
-			}
+				ses.m_alerts.post_alert(peer_blocked_alert(m_torrent->get_handle(), remote.address()));
+#ifndef TORRENT_DISABLE_EXTENSIONS
+			m_torrent->notify_extension_add_peer(remote, src, torrent_plugin::filtered);
+#endif
 			return 0;
 		}
 
 		iterator iter;
-		peer* i = 0;
-
-		int max_peerlist_size = m_torrent->is_paused()
-			?m_torrent->settings().max_paused_peerlist_size
-			:m_torrent->settings().max_peerlist_size;
+		peer* p = 0;
 
 		bool found = false;
 		if (m_torrent->settings().allow_multiple_connections_per_ip)
@@ -1038,29 +1349,13 @@ namespace libtorrent
 
 		if (!found)
 		{
-			if (max_peerlist_size
-				&& int(m_peers.size()) >= max_peerlist_size)
-			{
-				if (src == peer_info::resume_data) return 0;
-
-				erase_peers();
-				if (int(m_peers.size()) >= max_peerlist_size)
-					return 0;
-
-				// since some peers were removed, we need to
-				// update the iterator to make it valid again
-				iter = std::lower_bound(
-					m_peers.begin(), m_peers.end()
-					, remote.address(), peer_address_compare()
-				);
-			}
-
 			// we don't have any info about this peer.
 			// add a new entry
+
 #if TORRENT_USE_IPV6
 			bool is_v6 = remote.address().is_v6();
 #endif
-			peer* p =
+			p =
 #if TORRENT_USE_IPV6
 				is_v6 ? (peer*)m_torrent->session().m_ipv6_peer_pool.malloc() :
 #endif
@@ -1080,89 +1375,37 @@ namespace libtorrent
 #endif
 				new (p) ipv4_peer(remote, true, src);
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 			p->in_use = true;
 #endif
 
-			iter = m_peers.insert(iter, p);
-
-			if (m_round_robin >= iter - m_peers.begin()) ++m_round_robin;
-
-			i = *iter;
-#ifndef TORRENT_DISABLE_ENCRYPTION
-			if (flags & 0x01) i->pe_support = true;
-#endif
-			if (flags & 0x02)
+			if (!insert_peer(p, iter, flags))
 			{
-				i->seed = true;
-				++m_num_seeds;
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
+				p->in_use = false;
+#endif
+#if TORRENT_USE_IPV6
+				if (is_v6) m_torrent->session().m_ipv6_peer_pool.destroy((ipv6_peer*)p);
+				else
+#endif
+				m_torrent->session().m_ipv4_peer_pool.destroy((ipv4_peer*)p);
+				return 0;
 			}
-
-#ifndef TORRENT_DISABLE_GEO_IP
-			int as = ses.as_for_ip(remote.address());
-#ifdef TORRENT_DEBUG
-			i->inet_as_num = as;
+#ifndef TORRENT_DISABLE_EXTENSIONS
+			m_torrent->notify_extension_add_peer(remote, src, torrent_plugin::first_time);
 #endif
-			i->inet_as = ses.lookup_as(as);
-#endif
-			if (is_connect_candidate(*i, m_finished))
-				++m_num_connect_candidates;
 		}
 		else
 		{
-			i = *iter;
-
-			bool was_conn_cand = is_connect_candidate(*i, m_finished);
-
-			i->connectable = true;
-
-			TORRENT_ASSERT(i->address() == remote.address());
-			i->port = remote.port();
-			i->source |= src;
-				
-			// if this peer has failed before, decrease the
-			// counter to allow it another try, since somebody
-			// else is appearantly able to connect to it
-			// only trust this if it comes from the tracker
-			if (i->failcount > 0 && src == peer_info::tracker)
-				--i->failcount;
-
-			// if we're connected to this peer
-			// we already know if it's a seed or not
-			// so we don't have to trust this source
-			if ((flags & 0x02) && !i->connection)
-			{
-				if (!i->seed) ++m_num_seeds;
-				i->seed = true;
-			}
-
-#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
-			if (i->connection)
-			{
-				// this means we're already connected
-				// to this peer. don't connect to
-				// it again.
-
-				error_code ec;
-				char hex_pid[41];
-				to_hex((char*)&i->connection->pid()[0], 20, hex_pid);
-				char msg[200];
-				snprintf(msg, 200, "already connected to peer: %s %s"
-					, print_endpoint(remote).c_str(), hex_pid);
-				m_torrent->debug_log(msg);
-
-				TORRENT_ASSERT(i->connection->associated_torrent().lock().get() == m_torrent);
-			}
+			p = *iter;
+			TORRENT_ASSERT(p->in_use);
+			update_peer(p, src, flags, remote, 0);
+#ifndef TORRENT_DISABLE_EXTENSIONS
+			m_torrent->notify_extension_add_peer(remote, src, 0);
 #endif
-
-			if (was_conn_cand != is_connect_candidate(*i, m_finished))
-			{
-				m_num_connect_candidates += was_conn_cand ? -1 : 1;
-				if (m_num_connect_candidates < 0) m_num_connect_candidates = 0;
-			}
 		}
 
-		return i;
+		return p;
 	}
 
 	bool policy::connect_one_peer(int session_time)
@@ -1174,6 +1417,7 @@ namespace libtorrent
 		iterator i = find_connect_candidate(session_time);
 		if (i == m_peers.end()) return false;
 		peer& p = **i;
+		TORRENT_ASSERT(p.in_use);
 
 		TORRENT_ASSERT(!p.banned);
 		TORRENT_ASSERT(!p.connection);
@@ -1202,15 +1446,20 @@ namespace libtorrent
 
 		peer* p = c.peer_info_struct();
 
-		TORRENT_ASSERT((std::find_if(
-			m_peers.begin()
-			, m_peers.end()
-			, match_peer_connection(c))
-			!= m_peers.end()) == (p != 0));
-		
 		// if we couldn't find the connection in our list, just ignore it.
 		if (p == 0) return;
 
+		TORRENT_ASSERT(p->in_use);
+
+		// web seeds are special, they're not connected via the peer list
+		// so they're not kept in m_peers
+		TORRENT_ASSERT(p->web_seed
+			|| std::find_if(
+				m_peers.begin()
+				, m_peers.end()
+				, match_peer_connection(c))
+				!= m_peers.end());
+		
 		TORRENT_ASSERT(p->connection == &c);
 		TORRENT_ASSERT(!is_connect_candidate(*p, m_finished));
 
@@ -1248,6 +1497,17 @@ namespace libtorrent
 		// to still exist when we get back there, to assign the new
 		// peer connection pointer to it. The peer list must
 		// be left intact.
+
+		// if we allow multiple connections per IP, and this peer
+		// was incoming and it never advertised its listen
+		// port, we don't really know which peer it was. In order
+		// to avoid adding one entry for every single connection
+		// the peer makes to us, don't save this entry
+		if (m_torrent->settings().allow_multiple_connections_per_ip
+			&& !p->connectable)
+		{
+			erase_peer(p);
+		}
 	}
 
 	void policy::peer_is_interesting(peer_connection& c)
@@ -1282,21 +1542,29 @@ namespace libtorrent
 		}
 	}
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 	bool policy::has_connection(const peer_connection* c)
 	{
 		INVARIANT_CHECK;
 
 		TORRENT_ASSERT(c);
 		error_code ec;
-		TORRENT_ASSERT(c->remote() == c->get_socket()->remote_endpoint(ec) || ec);
+		if (c->remote() != c->get_socket()->remote_endpoint(ec) && !ec)
+		{
+			fprintf(stderr, "c->remote: %s\nc->get_socket()->remote_endpoint: %s\n"
+				, print_endpoint(c->remote()).c_str()
+				, print_endpoint(c->get_socket()->remote_endpoint(ec)).c_str());
+			TORRENT_ASSERT(false);
+		}
 
 		return std::find_if(
 			m_peers.begin()
 			, m_peers.end()
 			, match_peer_connection_or_endpoint(*c)) != m_peers.end();
 	}
+#endif
 
+#ifdef TORRENT_DEBUG
 	void policy::check_invariant() const
 	{
 		TORRENT_ASSERT(m_num_connect_candidates >= 0);
@@ -1325,6 +1593,7 @@ namespace libtorrent
 					TORRENT_ASSERT((*prev)->address() < (*i)->address());
 			}
 			peer const& p = **i;
+			TORRENT_ASSERT(p.in_use);
 			if (is_connect_candidate(p, m_finished)) ++connect_candidates;
 #ifndef TORRENT_DISABLE_GEO_IP
 			TORRENT_ASSERT(p.inet_as == 0 || p.inet_as->first == p.inet_as_num);
@@ -1366,7 +1635,7 @@ namespace libtorrent
 			if ((*i)->is_disconnecting()) continue;
 			// ignore web_peer_connections since they are not managed
 			// by the policy class
-			if (dynamic_cast<web_peer_connection*>(*i)) continue;
+			if ((*i)->type() != peer_connection::bittorrent_connection) continue;
 			++num_torrent_peers;
 		}
 
@@ -1390,7 +1659,11 @@ namespace libtorrent
 			{
 				policy::peer* p = static_cast<policy::peer*>(*i);
 				if (p == 0) continue;
+				TORRENT_ASSERT(p->in_use);
 				if (p->connection == 0) continue;
+				// web seeds are special, they're not connected via the peer list
+				// so they're not kept in m_peers
+				if (p->connection->type() != peer_connection::bittorrent_connection) continue;
 				TORRENT_ASSERT(std::find_if(m_peers.begin(), m_peers.end()
 					, match_peer_connection_or_endpoint(*p->connection)) != m_peers.end());
 			}
@@ -1415,9 +1688,6 @@ namespace libtorrent
 	}
 #endif // TORRENT_DEBUG
 
-	policy::peer::peer()
-	{}
-
 	policy::peer::peer(boost::uint16_t port, bool conn, int src)
 		: prev_amount_upload(0)
 		, prev_amount_download(0)
@@ -1443,13 +1713,20 @@ namespace libtorrent
 #endif
 #if TORRENT_USE_IPV6
 		, is_v6_addr(false)
+#endif
+#if TORRENT_USE_I2P
+		, is_i2p_addr(false)
 #endif
 		, on_parole(false)
 		, banned(false)
 #ifndef TORRENT_DISABLE_DHT
 		, added_to_dht(false)
 #endif
-#ifdef TORRENT_DEBUG
+		, supports_utp(true) // assume peers support utp
+		, confirmed_supports_utp(false)
+		, supports_holepunch(false)
+		, web_seed(false)
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 		, in_use(false)
 #endif
 	{
@@ -1465,7 +1742,7 @@ namespace libtorrent
 		}
 		else
 		{
-			return prev_amount_download;
+			return size_type(prev_amount_download) << 10;
 		}
 	}
 
@@ -1478,13 +1755,20 @@ namespace libtorrent
 		}
 		else
 		{
-			return prev_amount_upload;
+			return size_type(prev_amount_upload) << 10;
 		}
 	}
 
 	// this returns true if lhs is a better erase candidate than rhs
 	bool policy::compare_peer_erase(policy::peer const& lhs, policy::peer const& rhs) const
 	{
+		TORRENT_ASSERT(lhs.connection == 0);
+		TORRENT_ASSERT(rhs.connection == 0);
+
+		// primarily, prefer getting rid of peers we've already tried and failed
+		if (lhs.failcount != rhs.failcount)
+			return lhs.failcount > rhs.failcount;
+
 		bool lhs_resume_data_source = lhs.source == peer_info::resume_data;
 		bool rhs_resume_data_source = rhs.source == peer_info::resume_data;
 
@@ -1492,8 +1776,10 @@ namespace libtorrent
 		if (lhs_resume_data_source != rhs_resume_data_source)
 			return lhs_resume_data_source > rhs_resume_data_source;
 
-		// prefer peers with higher failcount
-		return lhs.failcount > rhs.failcount;
+		if (lhs.connectable != rhs.connectable)
+			return lhs.connectable < rhs.connectable;
+
+		return lhs.trust_points < rhs.trust_points;
 	}
 
 	// this returns true if lhs is a better connect candidate than rhs
