@@ -40,13 +40,8 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/tracker_manager.hpp"
 #include "libtorrent/http_tracker_connection.hpp"
 #include "libtorrent/udp_tracker_connection.hpp"
-#include "libtorrent/entry.hpp"
-#include "libtorrent/bencode.hpp"
-#include "libtorrent/torrent.hpp"
-#include "libtorrent/peer_connection.hpp"
 #include "libtorrent/aux_/session_impl.hpp"
 
-using namespace libtorrent;
 using boost::tuples::make_tuple;
 using boost::tuples::tuple;
 
@@ -90,6 +85,9 @@ namespace libtorrent
 				: (std::min)(m_completion_timeout, timeout);
 		}
 
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("timeout_handler::timeout_callback");
+#endif
 		error_code ec;
 		m_timeout.expires_at(m_read_time + seconds(timeout), ec);
 		m_timeout.async_wait(boost::bind(
@@ -111,6 +109,9 @@ namespace libtorrent
 
 	void timeout_handler::timeout_callback(error_code const& error)
 	{
+#if defined TORRENT_ASIO_DEBUGGING
+		complete_async("timeout_handler::timeout_callback");
+#endif
 		if (m_abort) return;
 
 		ptime now = time_now_hires();
@@ -135,6 +136,9 @@ namespace libtorrent
 				? m_completion_timeout - total_seconds(m_read_time - m_start_time)
 				: (std::min)(m_completion_timeout  - total_seconds(m_read_time - m_start_time), timeout);
 		}
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("timeout_handler::timeout_callback");
+#endif
 		error_code ec;
 		m_timeout.expires_at(m_read_time + seconds(timeout), ec);
 		m_timeout.async_wait(
@@ -157,10 +161,11 @@ namespace libtorrent
 		return m_requester.lock();
 	}
 
-	void tracker_connection::fail(int code, char const* msg, int interval, int min_interval)
+	void tracker_connection::fail(error_code const& ec, int code
+		, char const* msg, int interval, int min_interval)
 	{
 		boost::shared_ptr<request_callback> cb = requester();
-		if (cb) cb->tracker_request_error(m_req, code, msg
+		if (cb) cb->tracker_request_error(m_req, code, ec, msg
 			, interval == 0 ? min_interval : interval);
 		close();
 	}
@@ -175,13 +180,6 @@ namespace libtorrent
 		m_man.received_bytes(bytes);
 	}
 
-	void tracker_connection::fail_timeout()
-	{
-		boost::shared_ptr<request_callback> cb = requester();
-		if (cb) cb->tracker_request_timed_out(m_req);
-		close();
-	}
-	
 	void tracker_connection::close()
 	{
 		cancel();
@@ -196,13 +194,13 @@ namespace libtorrent
 
 	void tracker_manager::sent_bytes(int bytes)
 	{
-//		aux::session_impl::mutex_t::scoped_lock l(m_ses.m_mutex);
+		TORRENT_ASSERT(m_ses.is_network_thread());
 		m_ses.m_stat.sent_tracker_bytes(bytes);
 	}
 
 	void tracker_manager::received_bytes(int bytes)
 	{
-		aux::session_impl::mutex_t::scoped_lock l(m_ses.m_mutex);
+		TORRENT_ASSERT(m_ses.is_network_thread());
 		m_ses.m_stat.received_tracker_bytes(bytes);
 	}
 
@@ -247,7 +245,11 @@ namespace libtorrent
 		{
 			con = new http_tracker_connection(
 				ios, cc, *this, req, c
-				, m_ses, m_proxy, auth);
+				, m_ses, m_proxy, auth
+#if TORRENT_USE_I2P
+				, &m_ses.m_i2p_conn
+#endif
+				);
 		}
 		else if (protocol == "udp")
 		{
@@ -259,8 +261,9 @@ namespace libtorrent
 		{
 			// we need to post the error to avoid deadlock
 			if (boost::shared_ptr<request_callback> r = c.lock())
-				ios.post(boost::bind(&request_callback::tracker_request_error, r, req, -1
-					, "unknown protocol in tracker url: " + req.url, 0));
+				ios.post(boost::bind(&request_callback::tracker_request_error, r, req
+					, -1, error_code(errors::unsupported_url_protocol)
+					, "", 0));
 			return;
 		}
 
@@ -271,6 +274,34 @@ namespace libtorrent
 		con->start();
 	}
 
+	bool tracker_manager::incoming_udp(error_code const& e
+		, udp::endpoint const& ep, char const* buf, int size)
+	{
+		for (tracker_connections_t::iterator i = m_connections.begin();
+			i != m_connections.end();)
+		{
+			boost::intrusive_ptr<tracker_connection> p = *i;
+			++i;
+			// on_receive() may remove the tracker connection from the list
+			if (p->on_receive(e, ep, buf, size)) return true;
+		}
+		return false;
+	}
+
+	bool tracker_manager::incoming_udp(error_code const& e
+		, char const* hostname, char const* buf, int size)
+	{
+		for (tracker_connections_t::iterator i = m_connections.begin();
+			i != m_connections.end();)
+		{
+			boost::intrusive_ptr<tracker_connection> p = *i;
+			++i;
+			// on_receive() may remove the tracker connection from the list
+			if (p->on_receive_hostname(e, hostname, buf, size)) return true;
+		}
+		return false;
+	}
+
 	void tracker_manager::abort_all_requests(bool all)
 	{
 		// removes all connections from m_connections
@@ -278,34 +309,30 @@ namespace libtorrent
 		mutex_t::scoped_lock l(m_mutex);
 
 		m_abort = true;
-		tracker_connections_t keep_connections;
+		tracker_connections_t close_connections;
 
-		while (!m_connections.empty())
+		for (tracker_connections_t::iterator i = m_connections.begin()
+			, end(m_connections.end()); i != end; ++i)
 		{
-			boost::intrusive_ptr<tracker_connection>& c = m_connections.back();
-			if (!c)
-			{
-				m_connections.pop_back();
-				continue;
-			}
+			intrusive_ptr<tracker_connection> c = *i;
 			tracker_request const& req = c->tracker_req();
 			if (req.event == tracker_request::stopped && !all)
-			{
-				keep_connections.push_back(c);
-				m_connections.pop_back();
 				continue;
-			}
-			// close will remove the entry from m_connections
-			// so no need to pop
+
+			close_connections.push_back(c);
 
 #if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING || defined TORRENT_ERROR_LOGGING
 			boost::shared_ptr<request_callback> rc = c->requester();
 			if (rc) rc->debug_log("aborting: " + req.url);
 #endif
-			c->close();
 		}
+		l.unlock();
 
-		std::swap(m_connections, keep_connections);
+		for (tracker_connections_t::iterator i = close_connections.begin()
+			, end(close_connections.end()); i != end; ++i)
+		{
+			(*i)->close();
+		}
 	}
 	
 	bool tracker_manager::empty() const
