@@ -30,9 +30,13 @@ POSSIBILITY OF SUCH DAMAGE.
 
 */
 
+#include "libtorrent/config.hpp"
+#include "libtorrent/socket.hpp"
 #include "libtorrent/udp_socket.hpp"
 #include "libtorrent/connection_queue.hpp"
 #include "libtorrent/escape_string.hpp"
+#include "libtorrent/socket_io.hpp"
+#include "libtorrent/error.hpp"
 #include <stdlib.h>
 #include <boost/bind.hpp>
 #include <boost/array.hpp>
@@ -42,14 +46,27 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <boost/asio/read.hpp>
 #endif
 
+#if defined TORRENT_ASIO_DEBUGGING
+#include "libtorrent/debug.hpp"
+#endif
+
 using namespace libtorrent;
 
-udp_socket::udp_socket(asio::io_service& ios, udp_socket::callback_t const& c
+udp_socket::udp_socket(asio::io_service& ios
+	, udp_socket::callback_t const& c
+	, udp_socket::callback2_t const& c2
 	, connection_queue& cc)
 	: m_callback(c)
+	, m_callback2(c2)
 	, m_ipv4_sock(ios)
+	, m_v4_buf_size(0)
+	, m_v4_buf(0)
+	, m_reallocate_buffer4(false)
 #if TORRENT_USE_IPV6
 	, m_ipv6_sock(ios)
+	, m_v6_buf_size(0)
+	, m_v6_buf(0)
+	, m_reallocate_buffer6(false)
 #endif
 	, m_bind_port(0)
 	, m_v4_outstanding(0)
@@ -65,28 +82,40 @@ udp_socket::udp_socket(asio::io_service& ios, udp_socket::callback_t const& c
 	, m_abort(false)
 	, m_outstanding_ops(0)
 {
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 	m_magic = 0x1337;
 	m_started = false;
 	m_outstanding_when_aborted = -1;
+#if defined BOOST_HAS_PTHREADS
+	m_thread = 0;
+#endif
+#endif
+
+	m_v4_buf_size = 2000;
+	m_v4_buf = (char*)malloc(m_v4_buf_size);
+#if TORRENT_USE_IPV6
+	m_v6_buf_size = 2000;
+	m_v6_buf = (char*)malloc(m_v6_buf_size);
 #endif
 }
 
 udp_socket::~udp_socket()
 {
+	free(m_v4_buf);
 #if TORRENT_USE_IPV6
-	TORRENT_ASSERT(m_v6_outstanding == 0);
+	free(m_v6_buf);
+	TORRENT_ASSERT_VAL(m_v6_outstanding == 0, m_v6_outstanding);
 #endif
-	TORRENT_ASSERT(m_v4_outstanding == 0);
+	TORRENT_ASSERT_VAL(m_v4_outstanding == 0, m_v4_outstanding);
 	TORRENT_ASSERT(m_magic == 0x1337);
 	TORRENT_ASSERT(!m_callback || !m_started);
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 	m_magic = 0;
 #endif
 	TORRENT_ASSERT(m_outstanding_ops == 0);
 }
 
-#ifdef TORRENT_DEBUG
+#if defined TORRENT_DEBUG || TORRENT_RELEASE_ASSERTS
 	#define CHECK_MAGIC check_magic_ cm_(m_magic)
 	struct check_magic_
 	{
@@ -98,25 +127,13 @@ udp_socket::~udp_socket()
 	#define CHECK_MAGIC do {} while (false)
 #endif
 
-bool udp_socket::maybe_clear_callback(mutex_t::scoped_lock& l)
-{
-	if (m_outstanding_ops + m_v4_outstanding + m_v6_outstanding == 0)
-	{
-		// "this" may be destructed in the callback
-		// that's why we need to unlock
-		callback_t tmp = m_callback;
-		m_callback.clear();
-		l.unlock();
-		return true;
-	}
-	return false;
-}
-
-void udp_socket::send(udp::endpoint const& ep, char const* p, int len, error_code& ec)
+void udp_socket::send_hostname(char const* hostname, int port
+	, char const* p, int len, error_code& ec)
 {
 	CHECK_MAGIC;
 
 	TORRENT_ASSERT(is_open());
+	TORRENT_ASSERT(is_single_thread());
 
 	// if the sockets are closed, the udp_socket is closing too
 	if (!is_open()) return;
@@ -124,17 +141,67 @@ void udp_socket::send(udp::endpoint const& ep, char const* p, int len, error_cod
 	if (m_tunnel_packets)
 	{
 		// send udp packets through SOCKS5 server
-		wrap(ep, p, len, ec);
+		wrap(hostname, port, p, len, ec);
 		return;	
 	}
 
-	if (m_queue_packets)
+	// this function is only supported when we're using a proxy
+	TORRENT_ASSERT(m_queue_packets);
+	if (!m_queue_packets) return;
+
+	m_queue.push_back(queued_packet());
+	queued_packet& qp = m_queue.back();
+	qp.ep.port(port);
+	qp.hostname = strdup(hostname);
+	qp.buf.insert(qp.buf.begin(), p, p + len);
+	qp.flags = 0;
+}
+
+bool udp_socket::maybe_clear_callback()
+{
+	if (m_outstanding_ops + m_v4_outstanding
+#if TORRENT_USE_IPV6
+	 	+ m_v6_outstanding
+#endif
+		== 0)
 	{
-		m_queue.push_back(queued_packet());
-		queued_packet& qp = m_queue.back();
-		qp.ep = ep;
-		qp.buf.insert(qp.buf.begin(), p, p + len);
-		return;
+		// "this" may be destructed in the callback
+		m_callback.clear();
+		return true;
+	}
+	return false;
+}
+
+void udp_socket::send(udp::endpoint const& ep, char const* p, int len
+	, error_code& ec, int flags)
+{
+	CHECK_MAGIC;
+
+	TORRENT_ASSERT(is_open());
+	TORRENT_ASSERT(is_single_thread());
+
+	// if the sockets are closed, the udp_socket is closing too
+	if (!is_open()) return;
+
+	if (!(flags & peer_connection) || m_proxy_settings.proxy_peer_connections)
+	{
+		if (m_tunnel_packets)
+		{
+			// send udp packets through SOCKS5 server
+			wrap(ep, p, len, ec);
+			return;	
+		}
+
+		if (m_queue_packets)
+		{
+			m_queue.push_back(queued_packet());
+			queued_packet& qp = m_queue.back();
+			qp.ep = ep;
+			qp.hostname = 0;
+			qp.flags = flags;
+			qp.buf.insert(qp.buf.begin(), p, p + len);
+			return;
+		}
 	}
 
 #if TORRENT_USE_IPV6
@@ -147,10 +214,52 @@ void udp_socket::send(udp::endpoint const& ep, char const* p, int len, error_cod
 #endif
 }
 
+void udp_socket::maybe_realloc_buffers(int which)
+{
+	TORRENT_ASSERT(is_single_thread());
+	bool no_mem = false;
+	if (m_reallocate_buffer4 && (which & 1) && m_v4_outstanding == 0)
+	{
+		TORRENT_ASSERT(m_v4_outstanding == 0);
+		void* tmp = realloc(m_v4_buf, m_v4_buf_size);
+		if (tmp != 0) m_v4_buf = (char*)tmp;
+		else no_mem = true;
+		m_reallocate_buffer4 = false;
+	}
+#if TORRENT_USE_IPV6
+	if (m_reallocate_buffer6 && (which & 2) && m_v6_outstanding == 0)
+	{
+		TORRENT_ASSERT(m_v6_outstanding == 0);
+		void* tmp = realloc(m_v6_buf, m_v6_buf_size);
+		if (tmp != 0) m_v6_buf = (char*)tmp;
+		else no_mem = true;
+		m_reallocate_buffer6 = false;
+	}
+#endif
+
+	if (no_mem)
+	{
+		free(m_v4_buf);
+		m_v4_buf = 0;
+		m_v4_buf_size = 0;
+#if TORRENT_USE_IPV6
+		free(m_v6_buf);
+		m_v6_buf = 0;
+		m_v6_buf_size = 0;
+#endif
+		if (m_callback) m_callback(error::no_memory, m_v4_ep, 0, 0);
+		close();
+	}
+}
+
 void udp_socket::on_read(udp::socket* s, error_code const& e, std::size_t bytes_transferred)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_read");
+#endif
+
 	TORRENT_ASSERT(m_magic == 0x1337);
-	mutex_t::scoped_lock l(m_mutex);	
+	TORRENT_ASSERT(is_single_thread());
 
 #if TORRENT_USE_IPV6
 	if (s == &m_ipv6_sock)
@@ -167,7 +276,7 @@ void udp_socket::on_read(udp::socket* s, error_code const& e, std::size_t bytes_
 
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
@@ -176,24 +285,16 @@ void udp_socket::on_read(udp::socket* s, error_code const& e, std::size_t bytes_
 
 	if (e)
 	{
-		l.unlock();
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
+		TORRENT_TRY {
 
 #if TORRENT_USE_IPV6
-		if (s == &m_ipv4_sock)
-#endif
-			m_callback(e, m_v4_ep, 0, 0);
-#if TORRENT_USE_IPV6
-		else
-			m_callback(e, m_v6_ep, 0, 0);
+			if (s == &m_ipv6_sock)
+				m_callback(e, m_v6_ep, 0, 0);
+			else
 #endif
+				m_callback(e, m_v4_ep, 0, 0);
 
-#ifndef BOOST_NO_EXCEPTIONS
-		} catch(std::exception&) {}
-#endif
-		l.lock();
+		} TORRENT_CATCH (std::exception&) {}
 
 		// don't stop listening on recoverable errors
 		if (e != asio::error::host_unreachable
@@ -208,28 +309,34 @@ void udp_socket::on_read(udp::socket* s, error_code const& e, std::size_t bytes_
 #endif
 			&& e != asio::error::message_size)
 		{
-			maybe_clear_callback(l);
+			maybe_clear_callback();
 			return;
 		}
 
 		if (m_abort) return;
 
-#if TORRENT_USE_IPV6
-		if (s == &m_ipv4_sock && m_v4_outstanding == 0)
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("udp_socket::on_read");
 #endif
-		{
-			++m_v4_outstanding;
-			s->async_receive_from(asio::buffer(m_v4_buf, sizeof(m_v4_buf))
-				, m_v4_ep, boost::bind(&udp_socket::on_read, self(), s, _1, _2));
-		}
 #if TORRENT_USE_IPV6
-		else if (m_v6_outstanding == 0)
+		if (s == &m_ipv6_sock && num_outstanding() == 0)
 		{
+			maybe_realloc_buffers(2);
+			if (m_abort) return;
 			++m_v6_outstanding;
-			s->async_receive_from(asio::buffer(m_v6_buf, sizeof(m_v6_buf))
-				, m_v6_ep, boost::bind(&udp_socket::on_read, self(), s, _1, _2));
+			s->async_receive_from(asio::buffer(m_v6_buf, m_v6_buf_size)
+				, m_v6_ep, boost::bind(&udp_socket::on_read, this, s, _1, _2));
 		}
+		else
 #endif
+		if (m_v4_outstanding == 0)
+		{
+			maybe_realloc_buffers(1);
+			if (m_abort) return;
+			++m_v4_outstanding;
+			s->async_receive_from(asio::buffer(m_v4_buf, m_v4_buf_size)
+				, m_v4_ep, boost::bind(&udp_socket::on_read, this, s, _1, _2));
+		}
 
 #ifdef TORRENT_DEBUG
 		m_started = true;
@@ -238,76 +345,72 @@ void udp_socket::on_read(udp::socket* s, error_code const& e, std::size_t bytes_
 	}
 
 #if TORRENT_USE_IPV6
-	if (s == &m_ipv4_sock)
+	if (s == &m_ipv6_sock)
+	{
+		TORRENT_TRY {
+
+			if (m_tunnel_packets)
+			{
+				// if the source IP doesn't match the proxy's, ignore the packet
+				if (m_v6_ep == m_proxy_addr)
+					unwrap(e, m_v6_buf, bytes_transferred);
+			}
+			else
+			{
+				m_callback(e, m_v6_ep, m_v6_buf, bytes_transferred);
+			}
+
+		} TORRENT_CATCH (std::exception&) {}
+
+		if (m_abort) return;
+
+		if (num_outstanding() == 0)
+		{
+			maybe_realloc_buffers(2);
+			if (m_abort) return;
+
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("udp_socket::on_read");
 #endif
+			++m_v6_outstanding;
+			s->async_receive_from(asio::buffer(m_v6_buf, m_v6_buf_size)
+				, m_v6_ep, boost::bind(&udp_socket::on_read, this, s, _1, _2));
+		}
+	}
+	else
+#endif // TORRENT_USE_IPV6
 	{
 
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
+		TORRENT_TRY {
 
-		if (m_tunnel_packets)
-		{
-			l.unlock();
-			// if the source IP doesn't match the proxy's, ignore the packet
-			if (m_v4_ep == m_proxy_addr)
-				unwrap(e, m_v4_buf, bytes_transferred);
-		}
-		else
-		{
-			l.unlock();
-			m_callback(e, m_v4_ep, m_v4_buf, bytes_transferred);
-		}
-		l.lock();
+			if (m_tunnel_packets)
+			{
+				// if the source IP doesn't match the proxy's, ignore the packet
+				if (m_v4_ep == m_proxy_addr)
+					unwrap(e, m_v4_buf, bytes_transferred);
+			}
+			else
+			{
+				m_callback(e, m_v4_ep, m_v4_buf, bytes_transferred);
+			}
 
-#ifndef BOOST_NO_EXCEPTIONS
-		} catch(std::exception&) {}
-#endif
+		} TORRENT_CATCH (std::exception&) {}
 
 		if (m_abort) return;
 
 		if (m_v4_outstanding == 0)
 		{
+			maybe_realloc_buffers(1);
+			if (m_abort) return;
+
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("udp_socket::on_read");
+#endif
 			++m_v4_outstanding;
-			s->async_receive_from(asio::buffer(m_v4_buf, sizeof(m_v4_buf))
-				, m_v4_ep, boost::bind(&udp_socket::on_read, self(), s, _1, _2));
+			s->async_receive_from(asio::buffer(m_v4_buf, m_v4_buf_size)
+				, m_v4_ep, boost::bind(&udp_socket::on_read, this, s, _1, _2));
 		}
 	}
-#if TORRENT_USE_IPV6
-	else
-	{
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
-
-		if (m_tunnel_packets)
-		{
-			l.unlock();
-			// if the source IP doesn't match the proxy's, ignore the packet
-			if (m_v6_ep == m_proxy_addr)
-				unwrap(e, m_v6_buf, bytes_transferred);
-		}
-		else
-		{
-			l.unlock();
-			m_callback(e, m_v6_ep, m_v6_buf, bytes_transferred);
-		}
-
-#ifndef BOOST_NO_EXCEPTIONS
-		} catch(std::exception&) {}
-#endif
-		l.lock();
-
-		if (m_abort) return;
-
-		if (m_v6_outstanding == 0)
-		{
-			++m_v6_outstanding;
-			s->async_receive_from(asio::buffer(m_v6_buf, sizeof(m_v6_buf))
-				, m_v6_ep, boost::bind(&udp_socket::on_read, self(), s, _1, _2));
-		}
-	}
-#endif // TORRENT_USE_IPV6
 
 #ifdef TORRENT_DEBUG
 	m_started = true;
@@ -325,8 +428,38 @@ void udp_socket::wrap(udp::endpoint const& ep, char const* p, int len, error_cod
 	write_uint16(0, h); // reserved
 	write_uint8(0, h); // fragment
 	write_uint8(ep.address().is_v4()?1:4, h); // atyp
-	write_address(ep.address(), h);
-	write_uint16(ep.port(), h);
+	write_endpoint(ep, h);
+
+	boost::array<asio::const_buffer, 2> iovec;
+	iovec[0] = asio::const_buffer(header, h - header);
+	iovec[1] = asio::const_buffer(p, len);
+
+#if TORRENT_USE_IPV6
+	if (m_proxy_addr.address().is_v4() && m_ipv4_sock.is_open())
+#endif
+		m_ipv4_sock.send_to(iovec, m_proxy_addr, 0, ec);
+#if TORRENT_USE_IPV6
+	else
+		m_ipv6_sock.send_to(iovec, m_proxy_addr, 0, ec);
+#endif
+}
+
+void udp_socket::wrap(char const* hostname, int port, char const* p, int len, error_code& ec)
+{
+	CHECK_MAGIC;
+	using namespace libtorrent::detail;
+
+	char header[270];
+	char* h = header;
+
+	write_uint16(0, h); // reserved
+	write_uint8(0, h); // fragment
+	write_uint8(3, h); // atyp
+	int hostlen = (std::min)(strlen(hostname), size_t(255));
+	write_uint8(hostlen, h); // hostname len
+	memcpy(h, hostname, hostlen);
+	h += hostlen;
+	write_uint16(port, h);
 
 	boost::array<asio::const_buffer, 2> iovec;
 	iovec[0] = asio::const_buffer(header, h - header);
@@ -374,29 +507,51 @@ void udp_socket::unwrap(error_code const& e, char const* buf, int size)
 #endif
 	else
 	{
-		// domain name not supported
+		int len = read_uint8(p);
+		if (len > (buf + size) - p) return;
+		std::string hostname(p, p + len);
+		p += len;
+		m_callback2(e, hostname.c_str(), p, size - (p - buf));
 		return;
 	}
 
 	m_callback(e, sender, p, size - (p - buf));
 }
 
+#if !defined BOOST_ASIO_ENABLE_CANCELIO && defined TORRENT_WINDOWS
+#error BOOST_ASIO_ENABLE_CANCELIO needs to be defined when building libtorrent to enable cancel() in asio on windows
+#endif
+
 void udp_socket::close()
 {
-	mutex_t::scoped_lock l(m_mutex);	
+	TORRENT_ASSERT(is_single_thread());
 	TORRENT_ASSERT(m_magic == 0x1337);
 
 	error_code ec;
-	m_ipv4_sock.close(ec);
+	// if we close the socket here, we can't shut down
+	// utp connections or NAT-PMP. We need to cancel the
+	// outstanding operations
+	m_ipv4_sock.cancel(ec);
+	if (ec == error::operation_not_supported)
+		m_ipv4_sock.close(ec);
+	TORRENT_ASSERT_VAL(!ec || ec == error::bad_descriptor, ec);
 #if TORRENT_USE_IPV6
-	m_ipv6_sock.close(ec);
+	m_ipv6_sock.cancel(ec);
+	if (ec == error::operation_not_supported)
+		m_ipv6_sock.close(ec);
+	TORRENT_ASSERT_VAL(!ec || ec == error::bad_descriptor, ec);
 #endif
-	m_socks5_sock.close(ec);
+	m_socks5_sock.cancel(ec);
+	if (ec == error::operation_not_supported)
+		m_socks5_sock.close(ec);
+	TORRENT_ASSERT_VAL(!ec || ec == error::bad_descriptor, ec);
 	m_resolver.cancel();
 	m_abort = true;
+
 #ifdef TORRENT_DEBUG
-	m_outstanding_when_aborted = m_v4_outstanding + m_v6_outstanding;
+	m_outstanding_when_aborted = num_outstanding();
 #endif
+
 	if (m_connection_ticket >= 0)
 	{
 		m_cc.done(m_connection_ticket);
@@ -409,18 +564,32 @@ void udp_socket::close()
 		--m_outstanding_ops;
 		if (m_abort)
 		{
-			maybe_clear_callback(l);
+			maybe_clear_callback();
 			return;
 		}
 	}
 
-	maybe_clear_callback(l);
+	maybe_clear_callback();
+}
+
+void udp_socket::set_buf_size(int s)
+{
+	TORRENT_ASSERT(is_single_thread());
+	if (s > m_v4_buf_size)
+	{
+		m_v4_buf_size = s;
+		m_reallocate_buffer4 = true;
+#if TORRENT_USE_IPV6
+		m_v6_buf_size = s;
+		m_reallocate_buffer6 = true;
+#endif
+	}
 }
 
 void udp_socket::bind(udp::endpoint const& ep, error_code& ec)
 {
-	mutex_t::scoped_lock l(m_mutex);	
 	CHECK_MAGIC;
+	TORRENT_ASSERT(is_single_thread());
 
 	TORRENT_ASSERT(m_abort == false);
 	if (m_abort) return;
@@ -436,12 +605,17 @@ void udp_socket::bind(udp::endpoint const& ep, error_code& ec)
 		if (ec) return;
 		m_ipv4_sock.bind(ep, ec);
 		if (ec) return;
-
 		if (m_v4_outstanding == 0)
 		{
+			maybe_realloc_buffers(1);
+			if (m_abort) return;
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("udp_socket::on_read");
+#endif
 			++m_v4_outstanding;
-			m_ipv4_sock.async_receive_from(asio::buffer(m_v4_buf, sizeof(m_v4_buf))
-				, m_v4_ep, boost::bind(&udp_socket::on_read, self(), &m_ipv4_sock, _1, _2));
+			m_ipv4_sock.async_receive_from(asio::buffer(m_v4_buf, m_v4_buf_size)
+				, m_v4_ep, boost::bind(&udp_socket::on_read, this, &m_ipv4_sock
+				, _1, _2));
 		}
 	}
 #if TORRENT_USE_IPV6
@@ -453,9 +627,15 @@ void udp_socket::bind(udp::endpoint const& ep, error_code& ec)
 		if (ec) return;
 		if (m_v6_outstanding == 0)
 		{
+			maybe_realloc_buffers(2);
+			if (m_abort) return;
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("udp_socket::on_read");
+#endif
 			++m_v6_outstanding;
-			m_ipv6_sock.async_receive_from(asio::buffer(m_v6_buf, sizeof(m_v6_buf))
-				, m_v6_ep, boost::bind(&udp_socket::on_read, self(), &m_ipv6_sock, _1, _2));
+			m_ipv6_sock.async_receive_from(asio::buffer(m_v6_buf, m_v6_buf_size)
+				, m_v6_ep, boost::bind(&udp_socket::on_read, this, &m_ipv6_sock
+				, _1, _2));
 		}
 	}
 #endif
@@ -467,8 +647,8 @@ void udp_socket::bind(udp::endpoint const& ep, error_code& ec)
 
 void udp_socket::bind(int port)
 {
-	mutex_t::scoped_lock l(m_mutex);	
 	CHECK_MAGIC;
+	TORRENT_ASSERT(is_single_thread());
 
 	TORRENT_ASSERT(m_abort == false);
 	if (m_abort) return;
@@ -480,42 +660,54 @@ void udp_socket::bind(int port)
 	if (m_ipv6_sock.is_open()) m_ipv6_sock.close(ec);
 #endif
 
+	maybe_realloc_buffers();
+	if (m_abort) return;
+
 	m_ipv4_sock.open(udp::v4(), ec);
 	if (!ec)
 	{
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("udp_socket::on_read");
+#endif
 		m_ipv4_sock.bind(udp::endpoint(address_v4::any(), port), ec);
 		if (m_v4_outstanding == 0)
 		{
 			++m_v4_outstanding;
-			m_ipv4_sock.async_receive_from(asio::buffer(m_v4_buf, sizeof(m_v4_buf))
-				, m_v4_ep, boost::bind(&udp_socket::on_read, self(), &m_ipv4_sock, _1, _2));
+			m_ipv4_sock.async_receive_from(asio::buffer(m_v4_buf, m_v4_buf_size)
+				, m_v4_ep, boost::bind(&udp_socket::on_read, this, &m_ipv4_sock
+				, _1, _2));
 		}
 	}
 #if TORRENT_USE_IPV6
 	m_ipv6_sock.open(udp::v6(), ec);
 	if (!ec)
 	{
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("udp_socket::on_read");
+#endif
 		m_ipv6_sock.set_option(v6only(true), ec);
 		m_ipv6_sock.bind(udp::endpoint(address_v6::any(), port), ec);
+
 		if (m_v6_outstanding == 0)
 		{
 			++m_v6_outstanding;
-			m_ipv6_sock.async_receive_from(asio::buffer(m_v6_buf, sizeof(m_v6_buf))
-				, m_v6_ep, boost::bind(&udp_socket::on_read, self(), &m_ipv6_sock, _1, _2));
+			m_ipv6_sock.async_receive_from(asio::buffer(m_v6_buf, m_v6_buf_size)
+				, m_v6_ep, boost::bind(&udp_socket::on_read, this, &m_ipv6_sock
+				, _1, _2));
 		}
 	}
 #endif // TORRENT_USE_IPV6
 
 #ifdef TORRENT_DEBUG
-	m_started = true;
+		m_started = true;
 #endif
 	m_bind_port = port;
 }
 
 void udp_socket::set_proxy_settings(proxy_settings const& ps)
 {
-	mutex_t::scoped_lock l(m_mutex);	
 	CHECK_MAGIC;
+	TORRENT_ASSERT(is_single_thread());
 
 	error_code ec;
 	m_socks5_sock.close(ec);
@@ -523,6 +715,8 @@ void udp_socket::set_proxy_settings(proxy_settings const& ps)
 	
 	m_proxy_settings = ps;
 
+	if (m_abort) return;
+
 	if (ps.type == proxy_settings::socks5
 		|| ps.type == proxy_settings::socks5_pw)
 	{
@@ -531,40 +725,36 @@ void udp_socket::set_proxy_settings(proxy_settings const& ps)
 		tcp::resolver::query q(ps.hostname, to_string(ps.port).elems);
 		++m_outstanding_ops;
 		m_resolver.async_resolve(q, boost::bind(
-			&udp_socket::on_name_lookup, self(), _1, _2));
+			&udp_socket::on_name_lookup, this, _1, _2));
 	}
 }
 
 void udp_socket::on_name_lookup(error_code const& e, tcp::resolver::iterator i)
 {
-	mutex_t::scoped_lock l(m_mutex);	
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 	CHECK_MAGIC;
 
 	if (e == asio::error::operation_aborted) return;
 
+	TORRENT_ASSERT(is_single_thread());
+
 	if (e)
 	{
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
-			m_callback(e, udp::endpoint(), 0, 0);
-#ifndef BOOST_NO_EXCEPTIONS
-		} catch(std::exception&) {}
-#endif
+		TORRENT_TRY {
+			if (m_callback) m_callback(e, udp::endpoint(), 0, 0);
+		} TORRENT_CATCH (std::exception&) {}
 		return;
 	}
 
 	m_proxy_addr.address(i->endpoint().address());
 	m_proxy_addr.port(i->endpoint().port());
-	l.unlock();
 	// on_connect may be called from within this thread
 	// the semantics for on_connect and on_timeout is 
 	// a bit complicated. See comments in connection_queue.hpp
@@ -573,14 +763,21 @@ void udp_socket::on_name_lookup(error_code const& e, tcp::resolver::iterator i)
 	// To simplyfy this, it's probably a good idea to
 	// merge on_connect and on_timeout to a single function
 	++m_outstanding_ops;
-	m_cc.enqueue(boost::bind(&udp_socket::on_connect, self(), _1)
-		, boost::bind(&udp_socket::on_timeout, self()), seconds(10));
+	m_cc.enqueue(boost::bind(&udp_socket::on_connect, this, _1)
+		, boost::bind(&udp_socket::on_timeout, this), seconds(10));
 }
 
 void udp_socket::on_timeout()
 {
-	mutex_t::scoped_lock l(m_mutex);	
+	TORRENT_ASSERT(m_outstanding_ops > 0);
+	--m_outstanding_ops;
+	if (m_abort)
+	{
+		maybe_clear_callback();
+		return;
+	}
 	CHECK_MAGIC;
+	TORRENT_ASSERT(is_single_thread());
 
 	error_code ec;
 	m_socks5_sock.close(ec);
@@ -589,18 +786,27 @@ void udp_socket::on_timeout()
 
 void udp_socket::on_connect(int ticket)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+	TORRENT_ASSERT(is_single_thread());
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 	CHECK_MAGIC;
 
+	if (m_abort) return;
 	if (is_closed()) return;
+	if (ticket == -1)
+	{
+		close();
+		return;
+	}
 
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::on_connected");
+#endif
 	m_connection_ticket = ticket;
 	// at this point on_timeout may be called before on_connected
 	// so increment the outstanding ops
@@ -613,51 +819,49 @@ void udp_socket::on_connect(int ticket)
 	m_socks5_sock.open(m_proxy_addr.address().is_v4()?tcp::v4():tcp::v6(), ec);
 	++m_outstanding_ops;
 	m_socks5_sock.async_connect(tcp::endpoint(m_proxy_addr.address(), m_proxy_addr.port())
-		, boost::bind(&udp_socket::on_connected, self(), _1));
+		, boost::bind(&udp_socket::on_connected, this, _1));
 }
 
 void udp_socket::on_connected(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_connected");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
+
 	CHECK_MAGIC;
 
 	if (e == asio::error::operation_aborted) return;
 
+	TORRENT_ASSERT(is_single_thread());
 	m_cc.done(m_connection_ticket);
 	m_connection_ticket = -1;
 
-	// we just called done, which meand on_timeout
+	// we just called done, which means on_timeout
 	// won't be called. Decrement the outstanding
 	// ops counter for that
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
 	if (e)
 	{
-#ifndef BOOST_NO_EXCEPTIONS
-		try {
-#endif
-			m_callback(e, udp::endpoint(), 0, 0);
-#ifndef BOOST_NO_EXCEPTIONS
-		} catch(std::exception&) {}
-#endif
+		TORRENT_TRY {
+			if (m_callback) m_callback(e, udp::endpoint(), 0, 0);
+		} TORRENT_CATCH (std::exception&) {}
 		return;
 	}
 
-	if (is_closed()) return;
-
 	using namespace libtorrent::detail;
 
 	// send SOCKS5 authentication methods
@@ -675,38 +879,51 @@ void udp_socket::on_connected(error_code const& e)
 		write_uint8(0, p); // no authentication
 		write_uint8(2, p); // username/password
 	}
+	TORRENT_ASSERT_VAL(p - m_tmp_buf < int(sizeof(m_tmp_buf)), (p - m_tmp_buf));
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::on_handshake1");
+#endif
 	++m_outstanding_ops;
 	asio::async_write(m_socks5_sock, asio::buffer(m_tmp_buf, p - m_tmp_buf)
-		, boost::bind(&udp_socket::handshake1, self(), _1));
+		, boost::bind(&udp_socket::handshake1, this, _1));
 }
 
 void udp_socket::handshake1(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_handshake1");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
 	CHECK_MAGIC;
 	if (e) return;
 
+	TORRENT_ASSERT(is_single_thread());
+
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::on_handshake2");
+#endif
 	++m_outstanding_ops;
 	asio::async_read(m_socks5_sock, asio::buffer(m_tmp_buf, 2)
-		, boost::bind(&udp_socket::handshake2, self(), _1));
+		, boost::bind(&udp_socket::handshake2, this, _1));
 }
 
 void udp_socket::handshake2(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_handshake2");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 	CHECK_MAGIC;
@@ -715,6 +932,8 @@ void udp_socket::handshake2(error_code const& e)
 
 	using namespace libtorrent::detail;
 
+	TORRENT_ASSERT(is_single_thread());
+
 	char* p = &m_tmp_buf[0];
 	int version = read_uint8(p);
 	int method = read_uint8(p);
@@ -723,7 +942,7 @@ void udp_socket::handshake2(error_code const& e)
 
 	if (method == 0)
 	{
-		socks_forward_udp(l);
+		socks_forward_udp(/*l*/);
 	}
 	else if (method == 2)
 	{
@@ -741,9 +960,13 @@ void udp_socket::handshake2(error_code const& e)
 		write_string(m_proxy_settings.username, p);
 		write_uint8(m_proxy_settings.password.size(), p);
 		write_string(m_proxy_settings.password, p);
+		TORRENT_ASSERT_VAL(p - m_tmp_buf < int(sizeof(m_tmp_buf)), (p - m_tmp_buf));
+#if defined TORRENT_ASIO_DEBUGGING
+		add_outstanding_async("udp_socket::on_handshake3");
+#endif
 		++m_outstanding_ops;
 		asio::async_write(m_socks5_sock, asio::buffer(m_tmp_buf, p - m_tmp_buf)
-			, boost::bind(&udp_socket::handshake3, self(), _1));
+			, boost::bind(&udp_socket::handshake3, this, _1));
 	}
 	else
 	{
@@ -755,37 +978,48 @@ void udp_socket::handshake2(error_code const& e)
 
 void udp_socket::handshake3(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_handshake3");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
 	CHECK_MAGIC;
 	if (e) return;
 
+	TORRENT_ASSERT(is_single_thread());
+
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::on_handshake4");
+#endif
 	++m_outstanding_ops;
 	asio::async_read(m_socks5_sock, asio::buffer(m_tmp_buf, 2)
-		, boost::bind(&udp_socket::handshake4, self(), _1));
+		, boost::bind(&udp_socket::handshake4, this, _1));
 }
 
 void udp_socket::handshake4(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::on_handshake4");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
 	CHECK_MAGIC;
 	if (e) return;
 
+	TORRENT_ASSERT(is_single_thread());
+
 	using namespace libtorrent::detail;
 
 	char* p = &m_tmp_buf[0];
@@ -795,18 +1029,11 @@ void udp_socket::handshake4(error_code const& e)
 	if (version != 1) return;
 	if (status != 0) return;
 
-	socks_forward_udp(l);
+	socks_forward_udp(/*l*/);
 }
 
-void udp_socket::socks_forward_udp(mutex_t::scoped_lock& l)
+void udp_socket::socks_forward_udp()
 {
-	TORRENT_ASSERT(m_outstanding_ops > 0);
-	--m_outstanding_ops;
-	if (m_abort)
-	{
-		maybe_clear_callback(l);
-		return;
-	}
 	CHECK_MAGIC;
 	using namespace libtorrent::detail;
 
@@ -815,43 +1042,66 @@ void udp_socket::socks_forward_udp(mutex_t::scoped_lock& l)
 	write_uint8(5, p); // SOCKS VERSION 5
 	write_uint8(3, p); // UDP ASSOCIATE command
 	write_uint8(0, p); // reserved
-	write_uint8(1, p); // ATYP IPv4
-	write_uint32(0, p); // IP any
-	write_uint16(m_bind_port, p);
-
+	error_code ec;
+	tcp::endpoint local = m_socks5_sock.local_endpoint(ec);
+	write_uint8(local.address().is_v4() ? 1 : 4, p); // ATYP IPv4
+	detail::write_address(local.address(), p);
+	int port = 0;
+#if TORRENT_USE_IPV6
+	if (local.address().is_v4())
+#endif
+		port = m_ipv4_sock.local_endpoint(ec).port();
+#if TORRENT_USE_IPV6
+	else
+		port = m_ipv6_sock.local_endpoint(ec).port();
+#endif
+	detail::write_uint16(port , p);
+	TORRENT_ASSERT_VAL(p - m_tmp_buf < int(sizeof(m_tmp_buf)), (p - m_tmp_buf));
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::connect1");
+#endif
 	++m_outstanding_ops;
 	asio::async_write(m_socks5_sock, asio::buffer(m_tmp_buf, p - m_tmp_buf)
-		, boost::bind(&udp_socket::connect1, self(), _1));
+		, boost::bind(&udp_socket::connect1, this, _1));
 }
 
 void udp_socket::connect1(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::connect1");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
 	CHECK_MAGIC;
 	if (e) return;
 
+	TORRENT_ASSERT(is_single_thread());
+
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::connect2");
+#endif
 	++m_outstanding_ops;
 	asio::async_read(m_socks5_sock, asio::buffer(m_tmp_buf, 10)
-		, boost::bind(&udp_socket::connect2, self(), _1));
+		, boost::bind(&udp_socket::connect2, this, _1));
 }
 
 void udp_socket::connect2(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::connect2");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
 		m_queue.clear();
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
 
@@ -862,6 +1112,8 @@ void udp_socket::connect2(error_code const& e)
 		return;
 	}
 
+	TORRENT_ASSERT(is_single_thread());
+
 	using namespace libtorrent::detail;
 
 	char* p = &m_tmp_buf[0];
@@ -897,83 +1149,106 @@ void udp_socket::connect2(error_code const& e)
 	{
 		queued_packet const& p = m_queue.front();
 		error_code ec;
-		udp_socket::send(p.ep, &p.buf[0], p.buf.size(), ec);
+		if (p.hostname)
+		{
+			udp_socket::send_hostname(p.hostname, p.ep.port(), &p.buf[0], p.buf.size(), ec);
+			free(p.hostname);
+		}
+		else
+		{
+			udp_socket::send(p.ep, &p.buf[0], p.buf.size(), ec, p.flags);
+		}
 		m_queue.pop_front();
 	}
 
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("udp_socket::hung_up");
+#endif
 	++m_outstanding_ops;
 	asio::async_read(m_socks5_sock, asio::buffer(m_tmp_buf, 10)
-		, boost::bind(&udp_socket::hung_up, self(), _1));
+		, boost::bind(&udp_socket::hung_up, this, _1));
 }
 
 void udp_socket::hung_up(error_code const& e)
 {
-	mutex_t::scoped_lock l(m_mutex);	
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("udp_socket::hung_up");
+#endif
 	TORRENT_ASSERT(m_outstanding_ops > 0);
 	--m_outstanding_ops;
 	if (m_abort)
 	{
-		maybe_clear_callback(l);
+		maybe_clear_callback();
 		return;
 	}
-
+  
 	CHECK_MAGIC;
-	if (e == asio::error::operation_aborted || m_abort) return;
+	TORRENT_ASSERT(is_single_thread());
 
-	l.unlock();
+	if (e == asio::error::operation_aborted || m_abort) return;
 
 	// the socks connection was closed, re-open it
 	set_proxy_settings(m_proxy_settings);
 }
 
 rate_limited_udp_socket::rate_limited_udp_socket(io_service& ios
-	, callback_t const& c, connection_queue& cc)
-	: udp_socket(ios, c, cc)
+	, callback_t const& c
+	, callback2_t const& c2
+	, connection_queue& cc)
+	: udp_socket(ios, c, c2, cc)
 	, m_timer(ios)
 	, m_queue_size_limit(200)
 	, m_rate_limit(4000)
 	, m_quota(4000)
 	, m_last_tick(time_now())
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("rate_limited_udp_socket::on_tick");
+#endif
 	error_code ec;
 	m_timer.expires_from_now(seconds(1), ec);
-	m_timer.async_wait(boost::bind(&rate_limited_udp_socket::on_tick
-		, boost::intrusive_ptr<rate_limited_udp_socket>(this), _1));
-	TORRENT_ASSERT(!ec);
+	m_timer.async_wait(boost::bind(&rate_limited_udp_socket::on_tick, this, _1));
+	TORRENT_ASSERT_VAL(!ec, ec);
 }
 
-bool rate_limited_udp_socket::send(udp::endpoint const& ep, char const* p, int len, error_code& ec, int flags)
+bool rate_limited_udp_socket::send(udp::endpoint const& ep, char const* p
+	, int len, error_code& ec, int flags)
 {
 	if (m_quota < len)
 	{
-		// bit 1 of flags means "don't drop"
-		if (int(m_queue.size()) >= m_queue_size_limit && (flags & 1) == 0)
+		if (int(m_queue.size()) >= m_queue_size_limit && (flags & dont_drop) == 0)
 			return false;
 		m_queue.push_back(queued_packet());
 		queued_packet& qp = m_queue.back();
 		qp.ep = ep;
+		qp.flags = flags;
 		qp.buf.insert(qp.buf.begin(), p, p + len);
 		return true;
 	}
 
 	m_quota -= len;
-	udp_socket::send(ep, p, len, ec);
+	udp_socket::send(ep, p, len, ec, flags);
 	return true;
 }
 
 void rate_limited_udp_socket::on_tick(error_code const& e)
 {
+#if defined TORRENT_ASIO_DEBUGGING
+	complete_async("rate_limited_udp_socket::on_tick");
+#endif
 	if (e) return;
 	if (is_closed()) return;
 	error_code ec;
 	ptime now = time_now_hires();
+#if defined TORRENT_ASIO_DEBUGGING
+	add_outstanding_async("rate_limited_udp_socket::on_tick");
+#endif
 	m_timer.expires_at(now + seconds(1), ec);
-	m_timer.async_wait(boost::bind(&rate_limited_udp_socket::on_tick
-		, boost::intrusive_ptr<rate_limited_udp_socket>(this), _1));
+	m_timer.async_wait(boost::bind(&rate_limited_udp_socket::on_tick, this, _1));
 
 	time_duration delta = now - m_last_tick;
 	m_last_tick = now;
-	if (m_quota < m_rate_limit) m_quota += m_rate_limit * total_milliseconds(delta) / 1000;
+	if (m_quota < m_rate_limit) m_quota += boost::uint64_t(m_rate_limit) * total_milliseconds(delta) / 1000;
 
 	if (m_queue.empty()) return;
 
@@ -983,7 +1258,7 @@ void rate_limited_udp_socket::on_tick(error_code const& e)
 		TORRENT_ASSERT(m_quota >= int(p.buf.size()));
 		m_quota -= p.buf.size();
 		error_code ec;
-		udp_socket::send(p.ep, &p.buf[0], p.buf.size(), ec);
+		udp_socket::send(p.ep, &p.buf[0], p.buf.size(), ec, p.flags);
 		m_queue.pop_front();
 	}
 }
