@@ -35,8 +35,6 @@ POSSIBILITY OF SUCH DAMAGE.
 #include <vector>
 #include <cctype>
 
-#include "zlib.h"
-
 #ifdef _MSC_VER
 #pragma warning(push, 1)
 #endif
@@ -53,6 +51,8 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/io.hpp"
 #include "libtorrent/aux_/session_impl.hpp"
 #include "libtorrent/escape_string.hpp"
+#include "libtorrent/broadcast_socket.hpp" // for is_any
+#include "libtorrent/random.hpp"
 
 namespace libtorrent
 {
@@ -60,7 +60,7 @@ namespace libtorrent
 	std::map<address, udp_tracker_connection::connection_cache_entry>
 		udp_tracker_connection::m_connection_cache;
 
-	boost::mutex udp_tracker_connection::m_cache_mutex;
+	mutex udp_tracker_connection::m_cache_mutex;
 
 	udp_tracker_connection::udp_tracker_connection(
 		io_service& ios
@@ -68,19 +68,17 @@ namespace libtorrent
 		, tracker_manager& man
 		, tracker_request const& req
 		, boost::weak_ptr<request_callback> c
-		, aux::session_impl const& ses
+		, aux::session_impl& ses
 		, proxy_settings const& proxy)
 		: tracker_connection(man, req, ios, c)
-		, m_man(man)
-		, m_name_lookup(ios)
-		, m_socket(new udp_socket(ios, boost::bind(&udp_tracker_connection::on_receive, self(), _1, _2, _3, _4), cc))
+//		, m_man(man)
+		, m_abort(false)
 		, m_transaction_id(0)
 		, m_ses(ses)
 		, m_attempts(0)
 		, m_state(action_error)
+		, m_proxy(proxy)
 	{
-		TORRENT_ASSERT(refcount() == 1);
-		m_socket->set_proxy_settings(proxy);
 	}
 
 	void udp_tracker_connection::start()
@@ -96,34 +94,53 @@ namespace libtorrent
 		if (ec)
 		{
 			// never call fail() when the session mutex is locked!
-			m_socket->get_io_service().post(boost::bind(
-				&tracker_connection::fail_disp, self(), -1, ec.message()));
+			m_ses.m_io_service.post(boost::bind(
+				&tracker_connection::fail_disp, self(), ec));
 			return;
 		}
 		
 		session_settings const& settings = m_ses.settings();
 
-		udp::resolver::query q(hostname, to_string(port).elems);
-		m_name_lookup.async_resolve(q
-			, boost::bind(
-			&udp_tracker_connection::name_lookup, self(), _1, _2));
+		if (m_proxy.proxy_hostnames
+			&& (m_proxy.type == proxy_settings::socks5
+				|| m_proxy.type == proxy_settings::socks5_pw))
+		{
+			m_hostname = hostname;
+			m_target.port(port);
+			start_announce();
+		}
+		else
+		{
+#if defined TORRENT_ASIO_DEBUGGING
+			add_outstanding_async("udp_tracker_connection::name_lookup");
+#endif
+			tcp::resolver::query q(hostname, to_string(port).elems);
+			m_ses.m_host_resolver.async_resolve(q
+				, boost::bind(
+				&udp_tracker_connection::name_lookup, self(), _1, _2));
+#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
+			boost::shared_ptr<request_callback> cb = requester();
+			if (cb) cb->debug_log(("*** UDP_TRACKER [ initiating name lookup: " + hostname + " ]").c_str());
+#endif
+		}
+
 		set_timeout(tracker_req().event == tracker_request::stopped
 			? settings.stop_tracker_timeout
 			: settings.tracker_completion_timeout
 			, settings.tracker_receive_timeout);
-#if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
-		boost::shared_ptr<request_callback> cb = requester();
-		if (cb) cb->debug_log(("*** UDP_TRACKER [ initiating name lookup: " + hostname + " ]").c_str());
-#endif
 	}
 
 	void udp_tracker_connection::name_lookup(error_code const& error
-		, udp::resolver::iterator i)
+		, tcp::resolver::iterator i)
 	{
+#if defined TORRENT_ASIO_DEBUGGING
+		complete_async("udp_tracker_connection::name_lookup");
+#endif
+		if (m_abort) return;
 		if (error == asio::error::operation_aborted) return;
-		if (error || i == udp::resolver::iterator())
+		if (error || i == tcp::resolver::iterator())
 		{
-			fail(-1, error.message().c_str());
+			fail(error);
 			return;
 		}
 
@@ -133,7 +150,7 @@ namespace libtorrent
 #endif
 		if (cancelled())
 		{
-			fail(-1, "aborted");
+			fail(error_code(errors::torrent_aborted));
 			return;
 		}
 
@@ -143,32 +160,36 @@ namespace libtorrent
 		// we're listening on. To make sure the tracker get our
 		// correct listening address.
 
-		std::transform(i, udp::resolver::iterator(), std::back_inserter(m_endpoints)
-			, boost::bind(&udp::resolver::iterator::value_type::endpoint, _1));
+		std::transform(i, tcp::resolver::iterator(), std::back_inserter(m_endpoints)
+			, boost::bind(&tcp::resolver::iterator::value_type::endpoint, _1));
 
-		// remove endpoints that are filtered by the IP filter
-		for (std::list<udp::endpoint>::iterator i = m_endpoints.begin();
-			i != m_endpoints.end();)
+		if (tracker_req().apply_ip_filter)
 		{
-			if (m_ses.m_ip_filter.access(i->address()) == ip_filter::blocked) 
+			// remove endpoints that are filtered by the IP filter
+			for (std::list<tcp::endpoint>::iterator k = m_endpoints.begin();
+				k != m_endpoints.end();)
 			{
+				if (m_ses.m_ip_filter.access(k->address()) == ip_filter::blocked) 
+				{
 #if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
-				if (cb) cb->debug_log("*** UDP_TRACKER [ IP blocked by filter: " + print_address(i->address()) + " ]");
+					if (cb) cb->debug_log("*** UDP_TRACKER [ IP blocked by filter: " + print_address(k->address()) + " ]");
 #endif
-				i = m_endpoints.erase(i);
+					k = m_endpoints.erase(k);
+				}
+				else
+					++k;
 			}
-			else
-				++i;
 		}
 
+		// if all endpoints were filtered by the IP filter, we can't connect
 		if (m_endpoints.empty())
 		{
-			fail(-1, "blocked by IP filter");
+			fail(error_code(errors::banned_by_ip_filter));
 			return;
 		}
 		
-		std::list<udp::endpoint>::iterator iter = m_endpoints.begin();
-		m_target = *iter;
+		std::list<tcp::endpoint>::iterator iter = m_endpoints.begin();
+		m_target = udp::endpoint(iter->address(), iter->port());
 
 		if (bind_interface() != address_v4::any())
 		{
@@ -195,21 +216,18 @@ namespace libtorrent
 			}
 			else
 			{
-				m_target = *iter;
+				m_target = udp::endpoint(iter->address(), iter->port());
 			}
 		}
 
 		if (cb) cb->m_tracker_address = tcp::endpoint(m_target.address(), m_target.port());
 
-		error_code ec;
-		m_socket->bind(udp::endpoint(bind_interface(), 0), ec);
-		if (ec)
-		{
-			fail(-1, ec.message().c_str());
-			return;
-		}
+		start_announce();
+	}
 
-		boost::mutex::scoped_lock l(m_cache_mutex);
+	void udp_tracker_connection::start_announce()
+	{
+		mutex::scoped_lock l(m_cache_mutex);
 		std::map<address, connection_cache_entry>::iterator cc
 			= m_connection_cache.find(m_target.address());
 		if (cc != m_connection_cache.end())
@@ -236,7 +254,7 @@ namespace libtorrent
 	{
 		if (ec)
 		{
-			fail(-1, ec.message().c_str());
+			fail(ec);
 			return;
 		}
 
@@ -246,32 +264,41 @@ namespace libtorrent
 		snprintf(msg, 200, "*** UDP_TRACKER [ timed out url: %s ]", tracker_req().url.c_str());
 		if (cb) cb->debug_log(msg);
 #endif
-		m_socket->close();
-		m_name_lookup.cancel();
-		fail_timeout();
+		m_abort = true;
+		fail(error_code(errors::timed_out));
 	}
 
 	void udp_tracker_connection::close()
 	{
 		error_code ec;
-		m_socket->close();
-		m_name_lookup.cancel();
 		tracker_connection::close();
 	}
 
-	void udp_tracker_connection::on_receive(error_code const& e
+	bool udp_tracker_connection::on_receive_hostname(error_code const& e
+		, char const* hostname, char const* buf, int size)
+	{
+		// just ignore the hostname this came from, pretend that
+		// it's from the same endpoint we sent it to (i.e. the same
+		// port). We have so many other ways of confirming this packet
+		// comes from the tracker anyway, so it's not a big deal
+		return on_receive(e, m_target, buf, size);
+	}
+
+	bool udp_tracker_connection::on_receive(error_code const& e
 		, udp::endpoint const& ep, char const* buf, int size)
 	{
 		// ignore resposes before we've sent any requests
-		if (m_state == action_error) return;
+		if (m_state == action_error) return false;
 
-		if (!m_socket->is_open()) return; // the operation was aborted
+		if (m_abort) return false;
 
 		// ignore packet not sent from the tracker
-		if (m_target != ep) return;
+		// if m_target is inaddr_any, it suggests that we
+		// sent the packet through a proxy only knowing
+		// the hostname, in which case this packet might be for us
+		if (!is_any(m_target.address()) && m_target != ep) return false;
 		
-		received_bytes(size + 28); // assuming UDP/IP header
-		if (e) fail(-1, e.message().c_str());
+		if (e) fail(e);
 
 #if defined TORRENT_VERBOSE_LOGGING || defined TORRENT_LOGGING
 		boost::shared_ptr<request_callback> cb = requester();
@@ -284,7 +311,7 @@ namespace libtorrent
 #endif
 
 		// ignore packets smaller than 8 bytes
-		if (size < 8) return;
+		if (size < 8) return false;
 
 		const char* ptr = buf;
 		int action = detail::read_int32(ptr);
@@ -300,16 +327,16 @@ namespace libtorrent
 #endif
 
 		// ignore packets with incorrect transaction id
-		if (m_transaction_id != transaction) return;
+		if (m_transaction_id != transaction) return false;
 
 		if (action == action_error)
 		{
-			fail(-1, std::string(ptr, size - 8).c_str());
-			return;
+			fail(error_code(errors::tracker_failure), -1, std::string(ptr, size - 8).c_str());
+			return true;
 		}
 
 		// ignore packets that's not a response to our message
-		if (action != m_state) return;
+		if (action != m_state) return false;
 
 		restart_read_timeout();
 
@@ -326,22 +353,20 @@ namespace libtorrent
 		switch (m_state)
 		{
 			case action_connect:
-				on_connect_response(buf, size);
-				break;
+				return on_connect_response(buf, size);
 			case action_announce:
-				on_announce_response(buf, size);
-				break;
+				return on_announce_response(buf, size);
 			case action_scrape:
-				on_scrape_response(buf, size);
-				break;
+				return on_scrape_response(buf, size);
 			default: break;
 		}
+		return false;
 	}
 	
-	void udp_tracker_connection::on_connect_response(char const* buf, int size)
+	bool udp_tracker_connection::on_connect_response(char const* buf, int size)
 	{
 		// ignore packets smaller than 16 bytes
-		if (size < 16) return;
+		if (size < 16) return false;
 
 		restart_read_timeout();
 		buf += 8; // skip header
@@ -351,7 +376,7 @@ namespace libtorrent
 		m_attempts = 0;
 		boost::uint64_t connection_id = detail::read_int64(buf);
 
-		boost::mutex::scoped_lock l(m_cache_mutex);
+		mutex::scoped_lock l(m_cache_mutex);
 		connection_cache_entry& cce = m_connection_cache[m_target.address()];
 		cce.connection_id = connection_id;
 		cce.expires = time_now() + seconds(m_ses.m_settings.udp_tracker_token_expiry);
@@ -360,6 +385,7 @@ namespace libtorrent
 			send_udp_announce();
 		else if (tracker_req().kind == tracker_request::scrape_request)
 			send_udp_scrape();
+		return true;
 	}
 
 	void udp_tracker_connection::send_udp_connect()
@@ -375,13 +401,13 @@ namespace libtorrent
 			cb->debug_log(msg);
 		}
 #endif
-		if (!m_socket || !m_socket->is_open()) return; // the operation was aborted
+		if (m_abort) return;
 
 		char buf[16];
 		char* ptr = buf;
 
 		if (m_transaction_id == 0)
-			m_transaction_id = std::rand() ^ (std::rand() << 16);
+			m_transaction_id = random() ^ (random() << 16);
 
 		detail::write_uint32(0x417, ptr);
 		detail::write_uint32(0x27101980, ptr); // connection_id
@@ -390,13 +416,20 @@ namespace libtorrent
 		TORRENT_ASSERT(ptr - buf == sizeof(buf));
 
 		error_code ec;
-		m_socket->send(m_target, buf, 16, ec);
+		if (!m_hostname.empty())
+		{
+			m_ses.m_udp_socket.send_hostname(m_hostname.c_str(), m_target.port(), buf, 16, ec);
+		}
+		else
+		{
+			m_ses.m_udp_socket.send(m_target, buf, 16, ec);
+		}
 		m_state = action_connect;
 		sent_bytes(16 + 28); // assuming UDP/IP header
 		++m_attempts;
 		if (ec)
 		{
-			fail(-1, ec.message().c_str());
+			fail(ec);
 			return;
 		}
 	}
@@ -404,9 +437,9 @@ namespace libtorrent
 	void udp_tracker_connection::send_udp_scrape()
 	{
 		if (m_transaction_id == 0)
-			m_transaction_id = std::rand() ^ (std::rand() << 16);
+			m_transaction_id = random() ^ (random() << 16);
 
-		if (!m_socket || !m_socket->is_open()) return; // the operation was aborted
+		if (m_abort) return;
 
 		std::map<address, connection_cache_entry>::iterator i
 			= m_connection_cache.find(m_target.address());
@@ -426,20 +459,27 @@ namespace libtorrent
 		TORRENT_ASSERT(out - buf == sizeof(buf));
 
 		error_code ec;
-		m_socket->send(m_target, buf, sizeof(buf), ec);
+		if (!m_hostname.empty())
+		{
+			m_ses.m_udp_socket.send_hostname(m_hostname.c_str(), m_target.port(), buf, sizeof(buf), ec);
+		}
+		else
+		{
+			m_ses.m_udp_socket.send(m_target, buf, sizeof(buf), ec);
+		}
 		m_state = action_scrape;
 		sent_bytes(sizeof(buf) + 28); // assuming UDP/IP header
 		++m_attempts;
 		if (ec)
 		{
-			fail(-1, ec.message().c_str());
+			fail(ec);
 			return;
 		}
 	}
 
-	void udp_tracker_connection::on_announce_response(char const* buf, int size)
+	bool udp_tracker_connection::on_announce_response(char const* buf, int size)
 	{
-		if (size < 20) return;
+		if (size < 20) return false;
 
 		buf += 8; // skip header
 		restart_read_timeout();
@@ -450,8 +490,8 @@ namespace libtorrent
 		int num_peers = (size - 20) / 6;
 		if ((size - 20) % 6 != 0)
 		{
-			fail(-1, "invalid udp tracker response length");
-			return;
+			fail(error_code(errors::invalid_tracker_response_length));
+			return false;
 		}
 
 		boost::shared_ptr<request_callback> cb = requester();
@@ -468,13 +508,15 @@ namespace libtorrent
 		if (!cb)
 		{
 			close();
-			return;
+			return true;
 		}
 
 		std::vector<peer_entry> peer_list;
 		for (int i = 0; i < num_peers; ++i)
 		{
-			// TODO: don't use a string here
+			// TODO: don't use a string here. The problem is that
+			// some trackers will respond with actual strings.
+			// Especially i2p trackers
 			peer_entry e;
 			char ip_string[100];
 			unsigned int a = detail::read_uint8(buf);
@@ -489,19 +531,20 @@ namespace libtorrent
 		}
 
 		std::list<address> ip_list;
-		for (std::list<udp::endpoint>::const_iterator i = m_endpoints.begin()
+		for (std::list<tcp::endpoint>::const_iterator i = m_endpoints.begin()
 			, end(m_endpoints.end()); i != end; ++i)
 		{
 			ip_list.push_back(i->address());
 		}
 
 		cb->tracker_response(tracker_req(), m_target.address(), ip_list
-			, peer_list, interval, min_interval, complete, incomplete, address());
+			, peer_list, interval, min_interval, complete, incomplete, address(), "" /*trackerid*/);
 
 		close();
+		return true;
 	}
 
-	void udp_tracker_connection::on_scrape_response(char const* buf, int size)
+	bool udp_tracker_connection::on_scrape_response(char const* buf, int size)
 	{
 		restart_read_timeout();
 		int action = detail::read_int32(buf);
@@ -509,26 +552,26 @@ namespace libtorrent
 
 		if (transaction != m_transaction_id)
 		{
-			fail(-1, "incorrect transaction id");
-			return;
+			fail(error_code(errors::invalid_tracker_transaction_id));
+			return false;
 		}
 
 		if (action == action_error)
 		{
-			fail(-1, std::string(buf, size - 8).c_str());
-			return;
+			fail(error_code(errors::tracker_failure), -1, std::string(buf, size - 8).c_str());
+			return true;
 		}
 
 		if (action != action_scrape)
 		{
-			fail(-1, "invalid action in announce response");
-			return;
+			fail(error_code(errors::invalid_tracker_action));
+			return true;
 		}
 
 		if (size < 20)
 		{
-			fail(-1, "got a message with size < 20");
-			return;
+			fail(error_code(errors::invalid_tracker_response_length));
+			return true;
 		}
 
 		int complete = detail::read_int32(buf);
@@ -539,21 +582,22 @@ namespace libtorrent
 		if (!cb)
 		{
 			close();
-			return;
+			return true;
 		}
 		
 		cb->tracker_scrape_response(tracker_req()
-			, complete, incomplete, downloaded);
+			, complete, incomplete, downloaded, -1);
 
 		close();
+		return true;
 	}
 
 	void udp_tracker_connection::send_udp_announce()
 	{
 		if (m_transaction_id == 0)
-			m_transaction_id = std::rand() ^ (std::rand() << 16);
+			m_transaction_id = random() ^ (random() << 16);
 
-		if (!m_socket || !m_socket->is_open()) return; // the operation was aborted
+		if (m_abort) return;
 
 		char buf[8 + 4 + 4 + 20 + 20 + 8 + 8 + 8 + 4 + 4 + 4 + 4 + 2 + 2];
 		char* out = buf;
@@ -580,10 +624,15 @@ namespace libtorrent
 		detail::write_int64(stats ? req.uploaded : 0, out); // uploaded
 		detail::write_int32(req.event, out); // event
 		// ip address
-		if (settings.announce_ip != address() && settings.announce_ip.is_v4())
-			detail::write_uint32(settings.announce_ip.to_v4().to_ulong(), out);
-		else
-			detail::write_int32(0, out);
+		address_v4 announce_ip;
+
+		if (!settings.announce_ip.empty())
+		{
+			error_code ec;
+			address ip = address::from_string(settings.announce_ip.c_str(), ec);
+			if (!ec && ip.is_v4()) announce_ip = ip.to_v4();
+		}
+		detail::write_uint32(announce_ip.to_ulong(), out);
 		detail::write_int32(req.key, out); // key
 		detail::write_int32(req.num_want, out); // num_want
 		detail::write_uint16(req.listen_port, out); // port
@@ -604,13 +653,20 @@ namespace libtorrent
 #endif
 
 		error_code ec;
-		m_socket->send(m_target, buf, sizeof(buf), ec);
+		if (!m_hostname.empty())
+		{
+			m_ses.m_udp_socket.send_hostname(m_hostname.c_str(), m_target.port(), buf, sizeof(buf), ec);
+		}
+		else
+		{
+			m_ses.m_udp_socket.send(m_target, buf, sizeof(buf), ec);
+		}
 		m_state = action_announce;
 		sent_bytes(sizeof(buf) + 28); // assuming UDP/IP header
 		++m_attempts;
 		if (ec)
 		{
-			fail(-1, ec.message().c_str());
+			fail(ec);
 			return;
 		}
 	}
