@@ -33,7 +33,7 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/pch.hpp"
 
 #include <vector>
-#include <limits>
+#include <boost/limits.hpp>
 #include <boost/bind.hpp>
 #include <stdlib.h>
 
@@ -48,76 +48,65 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/version.hpp"
 #include "libtorrent/aux_/session_impl.hpp"
 #include "libtorrent/parse_url.hpp"
+#include "libtorrent/peer_info.hpp"
 
 using boost::shared_ptr;
 using libtorrent::aux::session_impl;
 
 namespace libtorrent
 {
+	enum
+	{
+		request_size_overhead = 5000
+	};
+
 	web_peer_connection::web_peer_connection(
 		session_impl& ses
 		, boost::weak_ptr<torrent> t
 		, boost::shared_ptr<socket_type> s
 		, tcp::endpoint const& remote
 		, std::string const& url
-		, policy::peer* peerinfo)
-		: peer_connection(ses, t, s, remote, peerinfo)
+		, policy::peer* peerinfo
+		, std::string const& auth
+		, web_seed_entry::headers_t const& extra_headers)
+		: web_connection_base(ses, t, s, remote, url, peerinfo, auth, extra_headers)
 		, m_url(url)
-		, m_original_url(url)
-		, m_first_request(true)
+		, m_received_body(0)
 		, m_range_pos(0)
 		, m_block_pos(0)
+		, m_chunk_pos(0)
+		, m_partial_chunk_header(0)
 	{
 		INVARIANT_CHECK;
 
-		// we want large blocks as well, so
-		// we can request more bytes at once
-		request_large_blocks(true);
+		if (!ses.settings().report_web_seed_downloads)
+			ignore_stats(true);
 
-		// we only want left-over bandwidth
-		set_priority(1);
 		shared_ptr<torrent> tor = t.lock();
 		TORRENT_ASSERT(tor);
-		int blocks_per_piece = tor->torrent_file().piece_length() / tor->block_size();
 
-		// we always prefer downloading 1 MB chunks
-		// from web seeds
-		prefer_whole_pieces((1024 * 1024) / tor->torrent_file().piece_length());
+		// we always prefer downloading 1 MiB chunks
+		// from web seeds, or whole pieces if pieces
+		// are larger than a MiB
+		prefer_whole_pieces((std::min)((1024 * 1024) / tor->torrent_file().piece_length(), 1));
 		
-		// multiply with the blocks per piece since that many requests are
-		// merged into one http request
-		m_max_out_request_queue = ses.settings().urlseed_pipeline_size
-			* blocks_per_piece;
+		// we want large blocks as well, so
+		// we can request more bytes at once
+		// this setting will merge adjacent requests
+		// into single larger ones
+		request_large_blocks(true);
 
-		// since this is a web seed, change the timeout
-		// according to the settings.
-		set_timeout(ses.settings().urlseed_timeout);
 #ifdef TORRENT_VERBOSE_LOGGING
-		(*m_logger) << "*** web_peer_connection\n";
+		peer_log("*** web_peer_connection %s", url.c_str());
 #endif
-
-		std::string protocol;
-		error_code ec;
-		boost::tie(protocol, m_auth, m_host, m_port, m_path)
-			= parse_url_components(url, ec);
-		TORRENT_ASSERT(!ec);
-		
-		if (!m_auth.empty())
-			m_auth = base64encode(m_auth);
-
-		m_server_string = "URL seed @ ";
-		m_server_string += m_host;
 	}
 
-	void web_peer_connection::start()
+	void web_peer_connection::disconnect(error_code const& ec, int error)
 	{
-		set_upload_only(true);
-		if (is_disconnecting()) return;
-		peer_connection::start();
+		boost::shared_ptr<torrent> t = associated_torrent().lock();
+		peer_connection::disconnect(ec, error);
+		if (t) t->disconnect_web_seed(this);
 	}
-
-	web_peer_connection::~web_peer_connection()
-	{}
 	
 	boost::optional<piece_block_progress>
 	web_peer_connection::downloading_piece_progress() const
@@ -137,6 +126,8 @@ namespace libtorrent
 		// would otherwise point to one past the end
 		int correction = m_block_pos ? -1 : 0;
 		ret.block_index = (m_requests.front().start + m_block_pos + correction) / t->block_size();
+		TORRENT_ASSERT(ret.block_index < int(piece_block::invalid.block_index));
+		TORRENT_ASSERT(ret.piece_index < int(piece_block::invalid.piece_index));
 
 		ret.full_block_bytes = t->block_size();
 		const int last_piece = t->torrent_file().num_pieces() - 1;
@@ -146,20 +137,6 @@ namespace libtorrent
 		return ret;
 	}
 
-	void web_peer_connection::on_connected()
-	{
-		boost::shared_ptr<torrent> t = associated_torrent().lock();
-		TORRENT_ASSERT(t);
-	
-		// this is always a seed
-		incoming_have_all();
-
-		// it is always possible to request pieces
-		incoming_unchoke();
-
-		reset_recv_buffer(t->block_size() + 1024);
-	}
-
 	void web_peer_connection::write_request(peer_request const& r)
 	{
 		INVARIANT_CHECK;
@@ -184,7 +161,7 @@ namespace libtorrent
 			if (m_path.empty()) m_path += "/" + t->torrent_file().name();
 			else if (m_path[m_path.size() - 1] == '/')
 			{
-				std::string tmp = t->torrent_file().files().at(0).path.string();
+				std::string tmp = t->torrent_file().files().at(0).path;
 #ifdef TORRENT_WINDOWS
 				convert_path_to_posix(tmp);
 #endif
@@ -192,7 +169,7 @@ namespace libtorrent
 			}
 			else if (!m_url.empty() && m_url[m_url.size() - 1] == '/')
 			{
-				std::string tmp = t->torrent_file().files().at(0).path.string();
+				std::string tmp = t->torrent_file().files().at(0).path;
 #ifdef TORRENT_WINDOWS
 				convert_path_to_posix(tmp);
 #endif
@@ -219,9 +196,9 @@ namespace libtorrent
 			size -= pr.length;
 		}
 
-		proxy_settings const& ps = m_ses.web_seed_proxy();
-		bool using_proxy = ps.type == proxy_settings::http
-			|| ps.type == proxy_settings::http_pw;
+		proxy_settings const& ps = m_ses.proxy();
+		bool using_proxy = (ps.type == proxy_settings::http
+			|| ps.type == proxy_settings::http_pw) && !m_ssl;
 
 		if (single_file_request)
 		{
@@ -230,33 +207,11 @@ namespace libtorrent
 			// assumed to be encoded in the torrent file
 			request += using_proxy ? m_url : m_path;
 			request += " HTTP/1.1\r\n";
-			request += "Host: ";
-			request += m_host;
-			if (m_first_request)
-			{
-				request += "\r\nUser-Agent: ";
-				request += m_ses.settings().user_agent;
-			}
-			if (!m_auth.empty())
-			{
-				request += "\r\nAuthorization: Basic ";
-				request += m_auth;
-			}
-			if (ps.type == proxy_settings::http_pw)
-			{
-				request += "\r\nProxy-Authorization: Basic ";
-				request += base64encode(ps.username + ":" + ps.password);
-			}
-			if (using_proxy)
-			{
-				request += "\r\nProxy-Connection: keep-alive";
-			}
+			add_headers(request, ps, using_proxy);
 			request += "\r\nRange: bytes=";
 			request += to_string(size_type(r.piece) * info.piece_length() + r.start).elems;
 			request += "-";
 			request += to_string(size_type(r.piece) * info.piece_length() + r.start + r.length - 1).elems;
-			if (m_first_request || using_proxy)
-				request += "\r\nConnection: keep-alive";
 			request += "\r\n\r\n";
 			m_first_request = false;
 			m_file_requests.push_back(0);
@@ -270,48 +225,37 @@ namespace libtorrent
 				i != files.end(); ++i)
 			{
 				file_slice const& f = *i;
+				if (info.orig_files().internal_at(f.file_index).pad_file)
+				{
+					m_file_requests.push_back(f.file_index);
+					continue;
+				}
 
 				request += "GET ";
 				if (using_proxy)
 				{
 					request += m_url;
-					std::string path = info.orig_files().at(f.file_index).path.string();
+					std::string path = info.orig_files().file_path(info.orig_files().internal_at(f.file_index));
+#ifdef TORRENT_WINDOWS
+					convert_path_to_posix(path);
+#endif
 					request += escape_path(path.c_str(), path.length());
 				}
 				else
 				{
 					std::string path = m_path;
-					path += info.orig_files().at(f.file_index).path.string();
+					path += info.orig_files().file_path(info.orig_files().internal_at(f.file_index));
+#ifdef TORRENT_WINDOWS
+					convert_path_to_posix(path);
+#endif
 					request += escape_path(path.c_str(), path.length());
 				}
 				request += " HTTP/1.1\r\n";
-				request += "Host: ";
-				request += m_host;
-				if (m_first_request)
-				{
-					request += "\r\nUser-Agent: ";
-					request += m_ses.settings().user_agent;
-				}
-				if (!m_auth.empty())
-				{
-					request += "\r\nAuthorization: Basic ";
-					request += m_auth;
-				}
-				if (ps.type == proxy_settings::http_pw)
-				{
-					request += "\r\nProxy-Authorization: Basic ";
-					request += base64encode(ps.username + ":" + ps.password);
-				}
-				if (using_proxy)
-				{
-					request += "\r\nProxy-Connection: keep-alive";
-				}
+				add_headers(request, ps, using_proxy);
 				request += "\r\nRange: bytes=";
 				request += to_string(f.offset).elems;
 				request += "-";
 				request += to_string(f.offset + f.size - 1).elems;
-				if (m_first_request || using_proxy)
-					request += "\r\nConnection: keep-alive";
 				request += "\r\n\r\n";
 				m_first_request = false;
 				TORRENT_ASSERT(f.file_index >= 0);
@@ -320,7 +264,7 @@ namespace libtorrent
 		}
 
 #ifdef TORRENT_VERBOSE_LOGGING
-		(*m_logger) << request << "\n";
+		peer_log("==> %s", request.c_str());
 #endif
 
 		send_buffer(request.c_str(), request.size(), message_type_request);
@@ -341,17 +285,58 @@ namespace libtorrent
 		}
 	}
 
+	bool web_peer_connection::maybe_harvest_block()
+	{
+		peer_request const& front_request = m_requests.front();
+
+		if (int(m_piece.size()) < front_request.length) return false;
+		TORRENT_ASSERT(int(m_piece.size() == front_request.length));
+
+		// each call to incoming_piece() may result in us becoming
+		// a seed. If we become a seed, all seeds we're connected to
+		// will be disconnected, including this web seed. We need to
+		// check for the disconnect condition after the call.
+
+		boost::shared_ptr<torrent> t = associated_torrent().lock();
+		TORRENT_ASSERT(t);
+		buffer::const_interval recv_buffer = receive_buffer();
+
+		incoming_piece(front_request, &m_piece[0]);
+		m_requests.pop_front();
+		if (associated_torrent().expired()) return false;
+		TORRENT_ASSERT(m_block_pos >= front_request.length);
+		m_block_pos -= front_request.length;
+		cut_receive_buffer(m_body_start, t->block_size() + request_size_overhead);
+		m_body_start = 0;
+		recv_buffer = receive_buffer();
+//		TORRENT_ASSERT(m_received_body <= range_end - range_start);
+		m_piece.clear();
+		TORRENT_ASSERT(m_piece.empty());
+		return true;
+	}
+
 	void web_peer_connection::on_receive(error_code const& error
 		, std::size_t bytes_transferred)
 	{
 		INVARIANT_CHECK;
 
+#ifdef TORRENT_DEBUG
+		TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+			+ m_statistics.last_protocol_downloaded() + bytes_transferred < size_t(INT_MAX));
+		int dl_target = m_statistics.last_payload_downloaded()
+			+ m_statistics.last_protocol_downloaded() + bytes_transferred;
+#endif
+
 		if (error)
 		{
 			m_statistics.received_bytes(0, bytes_transferred);
 #ifdef TORRENT_VERBOSE_LOGGING
-			(*m_logger) << "*** web_peer_connection error: "
-				<< error.message() << "\n";
+			peer_log("*** web_peer_connection error: %s", error.message().c_str());
+#endif
+#ifdef TORRENT_DEBUG
+			TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+				+ m_statistics.last_protocol_downloaded()
+				== dl_target);
 #endif
 			return;
 		}
@@ -361,6 +346,12 @@ namespace libtorrent
 
 		for (;;)
 		{
+#ifdef TORRENT_DEBUG
+			TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+				+ m_statistics.last_protocol_downloaded() + int(bytes_transferred)
+				== dl_target);
+#endif
+
 			buffer::const_interval recv_buffer = receive_buffer();
 
 			int payload;
@@ -368,18 +359,24 @@ namespace libtorrent
 			bool header_finished = m_parser.header_finished();
 			if (!header_finished)
 			{
-				bool error = false;
-				boost::tie(payload, protocol) = m_parser.incoming(recv_buffer, error);
+				bool failed = false;
+				boost::tie(payload, protocol) = m_parser.incoming(recv_buffer, failed);
 				m_statistics.received_bytes(0, protocol);
+				TORRENT_ASSERT(int(bytes_transferred) >= protocol);
 				bytes_transferred -= protocol;
 
-				if (error)
+				if (failed)
 				{
 					m_statistics.received_bytes(0, bytes_transferred);
 #ifdef TORRENT_VERBOSE_LOGGING
-					(*m_logger) << "*** " << std::string(recv_buffer.begin, recv_buffer.end) << "\n";
+					peer_log("*** %s", std::string(recv_buffer.begin, recv_buffer.end).c_str());
 #endif
 					disconnect(errors::http_parse_error, 2);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded()
+						== dl_target);
+#endif
 					return;
 				}
 
@@ -392,30 +389,23 @@ namespace libtorrent
 				{
 					TORRENT_ASSERT(payload == 0);
 					TORRENT_ASSERT(bytes_transferred == 0);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded() + int(bytes_transferred)
+						== dl_target);
+#endif
 					break;
 				}
 
-				// if the status code is not one of the accepted ones, abort
-				if (!is_ok_status(m_parser.status_code()))
-				{
-					// temporarily unavailable, retry later
-					t->retry_web_seed(m_original_url, web_seed_entry::url_seed);
-					t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
-					std::string error_msg = to_string(m_parser.status_code()).elems
-						+ (" " + m_parser.message());
-					if (m_ses.m_alerts.should_post<url_seed_alert>())
-					{
-						m_ses.m_alerts.post_alert(url_seed_alert(t->get_handle(), url()
-							, error_msg));
-					}
-					m_statistics.received_bytes(0, bytes_transferred);
-					disconnect(errors::http_error, 1);
-					return;
-				}
 				if (!m_parser.header_finished())
 				{
 					TORRENT_ASSERT(payload == 0);
 					TORRENT_ASSERT(bytes_transferred == 0);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded() + int(bytes_transferred)
+						== dl_target);
+#endif
 					break;
 				}
 
@@ -427,13 +417,35 @@ namespace libtorrent
 			if (!header_finished)
 			{
 #ifdef TORRENT_VERBOSE_LOGGING
-				(*m_logger) << "*** STATUS: " << m_parser.status_code()
-					<< " " << m_parser.message() << "\n";
-				std::map<std::string, std::string> const& headers = m_parser.headers();
-				for (std::map<std::string, std::string>::const_iterator i = headers.begin()
+				peer_log("*** STATUS: %d %s", m_parser.status_code(), m_parser.message().c_str());
+				std::multimap<std::string, std::string> const& headers = m_parser.headers();
+				for (std::multimap<std::string, std::string>::const_iterator i = headers.begin()
 					, end(headers.end()); i != end; ++i)
-					(*m_logger) << "   " << i->first << ": " << i->second << "\n";
+					peer_log("   %s: %s", i->first.c_str(), i->second.c_str());
 #endif
+				// if the status code is not one of the accepted ones, abort
+				if (!is_ok_status(m_parser.status_code()))
+				{
+					int retry_time = atoi(m_parser.header("retry-after").c_str());
+					if (retry_time <= 0) retry_time = 5 * 60;
+					// temporarily unavailable, retry later
+					t->retry_web_seed(this, retry_time);
+					std::string error_msg = to_string(m_parser.status_code()).elems
+						+ (" " + m_parser.message());
+					if (m_ses.m_alerts.should_post<url_seed_alert>())
+					{
+						m_ses.m_alerts.post_alert(url_seed_alert(t->get_handle(), m_url
+							, error_msg));
+					}
+					m_statistics.received_bytes(0, bytes_transferred);
+					disconnect(error_code(m_parser.status_code(), get_http_category()), 1);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded()
+						== dl_target);
+#endif
+					return;
+				}
 				if (is_redirect(m_parser.status_code()))
 				{
 					// this means we got a redirection request
@@ -444,8 +456,13 @@ namespace libtorrent
 					if (location.empty())
 					{
 						// we should not try this server again.
-						t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
+						t->remove_web_seed(this);
 						disconnect(errors::missing_location, 2);
+#ifdef TORRENT_DEBUG
+						TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+							+ m_statistics.last_protocol_downloaded()
+							== dl_target);
+#endif
 						return;
 					}
 					
@@ -456,25 +473,37 @@ namespace libtorrent
 					// add the redirected url and remove the current one
 					if (!single_file_request)
 					{
-						m_statistics.received_bytes(0, bytes_transferred);
 						TORRENT_ASSERT(!m_file_requests.empty());
 						int file_index = m_file_requests.front();
 
 						torrent_info const& info = t->torrent_file();
-						std::string path = info.orig_files().at(file_index).path.string();
+						std::string path = info.orig_files().file_path(info.orig_files().at(file_index));
+#ifdef TORRENT_WINDOWS
+						convert_path_to_posix(path);
+#endif
 						path = escape_path(path.c_str(), path.length());
 						size_t i = location.rfind(path);
 						if (i == std::string::npos)
 						{
-							t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
+							t->remove_web_seed(this);
 							disconnect(errors::invalid_redirection, 2);
+#ifdef TORRENT_DEBUG
+							TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+								+ m_statistics.last_protocol_downloaded()
+								== dl_target);
+#endif
 							return;
 						}
 						location.resize(i);
 					}
-					t->add_web_seed(location, web_seed_entry::url_seed);
-					t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
+					t->add_web_seed(location, web_seed_entry::url_seed, m_external_auth, m_extra_headers);
+					t->remove_web_seed(this);
 					disconnect(errors::redirecting, 2);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded()
+						== dl_target);
+#endif
 					return;
 				}
 
@@ -496,7 +525,15 @@ namespace libtorrent
 			recv_buffer.begin += m_body_start;
 
 			// we only received the header, no data
-			if (recv_buffer.left() == 0) break;
+			if (recv_buffer.left() == 0)
+			{
+#ifdef TORRENT_DEBUG
+				TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+					+ m_statistics.last_protocol_downloaded()
+					== dl_target);
+#endif
+				break;
+			}
 
 			size_type range_start;
 			size_type range_end;
@@ -507,8 +544,13 @@ namespace libtorrent
 				{
 					m_statistics.received_bytes(0, bytes_transferred);
 					// we should not try this server again.
-					t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
+					t->remove_web_seed(this);
 					disconnect(errors::invalid_range);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded()
+						== dl_target);
+#endif
 					return;
 				}
 				// the http range is inclusive
@@ -522,8 +564,13 @@ namespace libtorrent
 				{
 					m_statistics.received_bytes(0, bytes_transferred);
 					// we should not try this server again.
-					t->remove_web_seed(m_original_url, web_seed_entry::url_seed);
+					t->remove_web_seed(this);
 					disconnect(errors::no_content_length, 2);
+#ifdef TORRENT_DEBUG
+					TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+						+ m_statistics.last_protocol_downloaded()
+						== dl_target);
+#endif
 					return;
 				}
 			}
@@ -532,11 +579,66 @@ namespace libtorrent
 			{
 				m_statistics.received_bytes(0, bytes_transferred);
 				disconnect(errors::http_error, 2);
+#ifdef TORRENT_DEBUG
+				TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+					+ m_statistics.last_protocol_downloaded()
+					== dl_target);
+#endif
 				return;
 			}
 
-			int left_in_response = range_end - range_start - m_range_pos;
-			int payload_transferred = (std::min)(left_in_response, int(bytes_transferred));
+			// =========================
+			// === CHUNKED ENCODING  ===
+			// =========================
+			while (m_parser.chunked_encoding()
+				&& m_chunk_pos >= 0
+				&& m_chunk_pos < recv_buffer.left())
+			{
+				int header_size = 0;
+				size_type chunk_size = 0;
+				buffer::const_interval chunk_start = recv_buffer;
+				chunk_start.begin += m_chunk_pos;
+				TORRENT_ASSERT(chunk_start.begin[0] == '\r' || is_hex(chunk_start.begin, 1));
+				bool ret = m_parser.parse_chunk_header(chunk_start, &chunk_size, &header_size);
+				if (!ret)
+				{
+					TORRENT_ASSERT(int(bytes_transferred) >= chunk_start.left() - m_partial_chunk_header);
+					bytes_transferred -= chunk_start.left() - m_partial_chunk_header;
+					m_statistics.received_bytes(0, chunk_start.left() - m_partial_chunk_header);
+					m_partial_chunk_header = chunk_start.left();
+					if (bytes_transferred == 0) return;
+					break;
+				}
+				else
+				{
+#ifdef TORRENT_VERBOSE_LOGGING
+					peer_log("*** parsed chunk: %d header_size: %d", chunk_size, header_size);
+#endif
+					TORRENT_ASSERT(int(bytes_transferred) >= header_size - m_partial_chunk_header);
+					bytes_transferred -= header_size - m_partial_chunk_header;
+					m_statistics.received_bytes(0, header_size - m_partial_chunk_header);
+					m_partial_chunk_header = 0;
+					TORRENT_ASSERT(chunk_size != 0 || chunk_start.left() <= header_size || chunk_start.begin[header_size] == 'H');
+					// cut out the chunk header from the receive buffer
+					TORRENT_ASSERT(m_body_start + m_chunk_pos < INT_MAX);
+					cut_receive_buffer(header_size, t->block_size() + request_size_overhead, int(m_body_start + m_chunk_pos));
+					recv_buffer = receive_buffer();
+					recv_buffer.begin += m_body_start;
+					m_chunk_pos += chunk_size;
+					if (chunk_size == 0)
+					{
+#ifdef TORRENT_DEBUG
+						chunk_start = recv_buffer;
+						chunk_start.begin += m_chunk_pos;
+						TORRENT_ASSERT(chunk_start.left() == 0 || chunk_start.begin[0] == 'H');
+#endif
+						m_chunk_pos = -1;
+					}
+				}
+			}
+
+			size_type left_in_response = range_end - range_start - m_range_pos;
+			int payload_transferred = int((std::min)(left_in_response, size_type(bytes_transferred)));
 
 			torrent_info const& info = t->torrent_file();
 
@@ -545,11 +647,12 @@ namespace libtorrent
 			TORRENT_ASSERT(m_block_pos >= 0);
 
 #ifdef TORRENT_VERBOSE_LOGGING
-			(*m_logger) << "*** payload_transferred: " << payload_transferred
-				<< " [" << front_request.piece << ":" << front_request.start
-				<< " = " << front_request.length <<  "]\n";
+			peer_log("*** payload_transferred: %d [ %d:%d = %d ]"
+				, payload_transferred, front_request.piece
+				, front_request.start, front_request.length);
 #endif
 			m_statistics.received_bytes(payload_transferred, 0);
+			TORRENT_ASSERT(int(bytes_transferred) >= payload_transferred);
 			bytes_transferred -= payload_transferred;
 			m_range_pos += payload_transferred;
 			m_block_pos += payload_transferred;
@@ -564,8 +667,11 @@ namespace libtorrent
 			peer_request in_range = info.orig_files().map_file(file_index, range_start
 				, int(range_end - range_start));
 
+			// request start
 			size_type rs = size_type(in_range.piece) * info.piece_length() + in_range.start;
+			// request end
 			size_type re = rs + in_range.length;
+			// file start
 			size_type fs = size_type(front_request.piece) * info.piece_length() + front_request.start;
 #if 0
 			size_type fe = fs + front_request.length;
@@ -609,35 +715,30 @@ namespace libtorrent
 				int piece_size = int(m_piece.size());
 				int copy_size = (std::min)((std::min)(front_request.length - piece_size
 					, recv_buffer.left()), int(range_end - range_start - m_received_body));
-				m_piece.resize(piece_size + copy_size);
-				TORRENT_ASSERT(copy_size > 0);
-				std::memcpy(&m_piece[0] + piece_size, recv_buffer.begin, copy_size);
-				TORRENT_ASSERT(int(m_piece.size()) <= front_request.length);
-				recv_buffer.begin += copy_size;
-				m_received_body += copy_size;
-				m_body_start += copy_size;
-				TORRENT_ASSERT(m_received_body <= range_end - range_start);
-				TORRENT_ASSERT(int(m_piece.size()) <= front_request.length);
-				incoming_piece_fragment(copy_size);
-				if (int(m_piece.size()) == front_request.length)
+				if (copy_size > m_chunk_pos && m_chunk_pos > 0) copy_size = m_chunk_pos;
+				if (copy_size > 0)
 				{
-					// each call to incoming_piece() may result in us becoming
-					// a seed. If we become a seed, all seeds we're connected to
-					// will be disconnected, including this web seed. We need to
-					// check for the disconnect condition after the call.
-
-					incoming_piece(front_request, &m_piece[0]);
-					m_requests.pop_front();
-					if (associated_torrent().expired()) return;
-					TORRENT_ASSERT(m_block_pos >= front_request.length);
-					m_block_pos -= front_request.length;
-					cut_receive_buffer(m_body_start, t->block_size() + 1024);
-					m_body_start = 0;
-					recv_buffer = receive_buffer();
+					TORRENT_ASSERT(m_piece.size() == m_received_in_piece);
+					m_piece.resize(piece_size + copy_size);
+					std::memcpy(&m_piece[0] + piece_size, recv_buffer.begin, copy_size);
+					TORRENT_ASSERT(int(m_piece.size()) <= front_request.length);
+					recv_buffer.begin += copy_size;
+					m_received_body += copy_size;
+					m_body_start += copy_size;
+					if (m_chunk_pos > 0)
+					{
+						TORRENT_ASSERT(m_chunk_pos >= copy_size);
+						m_chunk_pos -= copy_size;
+					}
 					TORRENT_ASSERT(m_received_body <= range_end - range_start);
-					m_piece.clear();
-					TORRENT_ASSERT(m_piece.empty());
+					TORRENT_ASSERT(int(m_piece.size()) <= front_request.length);
+					incoming_piece_fragment(copy_size);
+					TORRENT_ASSERT(m_piece.size() == m_received_in_piece);
 				}
+
+				if (maybe_harvest_block())
+					recv_buffer = receive_buffer();
+				if (associated_torrent().expired()) return;
 			}
 
 			// report all received blocks to the bittorrent engine
@@ -652,11 +753,17 @@ namespace libtorrent
 				incoming_piece(r, recv_buffer.begin);
 				m_requests.pop_front();
 				if (associated_torrent().expired()) return;
+				TORRENT_ASSERT(m_block_pos >= r.length);
 				m_block_pos -= r.length;
 				m_received_body += r.length;
 				TORRENT_ASSERT(receive_buffer().begin + m_body_start == recv_buffer.begin);
 				TORRENT_ASSERT(m_received_body <= range_end - range_start);
-				cut_receive_buffer(r.length + m_body_start, t->block_size() + 1024);
+				cut_receive_buffer(m_body_start + r.length, t->block_size() + request_size_overhead);
+				if (m_chunk_pos > 0)
+				{
+					TORRENT_ASSERT(m_chunk_pos >= r.length);
+					m_chunk_pos -= r.length;
+				}
 				m_body_start = 0;
 				recv_buffer = receive_buffer();
 			}
@@ -675,75 +782,95 @@ namespace libtorrent
 					TORRENT_ASSERT(copy_size >= 0);
 					if (copy_size > 0)
 					{
+						TORRENT_ASSERT(m_piece.size() == m_received_in_piece);
 						m_piece.resize(piece_size + copy_size);
 						std::memcpy(&m_piece[0] + piece_size, recv_buffer.begin, copy_size);
 						recv_buffer.begin += copy_size;
 						m_received_body += copy_size;
 						m_body_start += copy_size;
+						incoming_piece_fragment(copy_size);
+						TORRENT_ASSERT(m_piece.size() == m_received_in_piece);
 					}
 					TORRENT_ASSERT(m_received_body == range_end - range_start);
 				}
 			}
 
 			TORRENT_ASSERT(m_received_body <= range_end - range_start);
-			if (m_received_body == range_end - range_start)
+			// if we're in chunked encoding mode, we have to wait for the complete
+			// tail header before we can consider have received the block, otherwise
+			// we'll get out of sync with the next http response. m_chunk_pos is set
+			// to -1 when the tail header has been received
+			if (m_received_body == range_end - range_start
+				&& (!m_parser.chunked_encoding() || m_chunk_pos == -1))
 			{
-				cut_receive_buffer(recv_buffer.begin - receive_buffer().begin
-					, t->block_size() + 1024);
+				int size_to_cut = recv_buffer.begin - receive_buffer().begin;
+
+				TORRENT_ASSERT(receive_buffer().left() < size_to_cut + 1
+					|| receive_buffer()[size_to_cut] == 'H');
+
+				cut_receive_buffer(size_to_cut, t->block_size() + request_size_overhead);
+				if (m_chunk_pos > 0)
+				{
+					TORRENT_ASSERT(m_chunk_pos >= size_to_cut);
+					m_chunk_pos -= size_to_cut;
+				}
 				recv_buffer = receive_buffer();
 				m_file_requests.pop_front();
 				m_parser.reset();
 				m_body_start = 0;
 				m_received_body = 0;
+				m_chunk_pos = 0;
+				m_partial_chunk_header = 0;
+				
+				torrent_info const& info = t->torrent_file();
+				while (!m_file_requests.empty()
+					&& info.orig_files().internal_at(m_file_requests.front()).pad_file)
+				{
+					// the next file is a pad file. We didn't actually send
+					// a request for this since it most likely doesn't exist on
+					// the web server anyway. Just pretend that we received a
+					// bunch of zeroes here and pop it again
+					int file_index = m_file_requests.front();
+					m_file_requests.pop_front();
+					size_type file_size = info.orig_files().file_size(info.orig_files().internal_at(file_index));
+					TORRENT_ASSERT(m_block_pos < front_request.length);
+					int pad_size = (std::min)(file_size, size_type(front_request.length - m_block_pos));
+
+					// insert zeroes to represent the pad file
+					m_piece.resize(m_piece.size() + pad_size, 0);
+					m_block_pos += pad_size;
+					incoming_piece_fragment(pad_size);
+
+					if (maybe_harvest_block())
+						recv_buffer = receive_buffer();
+					if (associated_torrent().expired()) return;
+				}
 				continue;
 			}
-			if (bytes_transferred == 0) break;
+			if (bytes_transferred == 0 || payload_transferred == 0)
+			{
+#ifdef TORRENT_DEBUG
+				TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+					+ m_statistics.last_protocol_downloaded()
+					== dl_target);
+#endif
+				break;
+			}
 			TORRENT_ASSERT(payload_transferred > 0);
 		}
 		TORRENT_ASSERT(bytes_transferred == 0);
+#ifdef TORRENT_DEBUG
+		TORRENT_ASSERT(m_statistics.last_payload_downloaded()
+			+ m_statistics.last_protocol_downloaded() == dl_target);
+#endif
 	}
 
 	void web_peer_connection::get_specific_peer_info(peer_info& p) const
 	{
-		if (is_interesting()) p.flags |= peer_info::interesting;
-		if (is_choked()) p.flags |= peer_info::choked;
-		if (is_peer_interested()) p.flags |= peer_info::remote_interested;
-		if (has_peer_choked()) p.flags |= peer_info::remote_choked;
-		if (is_local()) p.flags |= peer_info::local_connection;
-		if (!is_connecting() && m_server_string.empty())
-			p.flags |= peer_info::handshake;
-		if (is_connecting() && !is_queued()) p.flags |= peer_info::connecting;
-		if (is_queued()) p.flags |= peer_info::queued;
-
-		p.client = m_server_string;
+		web_connection_base::get_specific_peer_info(p);
+		p.flags |= peer_info::local_connection;
 		p.connection_type = peer_info::web_seed;
 	}
 
-	bool web_peer_connection::in_handshake() const
-	{
-		return m_server_string.empty();
-	}
-
-	void web_peer_connection::on_sent(error_code const& error
-		, std::size_t bytes_transferred)
-	{
-		INVARIANT_CHECK;
-
-		if (error) return;
-		m_statistics.sent_bytes(0, bytes_transferred);
-	}
-
-
-#ifdef TORRENT_DEBUG
-	void web_peer_connection::check_invariant() const
-	{
-/*
-		TORRENT_ASSERT(m_num_pieces == std::count(
-			m_have_piece.begin()
-			, m_have_piece.end()
-			, true));
-*/	}
-#endif
-
 }
 
