@@ -36,20 +36,19 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "libtorrent/storage.hpp"
 #include "libtorrent/bencode.hpp"
 #include "libtorrent/create_torrent.hpp"
-#include <boost/thread.hpp>
+#include "libtorrent/thread.hpp"
 #include <boost/tuple/tuple.hpp>
-#include <boost/filesystem/operations.hpp>
 #include <fstream>
+#include <iostream>
 
 #include "test.hpp"
 #include "setup_transfer.hpp"
 
-using namespace boost::filesystem;
 using namespace libtorrent;
 
 // proxy: 0=none, 1=socks4, 2=socks5, 3=socks5_pw 4=http 5=http_pw
 void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
-	, int proxy, int port, char const* protocol, bool url_seed)
+	, int proxy, int port, char const* protocol, bool url_seed, bool chunked_encoding, bool test_ban)
 {
 	using namespace libtorrent;
 
@@ -57,14 +56,17 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 	session_settings settings;
 	settings.max_queued_disk_bytes = 256 * 1024;
 	ses.set_settings(settings);
-	ses.set_alert_mask(~alert::progress_notification);
-	ses.listen_on(std::make_pair(51000, 52000));
-	remove_all("./tmp2_web_seed");
+	ses.set_alert_mask(~(alert::progress_notification | alert::stats_notification));
+	error_code ec;
+	ses.listen_on(std::make_pair(51000, 52000), ec);
+	if (ec) fprintf(stderr, "listen_on failed: %s\n", ec.message().c_str());
+
+	remove_all("tmp2_web_seed", ec);
 
 	char const* test_name[] = {"no", "SOCKS4", "SOCKS5", "SOCKS5 password", "HTTP", "HTTP password"};
 
-	fprintf(stderr, "\n\n  ==== TESTING %s proxy ==== %s ==== %s ===\n\n\n"
-		, test_name[proxy], protocol, url_seed ? "URL seed" : "HTTP seed");
+	fprintf(stderr, "\n\n  ==== TESTING === proxy: %s ==== protocol: %s ==== seed: %s === transfer-encoding: %s === corruption: %s\n\n\n"
+		, test_name[proxy], protocol, url_seed ? "URL seed" : "HTTP seed", chunked_encoding ? "chunked": "none", test_ban ? "yes" : "no");
 	
 	if (proxy)
 	{
@@ -75,14 +77,16 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 		ps.username = "testuser";
 		ps.password = "testpass";
 		ps.type = (proxy_settings::proxy_type)proxy;
-		ses.set_web_seed_proxy(ps);
+		ses.set_proxy(ps);
 	}
 
 	add_torrent_params p;
+	p.flags &= ~add_torrent_params::flag_paused;
+	p.flags &= ~add_torrent_params::flag_auto_managed;
 	p.ti = torrent_file;
-	p.save_path = "./tmp2_web_seed";
+	p.save_path = "tmp2_web_seed";
 	p.storage_mode = storage_mode_compact;
-	torrent_handle th = ses.add_torrent(p);
+	torrent_handle th = ses.add_torrent(p, ec);
 
 	std::vector<announce_entry> empty;
 	th.replace_trackers(empty);
@@ -94,6 +98,14 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 
 	cache_status cs;
 
+	file_storage const& fs = torrent_file->files();
+	int pad_file_size = 0;
+	for (int i = 0; i < fs.num_files(); ++i)
+	{
+		file_entry f = fs.at(i);
+		if (f.pad_file) pad_file_size += f.size;
+	}
+
 	for (int i = 0; i < 30; ++i)
 	{
 		torrent_status s = th.status();
@@ -104,7 +116,7 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 		cs = ses.get_cache_status();
 		if (cs.blocks_read < 1) cs.blocks_read = 1;
 		if (cs.blocks_written < 1) cs.blocks_written = 1;
-
+/*
 		std::cerr << (s.progress * 100.f) << " %"
 			<< " torrent rate: " << (s.download_rate / 1000.f) << " kB/s"
 			<< " session rate: " << (ss.download_rate / 1000.f) << " kB/s"
@@ -115,24 +127,33 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 			<< " rcache: " << cs.read_cache_size
 			<< " buffers: " << cs.total_used_buffers
 			<< std::endl;
+*/
+		print_alerts(ses, "  >>  ses", test_ban, false, false, 0, true);
 
-		print_alerts(ses, "  >>  ses");
-
-		if (th.is_seed()/* && ss.download_rate == 0.f*/)
+		if (test_ban && th.url_seeds().empty())
 		{
-			torrent_status st = th.status();
-			TEST_CHECK(st.total_payload_download - st.total_redundant_bytes == total_size);
+			// when we don't have any web seeds left, we know we successfully banned it
+			break;
+		}
+
+		if (s.is_seeding /* && ss.download_rate == 0.f*/)
+		{
+			TEST_EQUAL(s.total_payload_download - s.total_redundant_bytes, total_size - pad_file_size);
 			// we need to sleep here a bit to let the session sync with the torrent stats
 			test_sleep(1000);
-			TEST_CHECK(ses.status().total_payload_download - ses.status().total_redundant_bytes
-				== total_size);
+			TEST_EQUAL(ses.status().total_payload_download - ses.status().total_redundant_bytes
+				, total_size - pad_file_size);
 			break;
 		}
 		test_sleep(500);
 	}
 
-	TEST_CHECK(cs.cache_size == 0);
-	TEST_CHECK(cs.total_used_buffers == 0);
+	// for test_ban tests, make sure we removed
+	// the url seed (i.e. banned it)
+	TEST_CHECK(!test_ban || th.url_seeds().empty());
+
+	TEST_EQUAL(cs.cache_size, 0);
+	TEST_EQUAL(cs.total_used_buffers, 0);
 
 	std::cerr << "total_size: " << total_size
 		<< " rate_sum: " << rate_sum
@@ -146,12 +167,15 @@ void test_transfer(boost::intrusive_ptr<torrent_info> torrent_file
 //	TEST_CHECK(fabs(rate_sum - total_size) < total_size * .1f);
 //	TEST_CHECK(fabs(ses_rate_sum - total_size) < total_size * .1f);
 
-	TEST_CHECK(th.is_seed());
+	// if test_ban is true, we're not supposed to have completed the download
+	// otherwise, we are supposed to have
+	TEST_CHECK(th.status().is_seeding == !test_ban);
 
 	if (proxy) stop_proxy(8002);
 
-	TEST_CHECK(exists("./tmp2_web_seed" / torrent_file->file_at(0).path));
-	remove_all("./tmp2_web_seed");
+	TEST_CHECK(exists(combine_path("tmp2_web_seed", torrent_file->files().file_path(
+		torrent_file->file_at(0)))) || test_ban);
+	remove_all("tmp2_web_seed", ec);
 }
 
 void save_file(char const* filename, char const* data, int size)
@@ -175,55 +199,72 @@ void save_file(char const* filename, char const* data, int size)
 
 }
 
+sha1_hash file_hash(std::string const& name)
+{
+	std::vector<char> buf;
+	error_code ec;
+	load_file(name, buf, ec);
+	if (buf.empty()) return sha1_hash(0);
+	hasher h(&buf[0], buf.size());
+	return h.final();
+}
+
 // test_url_seed determines whether to use url-seed or http-seed
-int run_suite(char const* protocol, bool test_url_seed)
+int run_suite(char const* protocol, bool test_url_seed, bool chunked_encoding, bool test_ban)
 {
 	using namespace libtorrent;
-	using namespace boost::filesystem;
 
-	try {
-		create_directory("./tmp1_web_seed");
-	} catch (std::exception&) {}
-
-	try {
-		create_directory("./tmp1_web_seed/test_torrent_dir");
-	} catch (std::exception&) {}
+	error_code ec;
+	create_directories("tmp1_web_seed/test_torrent_dir", ec);
 
 	file_storage fs;
-	int piece_size = 16;
+	std::srand(10);
+	int piece_size = 0x4000;
+	static const int file_sizes[] =
+	{ 5, 16 - 5, 16000, 17, 10, 8000, 8000, 1,1,1,1,1,100,1,1,1,1,100,1,1,1,1,1,1
+		,1,1,1,1,1,1,13,65000,34,75,2,30,400,500,23000,900,43000,400,4300,6, 4};
+
 	if (test_url_seed)
 	{
-		int file_sizes[] =
-		{ 5, 16 - 5, 16, 17, 10, 30, 30, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
-			,1,1,1,1,1,1,13,65,34,75,2,3,4,5,23,9,43,4,43,6, 4};
-
 		char* random_data = (char*)malloc(300000);
-		std::srand(10);
 		for (int i = 0; i != sizeof(file_sizes)/sizeof(file_sizes[0]); ++i)
 		{
 			std::generate(random_data, random_data + 300000, &std::rand);
 			char filename[200];
-			snprintf(filename, sizeof(filename), "./tmp1_web_seed/test_torrent_dir/test%d", i);
-			save_file(filename, random_data, file_sizes[i]);
+			snprintf(filename, sizeof(filename), "tmp1_web_seed/test_torrent_dir/test%d", i);
+			int to_write = file_sizes[i];
+			file f(filename, file::write_only, ec);
+			size_type offset = 0;
+			while (to_write > 0)
+			{
+				int s = (std::min)(to_write, 300000);
+				file::iovec_t b = { random_data, s};
+				f.writev(offset, &b, 1, ec);
+				offset += s;
+				to_write -= s;
+			}
 		}
 
-		add_files(fs, "./tmp1_web_seed/test_torrent_dir");
+		add_files(fs, "tmp1_web_seed/test_torrent_dir");
 		free(random_data);
 	}
 	else
 	{
 		piece_size = 64 * 1024;
-		char* random_data = (char*)malloc(64 * 1024 * 50);
-		std::srand(10);
-		std::generate(random_data, random_data + 64 * 1024 * 50, &std::rand);
-		save_file("./tmp1_web_seed/seed", random_data, 64 * 1024 * 50);
-		fs.add_file("seed", 64 * 1024 * 50);
+		char* random_data = (char*)malloc(64 * 1024 * 25);
+		std::generate(random_data, random_data + 64 * 1024 * 25, &std::rand);
+		save_file("tmp1_web_seed/seed", random_data, 64 * 1024 * 25);
+		fs.add_file("seed", 64 * 1024 * 25);
 		free(random_data);
 	}
 
-	int port = start_web_server();
+	int port = start_web_server(strcmp(protocol, "https") == 0, chunked_encoding);
+
+	// generate a torrent with pad files to make sure they
+	// are not requested web seeds
+	libtorrent::create_torrent t(fs, piece_size, 0x4000, libtorrent::create_torrent::optimize
+		| libtorrent::create_torrent::calculate_file_hashes);
 
-	libtorrent::create_torrent t(fs, piece_size);
 	char tmp[512];
 	if (test_url_seed)
 	{
@@ -232,13 +273,21 @@ int run_suite(char const* protocol, bool test_url_seed)
 	}
 	else
 	{
-		snprintf(tmp, sizeof(tmp), "http://127.0.0.1:%d/seed", port);
+		snprintf(tmp, sizeof(tmp), "%s://127.0.0.1:%d/seed", protocol, port);
 		t.add_http_seed(tmp);
 	}
+	fprintf(stderr, "testing: %s\n", tmp);
+
+	for (int i = 0; i < fs.num_files(); ++i)
+	{
+		file_entry f = fs.at(i);
+		fprintf(stderr, "  %04x: %d %s\n", int(f.offset), f.pad_file, f.path.c_str());
+	}
+
+//	for (int i = 0; i < 1000; ++i) sleep(1000);
 
-	error_code ec;
 	// calculate the hash for all pieces
-	set_piece_hashes(t, "./tmp1_web_seed", ec);
+	set_piece_hashes(t, "tmp1_web_seed", ec);
 
 	if (ec)
 	{
@@ -248,30 +297,90 @@ int run_suite(char const* protocol, bool test_url_seed)
 		return 0;
 	}
 
+	if (test_ban)
+	{
+		// corrupt the files now, so that the web seed will be banned
+		if (test_url_seed)
+		{
+			char* random_data = (char*)malloc(300000);
+			for (int i = 0; i != sizeof(file_sizes)/sizeof(file_sizes[0]); ++i)
+			{
+				std::generate(random_data, random_data + 300000, &std::rand);
+				char filename[200];
+				snprintf(filename, sizeof(filename), "tmp1_web_seed/test_torrent_dir/test%d", i);
+				int to_write = file_sizes[i];
+				file f(filename, file::write_only, ec);
+				size_type offset = 0;
+				while (to_write > 0)
+				{
+					int s = (std::min)(to_write, 300000);
+					file::iovec_t b = { random_data, s};
+					f.writev(offset, &b, 1, ec);
+					offset += s;
+					to_write -= s;
+				}
+			}
+			free(random_data);
+		}
+		else
+		{
+			piece_size = 64 * 1024;
+			char* random_data = (char*)malloc(64 * 1024 * 25);
+			std::generate(random_data, random_data + 64 * 1024 * 25, &std::rand);
+			save_file("tmp1_web_seed/seed", random_data, 64 * 1024 * 25);
+			free(random_data);
+		}
+	}
+
 	std::vector<char> buf;
 	bencode(std::back_inserter(buf), t.generate());
 	boost::intrusive_ptr<torrent_info> torrent_file(new torrent_info(&buf[0], buf.size(), ec));
 
+	// no point in testing the hashes since we know the data is corrupt
+	if (!test_ban)
+	{
+		// verify that the file hashes are correct
+		for (int i = 0; i < torrent_file->num_files(); ++i)
+		{
+			sha1_hash h1 = torrent_file->file_at(i).filehash;
+			sha1_hash h2 = file_hash(combine_path("tmp1_web_seed"
+				, torrent_file->file_at(i).path));
+//			fprintf(stderr, "%s: %s == %s\n"
+//				, torrent_file->file_at(i).path.c_str()
+//				, to_hex(h1.to_string()).c_str(), to_hex(h2.to_string()).c_str());
+			TEST_EQUAL(h1, h2);
+		}
+	}
+
 	for (int i = 0; i < 6; ++i)
-		test_transfer(torrent_file, i, port, protocol, test_url_seed);
+		test_transfer(torrent_file, i, port, protocol, test_url_seed, chunked_encoding, test_ban);
 	
 	if (test_url_seed)
 	{
-		torrent_file->rename_file(0, "./tmp2_web_seed/test_torrent_dir/renamed_test1");
-		test_transfer(torrent_file, 0, port, protocol, test_url_seed);
+		torrent_file->rename_file(0, "tmp2_web_seed/test_torrent_dir/renamed_test1");
+		test_transfer(torrent_file, 0, port, protocol, test_url_seed, chunked_encoding, test_ban);
 	}
 
 	stop_web_server();
-	remove_all("./tmp1_web_seed");
+	remove_all("tmp1_web_seed", ec);
 	return 0;
 }
 
 int test_main()
 {
 	int ret = 0;
-	for (int i = 0; i < 2; ++i)
+	for (int url_seed = 0; url_seed < 2; ++url_seed)
 	{
-		run_suite("http", i);
+		for (int chunked = 0; chunked < 2; ++chunked)
+		{
+			for (int ban = 0; ban < 2; ++ban)
+			{
+#ifdef TORRENT_USE_OPENSSL
+				run_suite("https", url_seed, chunked, ban);
+#endif
+				run_suite("http", url_seed, chunked, ban);
+			}
+		}
 	}
 	return ret;
 }
